/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"

/************************************************************************/
/************************************************************************/

#define MEMO_RECIPIENT   0
#define MEMO_SENDER      1
#define MEMO_CHAN        2
#define MEMO_MSG         3
#define MEMO_TIME        4
#define MEMO_NEW         5
#define MEMO_DEL         6
#define MEMO_RECEIPT     7


static void do_csend(const char *source);
static void do_del(const char *source);
static void do_delnews(const char *source);
static void do_forward(const char *source);
static void do_global(const char *source);
static void help(const char *source);
static void do_list(const char *source);
static void ohelp(const char *source);
static void do_osend(const char *source);
static void do_purge(const char *source);
static void do_read(const char *source);
static void do_resetnews(const char *source);
static void do_send(const char *source);
static void do_sent(const char *source);
static void do_set(const char *source);
static void do_set_emailmemos(const char *source, NickInfo *ni);
static long do_set_nomemo(const char *source, NickInfo *ni);
static void do_set_notify(const char *source, NickInfo *ni);
static long do_set_receipt(const char *source, NickInfo *ni);
static void do_undel(const char *source);


/************************************************************************/

typedef struct {
	const char *cmd;
	void (*func)(const char *source);
	unsigned int access;
	unsigned int readonly;
	const char *desc;
} Command;


static Command commands[] = {
	{ "CSEND",	do_csend,	BASIC,	1,	MS_OHELP_CSEND },
	{ "DEL",	do_del,		BASIC,	1,	MS_OHELP_DEL },
	{ "DELNEWS",	do_delnews,	SRA,	1,	MS_OHELP_DELNEWS },
	{ "FORWARD",	do_forward,	BASIC,	1,	MS_OHELP_FORWARD },
	{ "MASSMEMO",	do_global,	SRA,	1,	MS_OHELP_GLOBAL },
	{ "OHELP",	ohelp,		OPER,	0,	MS_OHELP_OHELP },
	{ "HELP",	help,		BASIC,	0,	MS_OHELP_HELP },
	{ "NEWS",	do_news,	BASIC,	0,	MS_OHELP_NEWS },
	{ "PURGE",	do_purge,	BASIC,	1,	MS_OHELP_PURGE },
	{ "READ",	do_read,	BASIC,	0,	MS_OHELP_READ },
	{ "RESETNEWS",	do_resetnews,	SRA,	1,	MS_OHELP_RESETNEWS },
	{ "SEND",	do_send,	BASIC,	1,	MS_OHELP_SEND },
	{ "OSEND",	do_osend,	SOP,	1,	MS_OHELP_OSEND },
	{ "SENT",	do_sent,	BASIC,	0,	MS_OHELP_SENT },
	{ "SET",	do_set,		BASIC,	1,	MS_OHELP_SET },
	{ "UNDEL",	do_undel,	BASIC,	1,	MS_OHELP_UNDEL },

	{ NULL }

};

int check_ms_command(const char *cmd) {
	Command *c;
	int i=0;

	for (c = commands; c->cmd; c++) {
		if (sstrcmp(cmd, c->cmd) == 0) {
			return i;
		}
		i++;
	}
	return -1;
}

inline void update_ms_table(int cmd, int newlevel) {
	commands[cmd].access = newlevel;
}


/************************************************************************/

/* If id is true, this user is identified to this user nick. If id is false, only tell about new memos
   if they are recognized. If not identified or recognized, do nothing. */

void check_memos(User *user, int id)
{
	long num_rows;

	if (!user || (!id && !(user->mode & U_RECOGNIZED)))
		return;
	

	num_rows = sql_select("recipient FROM memos WHERE recipient = '%s' AND new=1", user->nick);

	if (!num_rows)
		return;
	else
		notice(s_MemoServ, user->nick, MS_NEWMEMOS, num_rows);

	user->mode |= U_MEMO_NOTIFIED;
}

/************************************************************************/

static void check_receipt(const char *source)
{
	NickInfo *ni = findnick(row[MEMO_SENDER]);

	if (ni && (ni->flags & MS_MEMO_RCPT)) {
		User *u = finduser(ni->nick);

		if (row[MEMO_CHAN])
			send_sql("INSERT INTO memos VALUES "
				"('%s','%s','%s','%s',%d,1,0,1)",
				row[MEMO_SENDER], row[MEMO_RECIPIENT],
				row[MEMO_CHAN], row[MEMO_MSG], time(NULL));
		else
			send_sql("INSERT INTO memos VALUES "
				"('%s','%s',NULL,'%s',%d,1,0,1)",
				row[MEMO_SENDER], row[MEMO_RECIPIENT], row[MEMO_MSG], time(NULL));

		if (u)
			check_memos(u, is_identified(ni->nick, ni->nick) ? 1 : 0);

		else if (sstrcmp(source, row[MEMO_SENDER]) == 0)  /* reading owner's own memo */
			check_memos(finduser(source), 1);
	}
}

/************************************************************************/

void email_memo(NickInfo *ni, const char *memo, const char *sender)
{
	char esubj[256];
	char ememo[NICKMAX + 640];

	if (!ni || !memo)
		return;

	if (!sql_select("email FROM nick WHERE name = '%s'", ni->nick))
		return;

	row = fetch(res);

	snprintf(esubj, sizeof(esubj), "New %s memo.", NETWORK_NAME);
	snprintf(ememo, sizeof(ememo), "You have a new memo sent by '%s' from %s. \n"
			"Memo: %s", sender, NETWORK_NAME, memo);

	e_mail(ememo, esubj, row[0], 1);
}



/************************************************************************/


void memoserv(const char *source, char *buf)
{
	Command *c;
	const char *cmd = gettok(buf, " ", CMDMAX);
	int valid=0;


	if (!cmd)
		return;


#ifdef USE_SQL
	if (!sql_alive) {
		notice(s_MemoServ, source, "Sorry, we're currently experiencing temporary "
			"technical difficulties. Please try again shortly.");
		return;
	}
#endif

	if (is_frozen(source) && !is_oper(source)) {
		notice(s_MemoServ, source,
			"Your nick is currently frozen and may not be used.");
		return;
	}

	if (sstrcmp(cmd, "LIST") == 0) {
		do_list(source);
		return;
	}


	for (c = commands; c->cmd; c++) {
		if (sstrcmp(c->cmd, cmd) == 0) {
			valid=1;
		if (READ_ONLY && c->readonly)
			notice(s_MemoServ, source,
				"Sorry, that command is temporarly disabled while services are in "
				"read-only mode.");
			else if (!c->access || has_access(source, c->access))
				c->func(source);
			else
				notice(s_MemoServ, source, "Permission denied.");
			break;
		}
	}

	if (!valid)
		notice(s_MemoServ, source,
			"Unrecognized command \2%s\2. Type \"/%s HELP\" for help.",
			cmd, s_MemoServ);


}


/************************************************************************/

static void do_csend(const char *source)
{
	ChanInfo *ci;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *list = strtok(NULL, " ");
	const char *memo = strtok(NULL, "");
	char *ntrace;
	int acc=0, target=0;


	if (list)
		target = !sstrcmp(list, "ALL") ? 0 : !sstrcmp(list, "AVOICE") ? 3 :
			!sstrcmp(list, "AOP") ? 5 : !sstrcmp(list, "SOP") ? 10 :
			!sstrcmp(list, "CF") ? 13 : !sstrcmp(list, "CFOUNDER") ? 13 : -1;

	if (!u)
		return;

	if (!chan || !list || !memo) {
		notice(s_MemoServ, source,
                        "Syntax: /%s CSEND <#channel> <list> <message>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP CSEND\2 for more information.",
                        s_MemoServ);

	} else if (time(NULL) < (u->lastmemo + MEMO_WAIT)) {
		notice(s_MemoServ, source, MS_MEMO_WAIT);

	} else if (target==-1) {
		notice(s_MemoServ, source,
			"CSEND ERROR: No such list '%s' for CSEND. Must be AVOICE, AOP, SOP, "
			"CFOUNDER or ALL", list);


	} else if (!(ci = findchaninfo(chan))) {
		notice(s_MemoServ, source,
			"CSEND ERROR: Channel %s is not registered!", chan);

	} else if (is_frozen(chan)) {
		notice(s_MemoServ, source,
			"CSEND ERROR: Channel %s is frozen and may not receive memos.", chan);

	} else if (chan_closed(chan)) {
		notice(s_MemoServ, source,
			"CSEND ERROR: Channel %s is closed and may not be used.", chan);


	} else if ((acc = get_access(u, chan, &ntrace, 0)) < 3) {
		notice(s_MemoServ, source,
			"CSEND ERROR: You do not have access to channel %s to send channel memos!", chan);


	} else if (!ci->mlevel) {
		notice(s_MemoServ, source,
			"CSEND ERROR: Channel %s has disabled channel memos.", chan);


	} else if (strlen(memo) > (size_t)MAX_MEMO_LENGTH) {
		notice(s_MemoServ, source,
			"CSEND ERROR: Length of memo is too long! (Max: %d  Length: %d)",
			MAX_MEMO_LENGTH, strlen(memo));


	} else if (acc < ci->mlevel) {
		notice(s_MemoServ, source,
			"CSEND ERROR: The channel memo level for %s is set to %s. You are only "
			"recognized as having %s access.", chan,
			ci->mlevel==5 ? "AOP" : ci->mlevel==10 ? "SOP" : ci->mlevel==13 ? "CO-FOUNDER" :
			ci->mlevel==15 ? "FOUNDER" : "",
			acc==5 ? "AOP" : acc==10 ? "SOP" : acc==13 ? "CO-FOUNDER" : acc==15 ? "FOUNDER" : "");

	} else {
		NickInfo *ni, *nif;
		struct AccessList *a;
		char memobuf[512 + CHANMAX + 20];
		int sends=0, num_memos=0;
		time_t sendtime = time(NULL);  /* to ensure all memos have the same time */

		u->lastmemo = time(NULL);
		u->memos++;

		snprintf(memobuf, sizeof(memobuf), "%s (%s+) >> %s", chan, 
			target==0 ? "AVOICE" : target==3 ? "AVOICE" : target==5 ? "AOP" :
			target==10 ? "SOP" : target==13 ? "CFOUNDER" : "AVOICE", memo);


		ni = findnick(ci->founder);

		if (ni && ((nif = ni->forward)))
			ni = nif;

		if (ni && (ni->flags & MS_NOMEMO)) {
			num_memos = sql_select(
				"recipient FROM memos WHERE recipient = '%s'", ni->nick);

			if (num_memos <= MAX_MEMOS) {
				sends=1;

				send_sql("INSERT INTO memos VALUES "
					"('%s','%s','%s','%s',%d,1,0,0)",
					ni->nick, ntrace, chan, memobuf, sendtime);

				if (ni->flags && MS_MEMO_NEW && finduser(ni->nick)) {
					notice(s_MemoServ, ni->nick, MS_NEWMEMO);
					notice(s_MemoServ, ni->nick, MS_NEWMEMO2,
						s_MemoServ, num_memos+1);
				}
				if (ni->flags & MS_EMAILMEMOS)
					email_memo(ni, memobuf, ntrace);

			}
		}


		for (a = ci->entries; a; a = a->next) {
			if (a->level >= target) {
				const char *ntarget;
				num_memos=0;

				if ((ni = findnick(a->entry))) {
					if (ni->flags & MS_NOMEMO)
						continue;

					nif = ni->forward;

					if (nif && (nif->flags & MS_NOMEMO))
						continue;

					ntarget = nif ? nif->nick : a->entry;

					num_memos = sql_select(
						"recipient FROM memos WHERE recipient = '%s'", ntarget);

					if (num_memos > MAX_MEMOS)
						continue;

					sends++;

					send_sql("INSERT INTO memos VALUES "
						"('%s','%s','%s','%s',%d,1,0,0)",
						ntarget, ntrace, chan, memobuf, sendtime);


					if (((!nif && ni->flags & MS_MEMO_NEW) || (nif && nif->flags & MS_MEMO_NEW))
						&& finduser(ntarget)) {
						notice(s_MemoServ, ntarget, MS_NEWMEMO);
						notice(s_MemoServ, ntarget, MS_NEWMEMO2,
							s_MemoServ, num_memos+1);
					}
					if ((!nif && ni->flags & MS_EMAILMEMOS) || (nif && nif->flags & MS_EMAILMEMOS))
						email_memo(nif ? nif : ni, memobuf, source);

				}
			}
		}
		if (!sends)
			notice(s_MemoServ, source,
				"No memos could be sent for the channel %s. "
				"This could be because there is no entries on the channel access lists, "
				"or because the targets all do not wish to have memos being sent to them "
				"or because their memobox is full.", chan);
		else {
			notice(s_MemoServ, source,
				"%s+ Memo has been sent to channel %s. (%d memos)", list, chan, sends);

			notice(s_MemoServ, source,
				"Note: Some users may not receive this memo depending upon if they have "
				"the NOMEMO option set, or if their memobox is full.");

		}

	}
}

/************************************************************************/

static void do_del(const char *source)
{
	const char *num = gettok(NULL, " ", 10);

	if (!num) {
		notice(s_MemoServ, source,
                        "Syntax: /%s DEL <memo number>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP DEL\2 for more information.",
                        s_MemoServ);

	} else if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"DEL ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"DEL ERROR: You are not identified to your current nick '%s'!", source);


	} else if (sstrcmp(num, "ALL") == 0) {
		int num_memos = sql_select("* FROM memos WHERE recipient = '%s'", source);

		if (!num_memos)
			notice(s_MemoServ, source,
				"DEL ERROR: You have no memos to delete!");
		else {
			notice(s_MemoServ, source,
				"All your memos have been deleted and purged. (%d memo%s)",
				num_memos, num_memos==1 ? "" : "s");
			send_sql("DELETE FROM memos WHERE recipient = '%s'", source);
		}

	} else if (sstrcmp(num, "OLD") == 0) {
		int num_memos = sql_select("* FROM memos WHERE recipient = '%s' AND new=0 AND del=0", source);

		if (!num_memos)
			notice(s_MemoServ, source,
				"DEL ERROR: You have no old memos to delete!");
		else {
			notice(s_MemoServ, source,
				"All old memos have been \2flagged\2 as deleted. (%d memo%s)",
				num_memos, num_memos==1 ? "" : "s");
			send_sql("UPDATE memos SET del=1 WHERE recipient = '%s' AND new=0", source);
		}


	} else if (!strspn(num, "1234567890") || (atoi(num) < 1)) {
		notice(s_MemoServ, source,
			"DEL ERROR: Invalid memo number.");
	} else {
		int num_memos=sql_select("* FROM memos WHERE recipient = '%s' ORDER BY time", source);
		
		if (!num_memos)
			notice(s_MemoServ, source,
				"DEL ERROR: You have no memos to delete.");

		else if (atoi(num) > num_memos)
			notice(s_MemoServ, source,
				"DEL ERROR: Memo number #%s doesn't exist.", num);
		else {
			int i;
			for (i=1; i <= num_memos; i++) {
				row = fetch(res);
				if (i == atoi(num)) {
					if (atoi(row[MEMO_DEL])) {
						notice(s_MemoServ, source,
							"Memo #%s already marked as deleted!", num);
						return;
					}

					notice(s_MemoServ, source,
						"Memo #%s marked as deleted.", num);
					send_sql("UPDATE memos SET del=1 WHERE recipient='%s' AND "
						"sender='%s' AND time=%d", source,
						row[MEMO_SENDER], atol(row[MEMO_TIME]));
					return;
				}
			}
			notice(s_MemoServ, source,
				"DEL ERROR: Memo number #%s not found.", num);
		}
	}
}

/************************************************************************/

static void do_delnews(const char *source)
{
	const char *report = gettok(NULL, " ", 10);

	if (!report) {
		notice(s_MemoServ, source,
                        "Syntax: /%s DELNEWS <index number>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP DELNEWS\2 for more information.",
                        s_MemoServ);

	} else if (!strspn(report, "1234567890") || atoi(report) < 0) {
		notice(s_MemoServ, source,
			"DELNEWS ERROR: Invalid news index number.");

	} else if (!sql_select("* FROM news WHERE id = %s", report)) {
		notice(s_MemoServ, source,
			"DELNEWS ERROR: Index number #%s not found!", report);

	} else {
		row = fetch(res);

		if (!row[0]) {
			notice(s_MemoServ, source,
				"DELNEWS ERROR: Index number #%s not found!", report);
		} else {
			send_sql("DELETE FROM news WHERE id = %s", report);
			notice(s_MemoServ, source,
				"News index #%s has been deleted.", report);

			load_news();
		}
	}

}

/************************************************************************/

static void do_forward(const char *source)
{
	User *u = finduser(source);
	NickInfo *ni, *nif;
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *pass = strtok(NULL, " ");
	int ident=0;


	if (!u)
		return;

	if (!(ni = findnick(source))) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: Your nick '%s' is not registered! You need to be using a registered "
			"nick you are identified to before using the FORWARD command.", source);

	} else if (!is_identified(source, source)) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: You must be identified to your current nick before using FORWARD! "
			"Try again after typing /%s IDENTIFY <your-password>", s_NickServ);


	} else if (!nick && !pass) {
		if (!sql_select("forward FROM nick WHERE name = '%s'", source)) {
			notice(s_MemoServ, source,
				"FORWARD ERROR: Unknown error - please try again later.");
			return;
		}
		row = fetch(res);
		if (*row[0] == '0') {
			notice(s_MemoServ, source,
				"Your nick is not set to forward memos.");
		} else {                
			notice(s_MemoServ, source,
				"Your memos will now no longer be forwarded to '%s'.", row[0]);
			send_sql("UPDATE nick SET forward=0 WHERE id = %d", ni->id);
			ni->forward = NULL;
				/* forward=0 should be fine since nicks can't start with a number. */
		}


	} else if (ni->flags & MS_NOMEMO) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: You may not forward memos when you have the NOMEMO "
			"option set!", nick);


	} else if (sstrcmp(nick, source) == 0) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: You may not forward memos to the same nick!", nick);

	} else if (!(nif = findnick(nick))) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: The nick '%s' is not registered!", nick);

	} else if (is_frozen(nick)) {
		notice(s_MemoServ, source,
			"FORWARD ERROR: The nick '%s' is frozen and may not be sent memos!", nick);


	} else if (!pass && !(ident=is_identified(source, nick))) {
			notice(s_MemoServ, source,
				"FORWARD ERROR: You may not forward memos to another nick unless "
				"you have identified to that nick, or supply the password after "
				"the nick in the FORWARD command.");

	} else if (pass || ident) {
		if (pass) {
			if (!sql_select("pass FROM nick WHERE name = '%s'", nick)) {
				notice(s_MemoServ, source,
					"FORWARD ERROR: Unknown error - please try again later.");
				return;
			} else {
				row = fetch(res);
				if (row[0] && strcmp(row[0], pass) != 0) { /* case sensitive */
					notice(s_MemoServ, source,
					"FORWARD ERROR: Invalid password for nick '%s'.", nick);
					blog("NS_I", "[NS] I* %s (%s!%s@%s) [MS FORWARD]",
						nick, source, u->username, u->host);
					bad_password(u, source);
					return;
				} else {
					add_idnick(u, nick);
					generateid(u,nick,nif->id);
				}
			}
		}
		ni->forward = nif;
		send_sql("UPDATE nick SET forward = '%s' WHERE id = %d", nick, ni->id);
		notice(s_MemoServ, source,
			"Your memos will now be forwarded to '%s'", nick);
	}
}

/************************************************************************/

static void do_global(const char *source)
{
	User *u = finduser(source);
	const char *memo = strtok(NULL, "");


	if (!u)
		return;

	if (!memo) {
		notice(s_MemoServ, source,
                        "Syntax: /%s MASSMEMO <memo>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP MASSMEMO\2 for more information.",
                        s_MemoServ);

	} else if (time(NULL) < (u->lastmemo + MEMO_WAIT)) {
		notice(s_MemoServ, source, MS_MEMO_WAIT);

	} else if (strlen(memo) > (size_t)MAX_MEMO_LENGTH) {
		notice(s_MemoServ, source,
			"SEND ERROR: Length of memo is too long! (Max: %d  Length: %d)",
			MAX_MEMO_LENGTH, strlen(memo));

	} else {
		NickInfo *ni, *nif;
		int num_rows, num_memos=0, i;
		char buf[NICKMAX + 256];

		u->lastmemo = time(NULL);
		u->memos++;

		num_rows = sql_select("name FROM nick");

		for (i=0; i < num_rows; i++) {
			row = fetch(res);

			if (!(ni = findnick(row[0])))
				continue;

			if (!(nif = ni->forward))
				nif = ni;
			if (nif->flags & MS_NOMEMO)
				continue;

			snprintf(buf, sizeof(buf),
				"SELECT recipient FROM memos WHERE recipient = '%s'", nif->nick);

			if (!mysql_query(&mysql, buf)) {
				if ((res2 = mysql_store_result(&mysql)) != NULL) {
					if (mysql_num_rows(res2) >= 50)
						continue;
					mysql_free_result(res2);
				}
			}


			send_sql("INSERT INTO memos VALUES "
				"('%s','%s',NULL,'%s',%d,1,0,0)",
				nif->nick, source, memo, time(NULL));

			if ((nif->flags & MS_MEMO_NEW) && finduser(nif->nick)) {
				notice(s_MemoServ, nif->nick, MS_NEWMEMO);
				notice(s_MemoServ, nif->nick, MS_NEWMEMO2, s_MemoServ, num_memos+1);
			}

			if (nif->flags & MS_EMAILMEMOS)
				email_memo(nif, memo, source);

		}
		notice(s_MemoServ, source, "All registered nicks have been sent the memo.");
		globops(s_MemoServ, "%s has used the MASSMEMO command.", opertrace(u,0));
	}
}

/************************************************************************/


static void help(const char *source)
{
	char *param = gettok(NULL, "", CMDMAX);
	char *s, *topic;
	int level=0;


	if (param) {
		s = sstrdup(param);
		topic = strtok(s, " ");
		if (topic) {
			int i=0;
			if ((i=check_ms_command(topic)) > -1)
				level = commands[i].access;

		}
		free(s);
	}
	do_help(s_MemoServ, source, param, level);


}

/************************************************************************/

static void ohelp(const char *source)
{
	Command *c;
	int shown=0;
	unsigned int level;
	User *u = finduser(source);

	if (!u)
		return;

	notice(s_MemoServ, source,
		"This is a listing of %s's privileged commands.", s_MemoServ);
	notice(s_MemoServ, source,
		"Your current access level is: %s.",
		u->os_level==0 ? "IRC Operator" :
		u->os_level==1 ? "Services Operator" :
		u->os_level==2 ? "Services Administrator" :
		u->os_level==3 ? "Services Root" :
		u->os_level>3 ? "Services Root Master" : "");

	notice(s_MemoServ, source, "For a detailed explanation for a command, type:");
	notice(s_MemoServ, source, "\2/%s HELP <command>", s_MemoServ);

	for (level=1; level < 5; level++, shown=0) {
		notice(s_MemoServ, source, " ");
		notice(s_MemoServ, source, "%s (%s)",
			level==1 ? "Commands available for \2IRC Operators\2:" :
			level==2 ? "Commands available for \2Services Operators\2:" :
			level==3 ? "Commands available for \2Services Administrators\2:" :
			level>3 ? "Commands available for \2Services Roots\2:" : "",
			u->os_level+1 >= level ? "Access Granted" : "Access Denied");
		notice(s_MemoServ, source, " ");

		for (c = commands; c->cmd; c++) {
			if (c->access == level) {
				notice(s_MemoServ, source, "    \2%s\2   %s\n", c->cmd, c->desc);
				shown++;
			}
		}
		if (!shown)
			notice(s_MemoServ, source, "    No commands available.");
	}
	notice(s_MemoServ, source, " ");
	notice(s_MemoServ, source, "Note: Most of these commands are logged!");
	notice(s_MemoServ, source, "*** END OF HELP ***");
}


/************************************************************************/

static void do_list(const char *source)
{
	const char *param = strtok(NULL, " ");
	int num_rows, all=1;

	if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"LIST ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"LIST ERROR: You are not identified to your current nick '%s'!", source);

	} else if (param && !(all=sstrcmp(param, "NEW")) &&
		 !sql_select("* FROM memos WHERE recipient = '%s' AND new=1", source)) {
		notice(s_MemoServ, source,
			"You do not have any new memos.");

	} else if (!(num_rows = sql_select("* FROM memos WHERE recipient = '%s' ORDER BY time", source))) {
		notice(s_MemoServ, source,
			"You do not have any memos.");
	} else {
		int i;

		notice(s_MemoServ, source, "\37New\37 \37Del\37 \37Num\37         \37Date/Time\37         \37Sender\37");

		for (i=0; i < num_rows && i < 25; i++) {
			row = fetch(res);
			if (all || atoi(row[MEMO_NEW]))
			notice(s_MemoServ, source,
				"[\2%c\2] [\2%c\2] (\2%d\2)   %s %s",
				atoi(row[MEMO_NEW]) ? 'x' : ' ', atoi(row[MEMO_DEL]) ? 'x' : ' ',
				i+1, timebuf(atol(row[MEMO_TIME])), row[MEMO_SENDER]);
		}
		if (num_rows > 25)
			notice(s_MemoServ, source, "25 of %d memos listed.", num_rows);
		notice(s_MemoServ, source, "Type \2/%s READ <num>\2 to read a memo.", s_MemoServ);
	}
}

/************************************************************************/

static void do_set_notify(const char *source, NickInfo *ni)
{
	const char *param = gettok(NULL, " ", CMDMAX);

	if (!param) {
		notice(s_MemoServ, source,
			"Syntax: /%s NOTIFY <option>", s_MemoServ);
		notice(s_MemoServ, source,
			"\2/%s HELP NOTIFY\2 for more information.",
			s_MemoServ);
		return;

	} else if (sstrcmp(param, "OFF") == 0) {
		ni->flags &= ~MS_MEMO_LOGON;
		ni->flags &= ~MS_MEMO_NEW;
		notice(s_MemoServ, source, "Notification of memos have been disabled for for the nick %s.",
			source);

	} else if (sstrcmp(param, "ON") == 0) {
		ni->flags |= MS_MEMO_LOGON;
		ni->flags |= MS_MEMO_NEW;
		notice(s_MemoServ, source, "You will now be notified of new memos when logging on and "
			"as they are sent to you.");

	} else if (sstrcmp(param, "NEW") == 0) {
		ni->flags &= ~MS_MEMO_LOGON;
		ni->flags |= MS_MEMO_NEW;
		notice(s_MemoServ, source, "You will now only be notified of new memos as they are sent to you.");

	} else if (sstrcmp(param, "LOGON") == 0) {
		ni->flags |= MS_MEMO_LOGON;
		ni->flags &= ~MS_MEMO_NEW;
		notice(s_MemoServ, source, "You will now only be notified of new memos when you logon.");

	} else {
		notice(s_MemoServ, source, "NOTIFY ERROR: Invalid option \2\"%s\"\2.", param);
		return;
	}
	send_sql("UPDATE nick SET flags = %d WHERE id = %d",
		ni->flags, ni->id);

}

/************************************************************************/

static void do_purge(const char *source)
{
	int num_memos=0;


	if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"PURGE ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"PURGE ERROR: You are not identified to your current nick '%s'!", source);

	} else if (!(num_memos = sql_select("* FROM memos WHERE recipient = '%s' AND del=1", source))) {
		notice(s_MemoServ, source,
			"PURGE ERROR: No memos to purge.");
	} else {
		notice(s_MemoServ, source,
			"All memos marked as deleted have been purged. (%d memo%s)",
			num_memos, num_memos==1 ? "" : "s");

		send_sql("DELETE FROM memos WHERE recipient='%s' AND del=1", source);
	}
}

/************************************************************************/

static void do_read(const char *source)
{
	const char *num = gettok(NULL, " ", 10);

	if (!num) {
		notice(s_MemoServ, source,
                        "Syntax: /%s READ <memo number>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP READ\2 for more information.",
                        s_MemoServ);


	} else if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"READ ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"READ ERROR: You are not identified to your current nick '%s'!", source);

	} else if (sstrcmp(num, "LAST") == 0) {
		int num_memos=sql_select("* FROM memos WHERE recipient = '%s' ORDER BY time", source);

		if (!num_memos)
			notice(s_MemoServ, source,
				"READ ERROR: You have no memos to read.");
		else {
			mysql_data_seek(res, num_memos-1);
			row = fetch(res);

				
			if (atoi(row[MEMO_RECEIPT])) {
				notice(s_MemoServ, source,
					"Memo \2%d\2 is a memo receipt that confirms user '%s' has "
					"read your memo on %s. Original Memo sent by '%s':", 
					num_memos, row[MEMO_SENDER], 
					timebuf(atol(row[MEMO_TIME])), row[MEMO_RECIPIENT]);
				notice(s_MemoServ, source, "%s", row[MEMO_MSG]);
			} else {
				notice(s_MemoServ, source,
					"%sMemo \2%d\2 from %s. Time/Date:%s",
					row[MEMO_CHAN] ? "Channel " : "",
					num_memos, row[MEMO_SENDER], timebuf(atol(row[MEMO_TIME])));

				notice(s_MemoServ, source, "%s", row[MEMO_MSG]);

			}

			if (atoi(row[MEMO_NEW])) {
				send_sql("UPDATE memos SET new=0 WHERE recipient='%s' AND "
					"sender='%s' AND time=%d",
					source, row[MEMO_SENDER], atol(row[MEMO_TIME]));
				if (!atoi(row[MEMO_RECEIPT]))
					check_receipt(source);

				if (!sql_select("new FROM memos WHERE recipient='%s' AND new=1"))
					finduser(source)->mode &= ~U_MEMO_NOTIFIED;
			}
		}


	} else if (!strspn(num, "1234567890") || (atoi(num) < 1)) {
		notice(s_MemoServ, source,
			"READ ERROR: Invalid memo number.");

	} else {
		int num_memos=sql_select("* FROM memos WHERE recipient = '%s' ORDER BY time", source);
		
		if (!num_memos)
			notice(s_MemoServ, source,
				"READ ERROR: You have no memos to read.");

		else if (atoi(num) > num_memos)
			notice(s_MemoServ, source,
				"READ ERROR: Memo number #%s doesn't exist.", num);
		else {
			int i;
			for (i=1; i <= num_memos; i++) {
				row = fetch(res);
				if (i == atoi(num)) {
					if (atoi(row[MEMO_RECEIPT])) {
						notice(s_MemoServ, source,
							"Memo \2%d\2 is a memo receipt that confirms user '%s' has "
							"read your memo on %s. Original Memo sent by '%s':", 
							i, row[MEMO_SENDER], 
							timebuf(atol(row[MEMO_TIME])), row[MEMO_RECIPIENT]);
						notice(s_MemoServ, source, "%s", row[MEMO_MSG]);
					} else {

						notice(s_MemoServ, source,
							"%sMemo \2%d\2 from %s. Time/Date:%s",
							row[MEMO_CHAN] ? "Channel " : "",
							i, row[MEMO_SENDER], timebuf(atol(row[MEMO_TIME])));

						notice(s_MemoServ, source, "%s", row[MEMO_MSG]);
					}

					if (atoi(row[MEMO_NEW])) {
						send_sql("UPDATE memos SET new=0 WHERE recipient='%s' AND "
							"sender='%s' AND time=%d",
							source, row[MEMO_SENDER], atol(row[MEMO_TIME]));
						if (!atoi(row[MEMO_RECEIPT]))
							check_receipt(source);
					}
					return;
				}
			}
			notice(s_MemoServ, source,
				"READ ERROR: Memo number #%s not found.", num);
		}
	}

}

/************************************************************************/

static long do_set_receipt(const char *source, NickInfo *ni)
{
	const char *param = gettok(NULL, " ", CMDMAX);

	if (!param) {
		if (ni->flags & MS_MEMO_RCPT)
			notice(s_MemoServ, source, "Memo receipts are currently activated for your nick.");
		else
			notice(s_MemoServ, source, "You are currently not set to be sent memo receipts.");
		return 0;

	} else if (sstrcmp(param, "OFF") == 0) {
		ni->flags &= ~MS_MEMO_RCPT;
		notice(s_MemoServ, source, "Memo receipts will no longer be sent to this nick.");

	} else if (sstrcmp(param, "ON") == 0) {
		ni->flags |= MS_MEMO_RCPT;
		notice(s_MemoServ, source, "You will now be notified when your memo recipients read "
			"their memos.");

	} else {
		notice(s_MemoServ, source, "RECEIPT ERROR: Invalid option \2\"%s\"\2.", param);
		return 0;
	}
	send_sql("UPDATE nick SET flags = %d WHERE id = %d",
		ni->flags, ni->id);
	return ni->flags;

}

/************************************************************************/

static void do_resetnews(const char *source)
{
	send_sql("UPDATE sirv,nick SET sirv.news=0, nick.news=0");
	send_sql("DELETE FROM news");

	notice(s_MemoServ, source, "News database has been reset.");
	globops(s_MemoServ, "%s has reset the news database.", opertrace(finduser(source),0));

	load_news();
}

/************************************************************************/

static void do_osend(const char *source)
{
	User *u = finduser(source);
	const char *paccess = strtok(NULL, " ");
	const char *memo = strtok(NULL, "");
	int oaccess=0, up=0;

	if (!u)
		return;

	if (paccess) {
		if (*paccess == '+') {
			*paccess++;
			up=1;
		}
		oaccess = !sstrcmp(paccess, "SOP") ? 1 : !sstrcmp(paccess, "SA") ? 2 :
			!sstrcmp(paccess, "SRA") ? 3 : !sstrcmp(paccess, "ALL") ? 1 : -1;

						/* Masters should get "SRA" and "+SRA" memos */
		if (sstrcmp(paccess, "ALL") == 0 || sstrcmp(paccess, "SRA") == 0)
			up=1;
	}


	if (!paccess || !memo) {
		notice(s_MemoServ, source,
                        "Syntax: /%s OSEND <access> <memo>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP OSEND\2 for more information.",
                        s_MemoServ);

	} else if (time(NULL) < (u->lastmemo + MEMO_WAIT)) {
		notice(s_MemoServ, source, MS_MEMO_WAIT);

	} else if (strlen(memo) > (size_t)MAX_MEMO_LENGTH) {
		notice(s_MemoServ, source,
			"SEND ERROR: Length of memo is too long! (Max: %d  Length: %d)",
			MAX_MEMO_LENGTH, strlen(memo));

	} else if (oaccess < 0) {
		notice(s_MemoServ, source,
			"OSEND ERROR: Unknown access list '%s'", paccess);

	} else {
		NickInfo *ni;
		char memobuf[524];
		time_t sendtime = time(NULL);
		int i, num_rows;

		u->lastmemo = time(NULL);
		u->memos++;

		snprintf(memobuf, sizeof(memobuf), "%s%s >> %s",
			oaccess==0 ? "SOP" : oaccess==1 ? "SOP" :
			oaccess==2 ? "SA" : oaccess==3 ? "SRA" : "SOP",
			up ? "+" : "", memo);


		num_rows = sql_select("name FROM oper WHERE access %s %d",
			up ? ">=" : "=", oaccess);

		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			ni = findnick(row[0]);
			if (!ni)
				continue;

			send_sql("INSERT INTO memos VALUES "
				"('%s','%s',NULL,'%s',%d,1,0,0)",
				ni->nick, u->os_nick, memobuf, sendtime);

			if (finduser(ni->nick)) {
				notice(s_MemoServ, ni->nick,
					"You have a new oper-access memo.");
			}

		}

		notice(s_MemoServ, source,
			"Memo has been sent to all %s%s access opers.", up ? "+" : "",
			oaccess==1 ? "SOP" : oaccess==2 ? "SA" : oaccess==3 ? "SRA" : "SOP");
	}
}

/************************************************************************/

static void do_send(const char *source)
{
	NickInfo *ni;
	User *u = finduser(source);
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *memo = strtok(NULL, "");
	int force=0;

	if (!u)
		return;

	if ((u->os_level > (unsigned int)FORCE_MEMO_LEVEL) && nick && (*nick == '+')) { *nick++; force=1; }

	if (!nick || !memo) {
		notice(s_MemoServ, source,
                        "Syntax: /%s SEND <nick> <memo>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP SEND\2 for more information.",
                        s_MemoServ);

	} else if (time(NULL) < (u->lastmemo + MEMO_WAIT)) {
		notice(s_MemoServ, source, MS_MEMO_WAIT);

	} else if (strlen(memo) > (size_t)MAX_MEMO_LENGTH) {
		notice(s_MemoServ, source,
			"SEND ERROR: Length of memo is too long! (Max: %d  Length: %d)",
			MAX_MEMO_LENGTH, strlen(memo));


	} else if (!is_identified(source, source)) {
		notice(s_MemoServ, source,
			"SEND ERROR: You must be identified to your current nick in order to send "
			"memos to others.");

	} else if (AUTH_EMAIL && !is_authed(source)) {
		notice(s_MemoServ, source,
			"SEND ERROR: You must authorize your nick before you may send memos.");


	} else if (!(ni = findnick(nick))) {
		notice(s_MemoServ, source,
			"SEND ERROR: Nick '%s' is not registered!", nick);


	} else if (!force && ni->flags & MS_NOMEMO) {
		notice(s_MemoServ, source,
			"SEND ERROR: Nick '%s' does not wish to receive memos.", nick);
		if (u->os_level >= (unsigned int)FORCE_MEMO_LEVEL)
			notice(s_MemoServ, source,
				"Oper notice: If you wish to force this memo, type: "
				"\2/%s SEND +%s %s\2", s_MemoServ, nick, memo);


	} else if (AUTH_EMAIL && !is_authed(nick)) {
		notice(s_MemoServ, source,
			"SEND ERROR: User '%s' has not authorized their nickname. "
			"You may not send memos to unauthorized nicknames.", nick);


	} else {
		NickInfo *nif = ni->forward;
		int num_memos;
		const char *target;

		u->lastmemo = time(NULL);
		u->memos++;

		if (nif && (nif->flags & MS_NOMEMO)) {
			notice(s_MemoServ, source,
				"SEND ERROR: Can't send memo because nick '%s' (forwarded) does not "
				"wish to recieve memos.", nif->nick);
			return;
		}

		target = nif ? nif->nick : nick;

		num_memos=sql_select("recipient FROM memos WHERE recipient = '%s'", target);

		if ((u->os_level < (unsigned int)FORCE_MEMO_LEVEL) && num_memos >= MAX_MEMOS) {
			notice(s_MemoServ, source,
				"SEND ERROR: Nick '%s'%s has a full memo box and can't receive any more.",
				target, nif ? "(forwarded)" : "");
			return;
		}


		send_sql("INSERT INTO memos VALUES "
			"('%s','%s',NULL,'%s',%d,1,0,0)",
			target, source, memo, time(NULL));

		notice(s_MemoServ, source,
			"Memo has been sent to \2%s\2%s.", target, nif ? " (forwarded)" : "");

		if (((!nif && ni->flags & MS_MEMO_NEW) || (nif && nif->flags & MS_MEMO_NEW))
			&& finduser(target)) {
			notice(s_MemoServ, target, MS_NEWMEMO);
			notice(s_MemoServ, target, MS_NEWMEMO2, s_MemoServ, num_memos+1);
		}

		if ((!nif && ni->flags & MS_EMAILMEMOS) || (nif && nif->flags & MS_EMAILMEMOS))
			email_memo(nif ? nif : ni, memo, source);
	}

}

/************************************************************************/

static void do_sent(const char *source)
{
	char *num = gettok(NULL, " ", 10);
	char *del = strtok(NULL, " ");
	long num_memos;

	if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"SENT ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"SENT ERROR: You are not identified to your current nick '%s'!", source);

	} else if (!(num_memos=sql_select("* FROM memos WHERE sender = '%s'", source))) {
		notice(s_MemoServ, source,
			"It appears no memos currently exist on file that you have sent.");

	} else if (num) {
		if (!strspn(num, "1234567890") || (atoi(num) < 1))
			notice(s_MemoServ, source, "SENT ERROR: Invalid memo number.");

		else if (atoi(num) > num_memos)
			notice(s_MemoServ, source, "SENT ERROR: Memo number #%s not found on your send list.",num);

		else {
			mysql_data_seek(res, atoi(num)-1);
			row = fetch(res);
			if (row[0]) {
				User *recp = finduser(row[MEMO_RECIPIENT]);

				if (del && sstrcmp(del, "DEL") == 0) {
					if (!atoi(row[MEMO_NEW])) {
						notice(s_MemoServ, source,
						"You may not delete sent-memo #%s because %s has already "
						"read it.", num, row[MEMO_RECIPIENT]);
					} else if (recp && (recp->mode & U_MEMO_NOTIFIED)) {
						notice(s_MemoServ, source,
						"You may not delete sent-memo #%s because %s has already "
						"been notified of it.", num, row[MEMO_RECIPIENT]);
					} else {
						notice(s_MemoServ, source,
							"Memo #%s has been successfully retrieved and "
							"deleted. User '%s' will not recieve this memo.",
							num, row[MEMO_RECIPIENT]);

						send_sql("DELETE FROM memos WHERE recipient='%s' AND "
						"sender='%s' AND time=%d AND new=1",
						row[MEMO_RECIPIENT], row[MEMO_SENDER], atol(row[MEMO_TIME]));
					}

				} else {
					notice(s_MemoServ, source,
						"%sMemo \2%s\2 from %s to \2%s\2. Time/Date:%s",
						row[MEMO_CHAN] ? "Channel " : "",
						num, row[MEMO_SENDER], row[MEMO_RECIPIENT], 
						timebuf(atol(row[MEMO_TIME])));

					notice(s_MemoServ, source, "%s", row[MEMO_MSG]);
				}
			} else
				notice(s_MemoServ, source,
					"SENT ERROR: Could not retrieve memo #%s on your send list.", num);
		}

	} else {
		int i;

		notice(s_MemoServ, source, "Memos on file sent from '%s':", source);
		for (i=0; i < num_memos && i < 25; i++) {
			row = fetch(res);
			notice(s_MemoServ, source,
				" %d)  To: \2%s\2  Time: %s  Read: \2%s\2   Chan-Memo: %s",
				i+1, row[MEMO_RECIPIENT], timebuf(atol(row[MEMO_TIME])),
				atoi(row[MEMO_NEW]) ? "No" : "Yes",
				row[MEMO_CHAN] ? row[MEMO_CHAN] : "No");
		}
		notice(s_MemoServ, source, "End of memo listing.");
		if (num_memos > 25)
			notice(s_MemoServ, source, "25 of %d memos listed.", num_memos);
	}
}

/************************************************************************/

static void do_set(const char *source)
{
	NickInfo *ni = findnick(source);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	long showopt=0;

	if (!cmd) {
		notice(s_MemoServ, source,
			"Snytax: \2SET\2 <command> <parameters>");
		notice(s_MemoServ, source,
			"\2/%s HELP SET\2 for more information.", s_MemoServ);
	} else if (!ni) {
		notice(s_MemoServ, source,
			"Your nickname '%s' is not registered.", source);
		notice(s_MemoServ, source,
			"For registration info, type: \2/%s HELP REGISTER\2",s_NickServ);

	} else if (!is_identified(source, source)) {
		notice(s_MemoServ, source,
			"Password authentication required for the SET command.");
		notice(s_MemoServ, source,
			"Retry after typing \2/%s IDENTIFY \37password\37.", s_NickServ);



	} else if (sstrcmp(cmd, "EMAILMEMOS") == 0) {
		do_set_emailmemos(source, ni);

	} else if (sstrcmp(cmd, "NOMEMO") == 0) {
		showopt=do_set_nomemo(source, ni);

	} else if (sstrcmp(cmd, "NOTIFY") == 0) {
		do_set_notify(source, ni);

	} else if (sstrcmp(cmd, "RECEIPT") == 0) {
		showopt=do_set_receipt(source, ni);

	} else {
		notice(s_MemoServ, source, "Unknown SET option \2%s\2.", strupper((char *)cmd));
		notice(s_MemoServ, source, "For more information type "
			"\2/%s HELP SET\2", s_MemoServ);
	}

	if (showopt) {
		char buf[512];
		notice(s_NickServ, source, "Options: %s", show_options(buf, showopt));
	}
}

/************************************************************************/

static void do_set_emailmemos(const char *source, NickInfo *ni)
{
	const char *param = gettok(NULL, " ", CMDMAX);
	int reg = sql_select("email FROM nick WHERE name = '%s'", source);

	if (reg)
		row = fetch(res);


	if (!reg || row[0] == NULL) {
		notice(s_MemoServ, source,
			"You have no vaild e-mail set for this nick. "
			"For more information, type: \2/%s HELP SET EMAIL\2",
			s_NickServ);
		return;
	}

	if (!param) {
		if (ni->flags & MS_EMAILMEMOS)
			notice(s_MemoServ, source, "All memos to this nick are being e-mailed to '%s'",
				row[0]);
		else
			notice(s_MemoServ, source, "Memos are not being e-mailed to your nick.");

		return;
	}

	if (sstrcmp(param, "ON") == 0) {
		if (ni->flags & MS_EMAILMEMOS) {
			notice(s_MemoServ, source,
				"Memos are already set to be e-mailed to you.");
			return;
		}
		ni->flags |= MS_EMAILMEMOS;
		notice(s_MemoServ, source, "All memos to this nick will now be e-mailed to '%s'",
		row[0]);

	} else if (sstrcmp(param, "OFF") == 0) {
		if (!(ni->flags & MS_EMAILMEMOS)) {
			notice(s_MemoServ, source,
				"You currently do not have memos being e-mailed.");
			return;
		}
		ni->flags &= ~MS_EMAILMEMOS;
		notice(s_MemoServ, source, "Memos to this nick will no longer be e-mailed to you.");

	} else {
		notice(s_MemoServ, source, "Syntax: \2SET EMAILMEMOS\2 [\37ON\37|\37OFF\37]");
		notice(s_MemoServ, source,
		"\2/%s HELP SET EMAILMEMOS\2 for more information.", s_NickServ);
		return;
        }
	send_sql("UPDATE nick SET flags = %d WHERE id = %d",
		ni->flags, ni->id);
}



/************************************************************************/

static long do_set_nomemo(const char *source, NickInfo *ni)
{
	const char *param = gettok(NULL, " ", CMDMAX);

	if (!param) {
		notice(s_MemoServ, source,
			"Syntax: /%s NOMEMO <option>", s_MemoServ);
		notice(s_MemoServ, source,
			"\2/%s HELP NOMEMO\2 for more information.",
			s_MemoServ);
		return 0;

	} else if (sstrcmp(param, "ON") == 0) {
		ni->flags |= MS_NOMEMO;
		notice(s_MemoServ, source, "You will no longer receieve memos to this nick.");

	} else if (sstrcmp(param, "OFF") == 0) {
		ni->flags &= ~MS_NOMEMO;
		notice(s_MemoServ, source, "Your nick may now receieve memos again.");

	} else {
		notice(s_MemoServ, source, "Syntax: \2SET NOMEMO\2 [\37ON\37|\37OFF\37]");
		notice(s_MemoServ, source,
			"\2/%s HELP SET NOMEMO\2 for more information.", s_MemoServ);
		return 0;
        }
	send_sql("UPDATE nick SET flags = %d WHERE id = %d",
		ni->flags, ni->id);
	return ni->flags;
}


/************************************************************************/

static void do_undel(const char *source)
{

	const char *num = gettok(NULL, " ", 10);

	if (!num) {
		notice(s_MemoServ, source,
                        "Syntax: /%s UNDEL <memo number>", s_MemoServ);
		notice(s_MemoServ, source,
                        "\2/%s HELP UNDEL\2 for more information.",
                        s_MemoServ);

	} else if (!nick_registered(source)) {
		notice(s_MemoServ, source,
			"UNDEL ERROR: Your current nick '%s' is not registered!", source);

	} else if (!is_recognized(source, NULL)) {
		notice(s_MemoServ, source,
			"UNDEL ERROR: You are not identified to your current nick '%s'!", source);


	} else if (sstrcmp(num, "ALL") == 0) {
		int num_memos = sql_select("* FROM memos WHERE recipient = '%s' AND del=1", source);

		if (!num_memos)
			notice(s_MemoServ, source,
				"UNDEL ERROR: You have no memos marked as 'deleted'.");
		else {
			notice(s_MemoServ, source,
				"All your memos marked as 'deleted' have been restored. (%d memo%s)",
				num_memos, num_memos==1 ? "" : "s");
			send_sql("UPDATE memos SET del=0 WHERE recipient = '%s'", source);
		}
	} else if (!strspn(num, "1234567890") || (atoi(num) < 1)) {
		notice(s_MemoServ, source,
			"UNDEL ERROR: Invalid memo number.");
	} else {
		int num_memos=sql_select("* FROM memos WHERE recipient = '%s' ORDER BY time", source);
		
		if (!num_memos)
			notice(s_MemoServ, source,
				"UNDEL ERROR: You have no memos to delete.");

		else if (atoi(num) > num_memos)
			notice(s_MemoServ, source,
				"UNDEL ERROR: Memo number #%s doesn't exist.", num);
		else {
			int i;
			for (i=1; i <= num_memos; i++) {
				row = fetch(res);
				if (i == atoi(num)) {
					if (!atoi(row[MEMO_DEL])) {
						notice(s_MemoServ, source,
							"Memo #%s is not marked as 'deleted'.", num);
						return;
					}


					notice(s_MemoServ, source,
						"Memo #%s no longer marked as 'deleted'.", num);
					send_sql("UPDATE memos SET del=0 WHERE recipient='%s' AND "
						"sender='%s' AND time=%d", source,
						row[MEMO_SENDER], atol(row[MEMO_TIME]));
					return;
				}
			}
			notice(s_MemoServ, source,
				"UNDEL ERROR: Memo number #%s not found.", num);
		}
	}

}

/************************************************************************/

void expire_memos()
{
	time_t expire = time(NULL) - (MEMO_EXPIRE*60*60*24);
	unsigned long memos = sql_select("* FROM memos");

	send_sql("DELETE FROM memos WHERE time <= %d", expire);

	if (DISPLAY_UPDATES)
		globops(NULL, "Completed memo expire: (%ld/%ld)",
			(long)(mysql_affected_rows(&mysql)), memos);
}

/************************************************************************/
/************************************************************************/

