/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"
#include "log.h"
#include "timeout.h"

/****************************************************************************/
/****************************************************************************/

#define FLAG_HOLD	1
#define FLAG_MARK	2
#define FLAG_FORBID	3
#define FLAG_FREEZE	4

static void do_register(const char *source);
static void do_auth(const char *source);
static void do_info(const char *source);
static void do_identify(const char *source);
static void do_set(const char *source);
static void do_set_password(const char *source, const char *param, unsigned long id);
static void do_set_email(const char *source, const char *param, unsigned long id);
static void do_set_showemail(const char *source, const char *param, unsigned long id);
static long do_set_enforce(const char *source, const char *param, unsigned long id);
static long do_set_secure(const char *source, const char *param, unsigned long id);
static void do_set_url(const char *source, const char *param, unsigned long id);
static void do_set_icq(const char *source, const char *param, unsigned long id);
static long do_set_noop(const char *source, const char *param, unsigned long id);
static long do_set_neverop(const char *source, const char *param, unsigned long id);
static long do_set_private(const char *source, const char *param, unsigned long id);
static void do_set_mailblock(const char *source, const char *param, unsigned long id);
static void do_access(const char *source);
static void do_drop(const char *source);
static void do_acc(const char *source);
static void do_ghost(const char *source);
static void do_sendpass(const char *source);
static void do_getpass(const char *source);
static void do_recover(const char *source);
static void do_release(const char *source);
static void help(const char *source);
static void ohelp(const char *source);
static void myquestion(const char *source);
static void do_list(const char *source);
static void do_flags(const char *source, int flags);
static void do_hold(const char *source) { do_flags(source, FLAG_HOLD); }
static void do_mark(const char *source) { do_flags(source, FLAG_MARK); }
static void do_forbid(const char *source) { do_flags(source, FLAG_FORBID); }
static void do_freeze(const char *source) { do_flags(source, FLAG_FREEZE); }
static void do_delete(const char *source);
static void add_forbid(const char *source, User *u, const char *nick, const char *reason);
static void do_flist(const char *source);
static void do_flush(const char *source);
static void do_getemail(const char *source);
static void do_listchans(const char *source);
static void do_authinfo(const char *source);
static void do_listemail(const char *source);


/****************************************************************************/


typedef struct nicknode NickNode;
struct nicknode {
	NickNode *left, *right;
	NickInfo *ni;
	unsigned long value;
};


NickNode *nicknodes[256];


typedef struct {
	const char *cmd;
	void (*func)(const char *source);
	unsigned int access;
	unsigned int readonly;
	const char *desc;
} Command;


static Command commands[] = {
	{ "ACC", 	do_acc,		BASIC,	0,	NS_OHELP_ACC },
	{ "ACCESS", 	do_access,	BASIC,	0,	NS_OHELP_ACCESS },
	{ "AUTH", 	do_auth,	BASIC,	1,	NS_OHELP_AUTH },
	{ "AUTHINFO", 	do_authinfo,	SA,	1,	NS_OHELP_AUTHINFO },
	{ "DROP", 	do_drop,	BASIC,	1, 	NS_OHELP_DROP },
	{ "DELETE",	do_delete,	SRA,	1,	NS_OHELP_DELETE },
	{ "FLIST", 	do_flist,	OPER,	1,	NS_OHELP_FLIST },
	{ "FLUSH", 	do_flush,	SRA,	1,	NS_OHELP_FLUSH },
	{ "FORBID", 	do_forbid,	SRA,	1,	NS_OHELP_FORBID },
	{ "FREEZE", 	do_freeze,	SOP,	1,	NS_OHELP_FREEZE },
	{ "GETEMAIL",	do_getemail,	SA,	1,	NS_OHELP_GETEMAIL },
	{ "GETPASS", 	do_getpass,	SA,	0,	NS_OHELP_GETPASS },
	{ "GHOST", 	do_ghost,	BASIC,	0,	NS_OHELP_GHOST },
	{ "HELP", 	help,		BASIC,	0,	NS_OHELP_HELP },
	{ "HOLD",	do_hold,	SA,	0,	NS_OHELP_HOLD },
	{ "IDENTIFY", 	do_identify,	BASIC,	0,	NS_OHELP_IDENTIFY },
	{ "ID", 	do_identify,	BASIC,	0,	NS_OHELP_ID },
	{ "INFO", 	do_info,	BASIC,	0,	NS_OHELP_INFO },
	{ "LIST", 	do_list,	BASIC,	0,	NS_OHELP_LIST },
	{ "LISTCHANS", 	do_listchans,	BASIC,	0,	NS_OHELP_LISTCHANS },
	{ "LISTEMAIL", 	do_listemail,	SA,	0,	NS_OHELP_LISTEMAIL },
	{ "MARK",	do_mark,	SA,	0,	NS_OHELP_MARK },
	{ "MYQUESTION",	myquestion,	BASIC,	0,	NS_OHELP_MYQUESTION },
	{ "OHELP",	ohelp,		OPER,	0,	NS_OHELP_OHELP },
	{ "RECOVER", 	do_recover,	BASIC,	0,	NS_OHELP_RECOVER },
	{ "REGISTER", 	do_register,	BASIC,	1,	NS_OHELP_REGISTER },
	{ "RELEASE", 	do_release,	BASIC,	0,	NS_OHELP_RELEASE },
	{ "SENDPASS", 	do_sendpass,	BASIC, 	0,	NS_OHELP_SENDPASS },
	{ "SET", 	do_set,		BASIC,	1,	NS_OHELP_SET },
	{ "SIDENTIFY",	do_identify,	BASIC,	0,	NS_OHELP_SIDENTIFY },

	{ NULL }
};


int check_ns_command(const char *cmd) {
        Command *c;
        int i=0;

        for (c = commands; c->cmd; c++) {
                if (sstrcmp(cmd, c->cmd) == 0) {
                        return i;
                }
                i++;
        }
        return -1;
}

inline void update_ns_table(int cmd, int newlevel) {
	commands[cmd].access = newlevel;
}


/****************************************************************************/

#define BLOCK_AUTH_EMAIL    1
#define BLOCK_GENERAL       2
#define BLOCK_SENDPASS      3

#define HELD_ID    1073741824


/****************************************************************************/

static inline unsigned long nick_key(const char *nick)
{ 
	NickInfo *ni;

	if (!nick || !(ni = findnick(nick)))
		return 0;

	return ni->id;
}

/****************************************************************************/

static void succeed_chans(const char *nick)
{
	NickInfo *ni = findnick(nick);
	int num_rows=0, i=0;


	if (!ni)
		return;


	num_rows = sql_select("name,id FROM chan WHERE fid = %d", ni->id);

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		if (!succeed_chan(row[0])) {
			send_sql("DELETE FROM chan WHERE id = %d", atol(row[1]));
			send_sql("DELETE FROM cflags WHERE name = '%s'", row[0]);
			send_sql("DELETE FROM reason WHERE name = '%s'", row[0]);
			send_sql("DELETE FROM caccess WHERE chan = %d", atol(row[1]));
			del_ci_chan(row[0]);
		}
	}
}


/****************************************************************************/

inline int nick_registered(const char *nick) { return findnick(nick) ? 1 : 0; }

/****************************************************************************/

/*
 *  If *nick is NULL, then we're only checking for recognition of active user.
 *  If *nick is valid, then we're checking if we're recognized on another
 *  nick's access list. First checks to see if we're identified to it, if not, then
 *  scans the target *nick's access list for a match with source *user's hostmask.
 *  Always returns false if nick security is set (unless identified is true).
 *
 */

int is_recognized(const char *user, const char *nick)
{
	NickInfo *ni;
	struct NickAccess *na;
	User *u = finduser(user);
	char *buf;


	if (!u)
		return 0;

	if (nick && is_identified(user, nick))
		return 1;


	if (!nick && is_identified(user, u->nick))
		return 1;


	if (!nick)
		return (u && (u->mode & U_RECOGNIZED));


	if (!(ni = findnick(nick)))
		return 0;


	/* Check for nick security, if enabled, return false */

	if (ni->flags & NS_SECURE)
		return 0;


        buf = smalloc(strlen(u->username) + strlen(u->host) + 2);
        sprintf(buf, "%s@%s", u->username, u->host);


	for (na = ni->entries; na; na = na->next) {
		if (match_wild_nocase(na->mask, buf)) {
			free(buf);
			return 1;
		}
	}

	free(buf);
	return 0;
}

/****************************************************************************/

int is_identified(const char *user, const char *nick)
{
	struct u_idlist *nl;
	User *u = finduser(user);

	if (!u)
		return 0;


	for (nl = u->nicks; nl; nl = nl->next) {
		if (sstrcmp(nl->nick, nick) == 0)
			return 1;
	}
	return 0;
}

/****************************************************************************/

int is_authed(const char *nick)
{
	NickInfo *ni = findnick(nick);
	int authed=0;
	char buf[512];

	if (!ni)
		return 0;

	snprintf(buf, sizeof(buf),
		"SELECT auth FROM nick WHERE id = %ld", ni->id);

	if (!mysql_query(&mysql, buf)) {
		if ((res2 = mysql_store_result(&mysql)) != NULL) {
			if (mysql_num_rows(res2)) {
				row2 = fetch(res2);
				if (atol(row2[0]) == 1)
					authed=1;
			}
		}
	}

	if (authed) {
		mysql_free_result(res2);
		return 1;
	}
	return 0;

}

/****************************************************************************/

/*

static void del_id_nick(User *u, const char *nick)
{
	struct u_idlist *nl;

	blog("NS_O", "[NS] D %s (%s!%s@%s)",
		nick, u->nick, u->username, u->host);

	if (u->os_nick && sstrcmp(nick, u->os_nick) == 0)
		u->os_level = 0;

	for (nl = u->nicks; nl; nl = nl->next) {
		if (sstrcmp(nl->nick, nick) == 0) {
			if (nl->prev)
				nl->prev->next = nl->next;
			else
				u->nicks = nl->next;
			if (nl->next)
				nl->next->prev = nl->prev;
			free(nl->nick);	free(nl);
			return;
		}
	}
}

*/


/****************************************************************************/

static void scan_id_list(User *u, const char *nick)
{
	struct u_idlist *nl;

	for (nl = u->nicks; nl; nl = nl->next) {
		if (sstrcmp(nl->nick, nick) == 0) {
			if (nl->prev)
				nl->prev->next = nl->next;
			else
				u->nicks = nl->next;
			if (nl->next)
				nl->next->prev = nl->prev;
			free(nl->nick); free(nl);
			return;
		}
	}
}

/****************************************************************************/

/* For instances like DROP or DELETE, this scans thru all id lists to make sure
 * no one is left identified to a non-existent nick */

static void del_all_id_nicks(const char *nick)
{
	User *u;
	int ulist=0;

	for (ulist=0; ulist < 256; ulist++) {
		for (u = userlist[ulist]; u; u = u->next)
			scan_id_list(u, nick);
	}
}



/****************************************************************************/

static void add_id_nick(User *u, const char *nick, int reg)
{
	NickInfo *ni = findnick(nick);
	struct u_idlist *nl;

	for (nl = u->nicks; nl; nl = nl->next) {
		if (sstrcmp(nl->nick, nick) == 0)
			return;
	}

	if (ni && ni->remind)
		ni->remind=0;

	if (!reg && LOG_IDENTIFIES)
		blog("NS_I", "[NS] I %s (%s!%s@%s)",
			nick, u->nick, u->username, u->host);


	nl = scalloc(sizeof(struct u_idlist), 1);
	nl->next = u->nicks;
	nl->prev = NULL;
	if (u->nicks)
		u->nicks->prev = nl;
	u->nicks = nl;
	nl->nick = sstrdup(nick);

}

void add_idnick(User *u, const char *nick)
{
	add_id_nick(u,nick, 0);
}


/****************************************************************************/

long generateid(User *u, const char *target, unsigned long id)
{
	int i=0, num_rows=0;
	long serviceid=0;

	num_rows = sql_select("stamp FROM nick");

	if (!num_rows && !u->timestamp)
		serviceid = getrandom(10000000,99999999);
	else if (num_rows && !u->timestamp) {   /* make sure services stamp is unique */
		int dupl=0;

		while (dupl != -1) {
			dupl=0;
			serviceid = getrandom(10000000,99999999);
			mysql_data_seek(res, 0);
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				if (atol(row[0]) == serviceid) {
					dupl=1;
					break;
				}
			}
			if (!dupl)
				dupl=-1;
		}
	}



	send_sql("UPDATE nick SET stamp = %d WHERE id = %d",
		u->timestamp>0 ? u->timestamp : serviceid, id);
	if (!u->timestamp)
		u->timestamp = serviceid;


	/* If user is identifying to active & authorized nick, set +r. Otherwise only set +d */

	if (ENABLE_R_MODES && (ENABLE_R_MODES != 2)) {
		if (AUTH_EMAIL)
			send_cmd(NULL, "SVSMODE %s +%sd %d", u->nick, 
			(is_authed(u->nick) && sstrcmp(target,u->nick)==0 ? "r" : ""), u->timestamp);
		else
			send_cmd(NULL, "SVSMODE %s +%sd %d", u->nick, 
			(sstrcmp(target,u->nick)==0 ? "r" : ""), u->timestamp);
	}

	return u->timestamp;
}

/****************************************************************************/

/* Turn off nick forwarding for any other nick registration that points to this nick. */

static void remove_forwards(const char *nick)
{
	NickInfo *ni;
	int num_rows=0, i;

	num_rows = sql_select("name FROM nick WHERE forward = '%s'", nick);

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		ni = findnick(row[0]);
		if (ni && ni->forward == ni)
			ni->forward = NULL;
	}

	send_sql("UPDATE nick SET forward='0' WHERE forward = '%s'", nick);

}

/****************************************************************************/

static void del_sql_nick(const char *nick)
{
	if (sql_select("id FROM nick WHERE name = '%s'", nick)) {
		row = fetch(res);
		send_sql("DELETE FROM naccess WHERE id = '%s'", row[0]);
		send_sql("DELETE FROM pending WHERE id = '%s'", row[0]);
	}
	send_sql("DELETE FROM nick WHERE name = '%s'", nick);
	send_sql("DELETE FROM flags WHERE name = '%s'", nick);
	send_sql("DELETE FROM reason WHERE name = '%s'", nick);
	send_sql("DELETE FROM oper WHERE name = '%s'", nick);
	send_sql("DELETE FROM memos WHERE recipient = '%s'", nick);
	send_sql("DELETE FROM caccess WHERE entry = '%s'", nick);
	send_sql("UPDATE chan SET successor=NULL WHERE successor = '%s'", nick);
}

/****************************************************************************/

inline long nick_flags(const char *nick)
{
	NickInfo *ni = findnick(nick);

	if (ni)
		return ni->flags;

	return 0;

}

/****************************************************************************/

void update_nick_flags(const char *nick, long flags)
{
	NickInfo *ni = findnick(nick);

	if (ni)
		ni->flags = flags;
}

/****************************************************************************/

static void del_ni_accesslist(NickInfo *ni)
{
	struct NickAccess *na, *na2;

	if (!ni)
		return;

	na = ni->entries;

	while (na) {
		na2 = na->next;
		free(na->mask); free(na);
		na = na2;
	}
	ni->entries = NULL;
}

/****************************************************************************/

static void del_ni_access(const char *source, const char *mask)
{
	NickInfo *ni = findnick(source);
	struct NickAccess *na, *na2;

	if (!ni)
		return;

	na = ni->entries;

	while (na) {
		na2 = na->next;
		if (sstrcmp(na->mask, mask) == 0) {
			if (na->prev)
				na->prev->next = na->next;
			else
				ni->entries = na->next;
			if (na->next)
				na->next->prev = na->prev;
			free(na->mask); free(na);
		}
		na = na2;
	}
}


/****************************************************************************/

static void add_ni_access(NickInfo *ni, const char *mask)
{
	struct NickAccess *na = scalloc(sizeof(struct NickAccess), 1);

	na->mask = sstrdup(mask);
	na->next = ni->entries;
	if (ni->entries)
		ni->entries->prev = na;
	ni->entries = na;
}

/****************************************************************************/

static void add_ni_nick(NickInfo *ni)
{
	NickNode *nn, *newnode = scalloc(sizeof(NickNode), 1);
	int alpha = toupper(*ni->nick) & 0x000000FF;
	unsigned long hvalue=0;


	nick_hash((char *)ni->nick, &hvalue);

	newnode->ni = ni;
	newnode->value = hvalue;

	if (!(nn = nicknodes[alpha])) {
		nicknodes[alpha] = scalloc(sizeof(NickNode), 1);
		nicknodes[alpha]->value = 50000000;
		nicknodes[alpha]->ni = scalloc(sizeof(NickInfo), 1);
		nicknodes[alpha]->ni->nick = sstrdup("*");
		if (hvalue < 50000000)
			nicknodes[alpha]->left = newnode;
		else
			nicknodes[alpha]->right = newnode;
		return;
	}

	while (nn != NULL) {
		if (hvalue < nn->value) {
			if (nn->left)
				nn = nn->left;
			else {
				nn->left = newnode;
				break;
			}
		} else if (hvalue >= nn->value) {
			if (nn->right)
				nn = nn->right;
			else {
				nn->right = newnode;
				break;
			}
		}
	}


}


/****************************************************************************/

static void retree_nicknode(NickNode *nn, NickNode *nptr)
{
	while (nptr != NULL) {
		if (nn->value < nptr->value) {
			if (nptr->left)
				nptr = nptr->left;
			else {
				nptr->left = nn;
				break;
			}
		} else if (nn->value >= nptr->value) {
			if (nptr->right)
				nptr = nptr->right;
			else {
				nptr->right = nn;
				break;
			}
		}
	}


}

/* If succeed is true, check to to see if there is a successor, otherwise let the expiry routine
 * catch it. The succeed int should be true for all calls to this function for the exception of
 * expire_nicks(). We don't want the nick expiry to succeed chans - let the expire_chans() function
 * take care of that. */


static void del_ni_nick(const char *nick, int succeed)
{
	NickNode *nn, *last;
	int alpha = toupper(*nick) & 0x000000FF;
	unsigned long hvalue=0;

	if (succeed)
		succeed_chans(nick);
	expire_chanlists(nick);
	del_all_id_nicks(nick);
	remove_forwards(nick);

	nick_hash((char *)nick, &hvalue);

	last = nn = nicknodes[alpha];


	while (nn != NULL) {
		if ((hvalue == nn->value) && sstrcmp(nick, nn->ni->nick) == 0) {
			freeze(nn->ni->nick, DEL);
			del_ni_accesslist(nn->ni);
			free(nn->ni->nick);
			free(nn->ni);

			if (last->right == nn)
				last->right = nn->right;
			else if (last->left == nn)
				last->left = nn->right;
			if (nn->left)
				retree_nicknode(nn->left, last);
			free(nn);

			break;
		}
		last = nn;
		nn = (hvalue<nn->value) ? nn->left : nn->right;
	}

}

/****************************************************************************/

void load_nickinfo()
{
	NickInfo *ni, *nif;
	unsigned long num_rows = sql_select("id,name,flags,auth,news,pass FROM nick");
	unsigned i;

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		ni = scalloc(sizeof(NickInfo), 1);
		ni->id = strtoul(row[0], NULL, 10);
		ni->nick = sstrdup(row[1]);
		ni->flags = atol(row[2]);
		ni->news = atoi(row[4]);
		add_ni_nick(ni);
#ifdef USE_ENCRYPTION
		if (!(ni->flags & NS_ENCRYPTED) && ni->flags != -1) {
			char passwd[PASSMAX];

			if (!encryptm(row[5], strlen(row[5]), passwd, PASSMAX, 1)) {
				ni->flags |= NS_ENCRYPTED;
				send_sql("UPDATE nick SET pass = '%s', flags = %d WHERE id = %d",
					passwd, ni->flags, ni->id);
			} else
				fatal_error("Could not encrypt password for nick '%s'", ni->nick);

		}
#else  /* NOT ENCRYPTED */
		if ((ni->flags > 0) && ni->flags & NS_ENCRYPTED)
			fatal_error("Nick '%s' has encrypted password and services are not in "
			"encryption mode!", ni->nick);
#endif
	}
	num_rows = sql_select("nick.name,naccess.mask FROM nick,naccess WHERE nick.id = naccess.id");
	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		ni = findnick(row[0]);
		if (ni)
			add_ni_access(ni, row[1]);
	}

	num_rows = sql_select("name,forward,id FROM nick WHERE forward <> '0'");
	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		if (row[1] && (nif=findnick(row[1]))) {
			if ((ni = findnick(row[0])))
				ni->forward = nif;
		} else
			send_sql("UPDATE nick SET forward=NULL WHERE id=%d", atol(row[2]));
	}
}

/****************************************************************************/

NickInfo *findnick(const char *nick)
{
	NickNode *nn;
	unsigned long hvalue=0;
	unsigned int alpha = toupper(*nick) & 0x000000FF;

	if ((alpha > 256) || (alpha < 0) || !(nn = nicknodes[alpha]))
		return NULL;

	nick_hash((char *)nick, &hvalue);


	while (nn != NULL) {
		if ((hvalue == nn->value) && sstrcmp(nick, nn->ni->nick) == 0)
			return nn->ni;
		nn = (hvalue<nn->value) ? nn->left : nn->right;
	}
	return NULL;
}

/****************************************************************************/

void nickserv(const char *source, char *buf)
{
	Command *c;
	const char *cmd = gettok(buf, " ", CMDMAX);
	int valid=0;

	if (!cmd)
		return;


#ifdef USE_SQL
	if (!sql_alive) {
		notice(s_NickServ, source, "Sorry, we're currently experiencing temporary "
		"technical difficulties. Please try again shortly.");
		return;
	}
#endif
	if (is_frozen(source) && !is_oper(source)) {
		notice(s_NickServ, source,
			"This nick is frozen and may not be used.");
		return;
	}


	for (c = commands; c->cmd; c++) {
		if (sstrcmp(c->cmd, cmd) == 0) {
			valid=1;
			if (READ_ONLY && c->readonly)
				notice(s_NickServ, source,
					"Sorry, that command is temporarly disabled while services are in "
					"read-only mode.");
			else if (!c->access || has_access(source, c->access))
				c->func(source);
			else
				notice(s_NickServ, source, "Permission denied.");
			break;
		}
	}

	if (!valid)
		notice(s_NickServ, source,
			"Unrecognized command \2%s\2. Type \"/%s HELP\" for help.",
				cmd, s_NickServ);


}


/****************************************************************************/

/*  create_ticket() will free the last result set which *email and *info points to.
 *  Creating pointers *pemail and *pinfo to newly allocated memory before calling create_ticket()
 *  gets around this. */

static void send_auth_email(const char *source, const char *email, const char *info, int type, unsigned long id)
{
	char emsg[1024];
	char esubj[128];
	char *pemail = sstrdup(email), *pinfo = sstrdup(info);
	char *code = create_ticket();

	send_sql("INSERT INTO pending VALUES "
		"(%d, '%s','%s','%d', %d)",
		id, pinfo, code, time(NULL), type);


	switch (type) {
		case 0:
			if (HIGH_NICK_SECURITY)
				snprintf(emsg, sizeof(emsg), "Your authorization code is %s.\n"
					"To fully authorize your nick, log back "
					"onto %s using the nick '%s' and type - /%s AUTH %s # <answer>.\n"
					"For example: /%s AUTH %s 2 Maple Street",
					code, NETWORK_NAME, source, s_NickServ, code, s_NickServ, code);

			else
				snprintf(emsg, sizeof(emsg),
					"Your nickname's e-mail authorization code is %s. "
					"Log back onto %s using the nick %s and type - "
					"/%s AUTH %s\n",
					code, NETWORK_NAME, source, s_NickServ, code);
			break;

		case 1:
			snprintf(emsg, sizeof(emsg),
				"Someone, possibly you, has requested an e-mail change"
				"for the nick '%s'. If you approve of this, log back onto "
				"%s and type: \"/%s AUTH %s\". If you did not request "
				"this password change, delete this e-mail and follow no "
				"further instructions - do NOT issue the AUTH command or else "
				"your password will be changed!",
				source, NETWORK_NAME, s_NickServ, code);
			break;
		case 2:
			snprintf(emsg, sizeof(emsg),
				"Someone, possibly you, has requested a password change"
				"for the nick '%s'. If you approve of this, log back onto "
				"%s and type: \"/%s AUTH %s\". If you did not request "
				"this password change, do not issue the AUTH command or else "
				"your password will be changed!",
				source, NETWORK_NAME, s_NickServ, code);
			break;
		case 3:
			snprintf(emsg, sizeof(emsg),
				"Your nickname's drop authorization code is %s. "
				"If you wish to confirm the complete deletion of your "
				"nick \"%s\" then log back onto %s using the nick %s and type - "
				"\"/%s AUTH %s\".\nOtherwise, delete this e-mail and "
				"follow no further instructions.",
				code, source, NETWORK_NAME, source, s_NickServ, code);
			break;
		case 4:
			snprintf(emsg, sizeof(emsg),
				"Your nickname's MAILBLCOK authorization code is %s. "
				"If you wish to confirm this setting change of your nick "
				"\"%s\" then log back onto %s using the nick %s and type - "
				"\"/%s AUTH %s\".\nOtherwise, delete this e-mail and "
				"follow no further instructions. If you authorize this "
				"change, you are confirming you no longer wish to have "
				"services e-mail your password to you. Make sure you either "
				"remember your password or have it written down somewhere.",
				code, source, NETWORK_NAME, source, s_NickServ, code);

			break;
		default:
			if (debug)
				log("Unknown send auth type %d for nick %s", type, source);
			return;
			break;


	}
	snprintf(esubj, sizeof(esubj), "%s nick authorization.", NETWORK_NAME);
	e_mail(emsg, esubj, pemail, type);
	free(code); free(pemail); free(pinfo);

}

/****************************************************************************/

static inline int check_maxnicks(const char *email)
{
	return sql_select("name FROM nick WHERE email = '%s'", email);
}


/****************************************************************************/


static void do_register(const char *source)
{
	User *u;
	const char *pass = strtok(NULL, " ");
	const char *email = strtok(NULL, " ");

	if (!pass || (!email && AUTH_EMAIL)) {
		notice(s_NickServ, source, "Syntax: \2REGISTER\2 <password> %s",
			AUTH_EMAIL==1 ? "<e-mail>" : "");
		notice(s_NickServ, source,
			"\2/%s HELP REGISTER\2 for more information.",
				s_NickServ);

	} else if (!(u = finduser(source))) {
		if (debug)
			log("%s registration failed! %s not online!",
				s_NickServ, source);
		notice(s_NickServ, source, "Sorry, registration failed.");



	} else if (sstrcmp(source, pass)==0 || strlen(pass) < 5) {
		notice(s_NickServ, source,
			"REGISTER failure: Please try again with a more obscure password. "
			"Passwords should be at least five characters in length, should not "
			"be something easily guessed (e.g. your real name or nick), and cannot "
			"contain the space or tab characters. "
			"\2/%s HELP REGISTER\2 for more information.", s_NickServ);

	} else if (strlen(pass) > PASSMAX) {
		notice(s_NickServ, source,
			"Sorry, nick passwords may only be up to %d characters in length.",
				PASSMAX);


	} else if (check_escape_codes(pass)) {
		notice(s_NickServ, source,
			"REGISTER ERROR: Your password has invalid characters. Please try again with "
			"common characters and don't use special colour/attribute codes.");


	} else if (time(NULL) < (u->lastreg + NS_REG_DELAY)) {
		notice(s_NickServ, source, NS_REGISTER_DELAY);

	} else if (AUTH_EMAIL && email && !check_valid_email(email)) {
		notice(s_NickServ, source, INVALID_EMAIL);

	} else if (AUTH_EMAIL && email && check_bad_email(email)) {
		notice(s_NickServ, source, BAD_EMAIL);

                 /* keep > rather than >= since we just incremented it with check_email */
	} else if (email && (check_email(email, BLOCK_GENERAL, 1) > MAX_EMAILS)) {
		notice(s_NickServ, source, MAX_EMAIL_REACHED);

	} else if (email && MAX_NICKS && (check_maxnicks(email) >= MAX_NICKS)) {
		notice(s_NickServ, source, MAX_NICKS_REACHED, email);

	} else {
		int reg=0;
		char temp[NICKMAX];

		strncpy(temp, source, NICKMAX);
		strupper(temp);

		if ((strstr(temp, "GUEST") == temp)) {
			notice(s_NickServ, source,
			"You may not register nicks starting with \"Guest\". "
			"\2/%s HELP REGISTER\2 for more information.", s_NickServ);


		} else if ((reg = nick_registered(source)) && (nick_flags(source) < 0))
			notice(s_NickServ, source, 
				"Nick '\2%s\2' may not be registered.", source);


		else if (reg)
			notice(s_NickServ,source, "Nickname \2%s\2 already registered!",
				source);
		else if (reg<0)  /* no result set from database */
			notice(s_NickServ, source, "Sorry, nickname registration failed.");

		else {     /* create nick */
			NickInfo *ni = scalloc(sizeof(NickInfo), 1);
			long stamp, flags=0;
			char *mask = create_mask(u);
#ifndef USE_ENCRYPTION
			const char *passwd = pass;

#else
			char passwd[PASSMAX];
			encryptm(pass, strlen(pass), passwd, PASSMAX, 1);
			flags |= NS_ENCRYPTED;
#endif

			flags |= NS_HIDE_EMAIL;
			flags |= MS_MEMO_LOGON;
			flags |= MS_MEMO_NEW;

			if (DEF_NS_ENFORCE)
				flags |= NS_ENFORCE;
			if (DEF_NS_SECURE)
				flags |= NS_SECURE;
			if (DEF_NS_HIDE)
				flags |= NS_HIDE_EMAIL;


			ni->id = generate_primary_id("nick");
			ni->nick = sstrdup(source);
			ni->flags = flags;
			add_ni_nick(ni);
			stamp = generateid(u, source, ni->id);


			u->lastreg = time(NULL);
			add_id_nick(u, source, 1);


			if (AUTH_EMAIL)
				send_sql("INSERT INTO nick VALUES "
				"(%d,'%s','%s','%s',NULL,NULL,'%s@%s','%d','%d','%d','%d',NULL,NULL,%d,0,0)",
					ni->id, source, passwd, email,
					u->username, u->host,
					time(NULL), time(NULL), flags, stamp, AUTH_EMAIL==1 ? 0 : -1);
			else
				send_sql("INSERT INTO nick VALUES "
				"(%d,'%s','%s',NULL,NULL,NULL,'%s@%s','%d','%d','%d','%d',NULL,NULL,%d,0,0)",
					ni->id, source, passwd,
					u->username, u->host,
					time(NULL), time(NULL), flags, stamp, AUTH_EMAIL==1 ? 0 : -1);

			send_sql("INSERT INTO naccess VALUES "
				"(%d, '%s')",
				ni->id, mask);
			add_ni_access(ni, mask);
			free(mask);

			if (AUTH_EMAIL) {
				add_email(email, BLOCK_GENERAL);
				send_auth_email(source, email, email,0, ni->id);
			}
			blog("NS_R", "[NS] R %s (%s@%s)",
				source, u->username, u->host);
			notice(s_NickServ, source, NS_REGISTER, source);
			notice(s_NickServ, source, NS_REGISTER2, pass);
			if (AUTH_EMAIL)
				notice(s_NickServ, source, NS_REGISTER3);
			if (NSECURE)
				notice(s_NickServ, source, NS_REGISTER4);

		}
	}




}

/****************************************************************************/

static void do_auth(const char *source)
{
	User *u = finduser(source);
	const char *code = strtok(NULL, " ");

	if (!u)
		return;

	if (!code) {
		notice(s_NickServ, source,
			"AUTH ERROR: Syntax: /%s AUTH <code>", s_NickServ);
		notice(s_NickServ, source,
			"\2/%s HELP AUTH\2 for more information.",
				s_NickServ);

	} else if (!is_identified(source, source)) {
		notice(s_NickServ, source,
			"Password authentication is required before you may give your AUTH code. "
			"Retry after typing \2/%s IDENTIFY <your-password>\2", s_NickServ);


	} else {
		int pend=0;
		NickInfo *ni = findnick(source);

		if (!ni) {
			notice(s_NickServ, source,
				"AUTH ERROR: Nick '%s' is not registered!", source);

		} else if (!(pend = sql_select("authcode,info,type FROM pending WHERE id = %d", ni->id))
			&& is_authed(source)) {
			notice(s_NickServ, source,
				"AUTH ERROR: Your nick is fully secured. You have no pending "
				"authorization requests.");

		} else if (!pend) {
			notice(s_NickServ, source,
				"You have not been issued an e-mail authorization "
				"code. Type \2/%s SET EMAIL <your@email.com>\2 "
				"to do so.", s_NickServ);


		} else {
			char *question, *answer;
			int type=0;
			row = fetch(res);

			type = atoi(row[2]);
			if (HIGH_NICK_SECURITY && type < 1) {
				question = strtok(NULL, " ");
				answer = strtok(NULL, "");

				if (!question || !answer) {
					notice(s_NickServ, source,
						"Invalid syntax for setting authorization and "
						"personal question.");
					notice(s_NickServ, source,
						"AUTH syntax: AUTH <code> <question#> <answer>");
					return;
				}
				if (!strspn(question, "1234567")) {
					notice(s_NickServ, source,
						"Invalid Question number! Only 1-7 valid!");
					return;
				}
			}

			if (sstrcmp(row[0], code) == 0) {
				switch(type) {
					case 0:       /* new authorization */
						if (HIGH_NICK_SECURITY) {
							send_sql("UPDATE nick SET auth = 1, question = '%s', "
								"answer = '%s' WHERE id = %d", 
								question, answer, ni->id);
						} else
							send_sql("UPDATE nick SET auth = 1 WHERE id = %d", 
								ni->id);
							generateid(u, source, ni->id);

						break;
					case 1:
						send_sql("UPDATE nick SET email = '%s' WHERE id = %d", 
							row[1], ni->id);
						break;
					case 2:
						send_sql("UPDATE nick SET pass = '%s' WHERE id = %d", 
							row[1], ni->id);
						break;
					case 3:
						del_ni_nick(source, 1);
						del_sql_nick(source);
						break;
					case 4:
						if (ni->flags & NS_MAILBLOCK)
							ni->flags &= ~NS_MAILBLOCK;
						else
							ni->flags |= NS_MAILBLOCK;
						send_sql("UPDATE nick SET flags = %d WHERE id = %d",
							ni->flags, ni->id);
						break;

					default:
						if (debug)
							log("Unknown auth type %d for nick %s",
								type, source);
				}


				send_sql("DELETE FROM pending WHERE id = %d", ni->id);

				notice(s_NickServ, source, "AUTH code accepted, %s%s%s%s",
					type <2 ?
					"your e-mail address is now authorized." : "",
					type==2 ?
					"your new password is now set. Remember to write it down!" : "",
					type==3 ?
					"your nick is now deleted from our records." : "",
					type==4 ?
					"your mailblock setting has been changed." : "");

				blog("NS_R", "[NS] A %s (%s@%s) [%s]",
					source, u->username, u->host,
					type==2 ? "pass" : type<2 ? "e-mail" : 
					type==3 ? "Drop" : type==4 ? "mailblock" : "");


			} else {
				notice(s_NickServ, source,
					"AUTH ERROR: Invalid authorization code.");
				blog("NS_R", "[NS] A* %s (%s@%s)",
					source, u->username, u->host);



			}

		}
	}

}

/****************************************************************************/

char *show_options(char *buf, long flags)
{

	*buf=0;
	if (flags & NS_ENFORCE)
		strcat(buf, "Enforced");
	if (flags & NS_SECURE) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Security");
	}
	if (flags & MS_MEMO_RCPT) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Memo Receipts");
	}
	if (flags & NS_NOOP) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "No Op");
	}
	if (flags & NS_NEVEROP) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Never Op");
	}
	if (flags & MS_NOMEMO) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "No Memo");
	}
	if (flags & NS_PRIVATE) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Privacy");
	}
	if (flags & NS_MAILBLOCK) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Mail-block");
	}
	if (!*buf)
		strcpy(buf, "None");

	return buf;

}

/****************************************************************************/
#define INFO_NAME     0
#define INFO_EMAIL    1
#define INFO_MASK     2
#define INFO_TIME     3
#define INFO_LAST     4
#define INFO_FLAGS    5
#define INFO_ICQ      6
#define INFO_URL      7
#define INFO_CODE     8
#define INFO_FORWARD  9


static void do_info(const char *source)
{

	const char *nick = gettok(NULL, " ", NICKMAX);
	unsigned long id = nick_key(nick);

	if (!nick) {
		notice(s_NickServ, source, "Syntax: \2INFO\2 <nick>");
		notice(s_NickServ, source,
			"\2/%s HELP INFO\2 for more information.",
				s_NickServ);
	} else {

		int num_rows=sql_select(
			"name,email,usermask,time,lasttime,flags,icq,url,auth,forward "
			"FROM nick WHERE id=%d", id);

		if (!num_rows)
			notice(s_NickServ, source, "INFO ERROR: nick '%s' is not registered.",
				nick);
		else {
			User *u = finduser(nick);
			char emailbuf[70];
			char buf[512];
			long flags;
			int ltime, is_owner=is_identified(source,nick);
			int isoper = is_oper(source);
			char *mask=0;

			row = fetch(res);
			flags = atol(row[INFO_FLAGS]);
			if (flags < 0) {
				notice(s_NickServ, source,
					"Nick %s is forbidden and may not be used.", nick);
				return;
			}
			notice(s_NickServ, source,
				"\2Information for nick '%s\2'", row[INFO_NAME]);

			if (u)
				notice(s_NickServ, source,
					"(\2Currently on IRC\2) - \"/whois %s\" for "
					"more information.", nick);
	
			if (MASK_HOSTNAME)
				mask = mask_host(row[INFO_MASK]);

			notice(s_NickServ, source,
				"Last seen address: %s\n",
				(MASK_HOSTNAME && !isoper && mask) ? mask : row[INFO_MASK]);
			if (mask)
				free(mask);


			notice(s_NickServ, source,
				"  Time registered: %s", timebuf(atol(row[INFO_TIME])));


			ltime = (time(NULL) - atol(row[INFO_LAST]));

			if (ltime > 86400)
				notice(s_NickServ, source,
					"   Last seen time: %s [over %d day%s]",
					timebuf(atol(row[INFO_LAST])),
					ltime/86400, (ltime/86400 == 1) ? "" : "s");
			else if (ltime > 3600)
				notice(s_NickServ, source,
					"   Last seen time: %s [over %d hour%s]",
					timebuf(atol(row[INFO_LAST])),
					ltime/3600, (ltime/3600 == 1) ? "" : "s");
			else
				notice(s_NickServ, source,
					"   Last seen time: %s [under 1 hour]",
					timebuf(atol(row[INFO_LAST])));


			notice(s_NickServ, source,
				"         Time now: %s", timebuf(0));



			if (row[INFO_EMAIL]) {
				snprintf(emailbuf, sizeof(emailbuf), " [%s]", row[INFO_EMAIL]);
				notice(s_NickServ, source,
					"   E-Mail address: %s %s", 
					(flags & NS_HIDE_EMAIL) ? "Hidden" : row[INFO_EMAIL],
					(flags & NS_HIDE_EMAIL) && is_owner ? emailbuf : "");
			}


			if (row[INFO_ICQ])
				notice(s_NickServ, source,
					"          ICQ uin: %s", row[INFO_ICQ]);

			if (row[INFO_URL])
				notice(s_NickServ, source,
					"              URL: %s", row[INFO_URL]);

			notice(s_NickServ, source,
				"          Options: %s", show_options(buf, flags));


			if (*row[INFO_FORWARD] != '0')
				notice(s_NickServ, source,
					"  * Memos are being forwarded to %s.", row[INFO_FORWARD]);


			if (isoper && AUTH_EMAIL) {
				if (atol(row[INFO_CODE]) == 1)
					notice(s_NickServ, source,
						"  * This nick has been authorized.");
				else if (atol(row[INFO_CODE]) == 0)
					notice(s_NickServ, source,
						"  * This nick has NOT been authorized.");
				else if (atol(row[INFO_CODE]) == -1)
					notice(s_NickServ, source,
						"  * This nick was registered before the implementation "
						"of the authorization system.");

				if (sql_select("type FROM pending WHERE id = %d", id)) {
					row = fetch(res);
					if (atol(row[0]) == 2)
						notice(s_NickServ, source,
						"  * This nick has a password authorization pending.");
					if (atol(row[0]) == 3)
						notice(s_NickServ, source,
						"  * This nick has a drop authorization pending.");

				}

			}


			if (isoper && (num_rows=sql_select(
				"type,instater FROM reason WHERE name = '%s'", nick))) {
				int i;
				for (i=0; i < num_rows; i++) {
					row = fetch(res);
						notice(s_NickServ, source,
						"  * This nick has been flagged as %s by %s.",
							row[0], row[1]);
				}
			} else {
				if (is_frozen(nick))
					notice(s_NickServ, source,
					"  * This nick has been frozen by the network.");
			}

			notice(s_NickServ, source, "\2*** End of Info ***\2");

		}
	}
}

/****************************************************************************/

static void do_identify(const char *source)
{
	char *cmd = strtok(NULL, "");
	const char *nick, *pass;
	User *u = finduser(source);
	int reg=0;
	unsigned long id;


	if (!u)
		return;


	if (!cmd) {
		notice(s_NickServ, source,
			"Syntax: \2IDENTIFY\2 [nick] <password>");
		notice(s_NickServ, source,
			"\2/%s HELP IDENTIFY\2 for more information.",
			s_NickServ);
		return;
	}

	if (strchr(cmd, ' ')) {
		nick = gettok(cmd, " ", NICKMAX);
		pass = strtok(NULL, " ");
	} else { pass=cmd; nick=source; }


	id = nick_key(nick);

	if (!(reg = sql_select("pass,auth,flags,news FROM nick WHERE id=%d", id))) {
		notice(s_NickServ,source, "IDENTIFY ERROR: Nick '%s' is not registered!",
			nick);

	} else if (is_frozen(nick)) {
		notice(s_NickServ, source,
			"IDENTIFY ERROR: You may not identify to the nick '%s' because it is currently frozen.", nick);

	} else if (reg>0) {
		row = fetch(res);
		if (atol(row[2]) < 0) {
			notice(s_NickServ, source,
				"ERROR: Nick '\2%s\2' may not be used.", nick);



#ifdef USE_ENCRYPTION
		} else if (check_password(pass, row[0]) < 1) {
#else
		} else if (strcmp(row[0], pass) != 0) { /* case sensitive */
#endif
			notice(s_NickServ, source, NS_BAD_PASS, nick);
			blog("NS_I", "[NS] I* %s (%s!%s@%s)",
				nick, source, u->username, u->host);
			bad_password(u, source);


		} else {
			long stamp;

			add_idnick(u, nick);

			notice(s_NickServ, source, NS_IDENTIFY, nick);

			if (AUTH_EMAIL && (atol(row[1]) == -1)) {
				notice(s_NickServ, source,
					"Your nick was registered before the implementation of "
					"the AUTH security feature. Take an extra step in "
					"securing your nick: %s%s%s"
					"type - /%s SET EMAIL <your@email.com>",
					sstrcmp(source,nick)==0 ? "" : "While using your nick '\2",
					sstrcmp(source,nick)==0 ? "" : nick,
					sstrcmp(source,nick)==0 ? "" : "\2', ",	s_NickServ);
				if (NSECURE)
					notice(s_NickServ, source, NS_REGISTER4);
			} else if (AUTH_EMAIL && (atoi(row[1]) == 0)) {
				notice(s_NickServ, source,
					"\2REMINDER\2: You have yet to fully authorize and secure your "
					"nick!. Remember to check your e-mail and follow the authorization "
					"instructions.");
			}

			if ((unsigned int)(atoi(row[3])) < last_news_id) {
				if (u->news == last_news_id) {
					send_sql("UPDATE nick SET news = %d WHERE id = %d",
						last_news_id, id);
				} else
					notice(s_MemoServ, source,
					"There is new network news to read! Type \"\2/%s news\2\" to read it.",
					s_MemoServ);

			} else if ((unsigned int)atoi(row[3]) == last_news_id)
				u->news = last_news_id;


			stamp=generateid(u, nick, id);
			send_sql("UPDATE nick SET lasttime = %d, "
				"usermask = '%s@%s' WHERE id = %d", 
				time(NULL), u->username, u->host, id);


			if (uburst && !(u->mode & U_MEMO_NOTIFIED) && (sstrcmp(u->nick, nick) == 0))
				check_memos(u, 1);
		}

	} else if (debug)
		log("Couldn't identify nick - SQL server returned no data!");


}


/****************************************************************************/

#define SET_AUTH	0
#define SET_PASS	1
#define SET_EMAIL	2
#define SET_FLAGS	3
#define SET_URL 	4
#define SET_ICQ 	5
#define SET_QUESTION    6
#define SET_ANSWER      7


static void do_set(const char *source)
{
	const char *cmd = gettok(NULL, " ", CMDMAX);
	const char *param = strtok(NULL, " ");
	int id = nick_key(source);
	int reg = sql_select("auth,pass,email,flags,url,icq,question,answer "
		"FROM nick WHERE id = %d", id);
	long showopt=0;

	if (reg)
		row = fetch(res);

	if (!param) {
		notice(s_NickServ, source,
			"Snytax: \2SET\2 <command> <parameters>");
		notice(s_NickServ, source,
	                "\2/%s HELP SET\2 for more information.", s_NickServ);
	} else if (!reg) {
		notice(s_NickServ, source,
			"Your nickname '%s' is not registered.", source);
		notice(s_NickServ, source, 
			"For registration info, type: \2/%s HELP REGISTER\2",s_NickServ);

	} else if (!is_identified(source, source)) {
		notice(s_NickServ, source,
			"Password authentication required for the SET command.");
		notice(s_NickServ, source,
			"Retry after typing \2/%s IDENTIFY \37password\37.", s_NickServ);

	} else if (AUTH_EMAIL && NSECURE && sstrcmp(cmd, "EMAIL") && (atoi(row[SET_AUTH]) < 1)) {
		notice(s_NickServ, source,
			"You may not change any nick settings until you have fully "
			"registered your nick via e-mail authorization. "
			"If you provided a valid e-mail during registration you should "
			"have received an e-mail with instructions. "
			"In you need assistance, contact a %s representitive.", NETWORK_NAME);

	} else if (sstrcmp(cmd, "PASSWORD") == 0) {
		do_set_password(source, param, id);
	} else if (sstrcmp(cmd, "EMAIL") == 0) {
		do_set_email(source, param, id);
	} else if (sstrcmp(cmd, "SHOWEMAIL") == 0) {
		do_set_showemail(source, param, id);
	} else if ((sstrcmp(cmd, "ENFORCE") == 0) || sstrcmp(cmd, "KILL") == 0) {
		showopt=do_set_enforce(source, param, id);
	} else if (sstrcmp(cmd, "SECURE") == 0) {
		showopt=do_set_secure(source, param, id);
	} else if (sstrcmp(cmd, "URL") == 0) {
		do_set_url(source, param, id);
	} else if (sstrcmp(cmd, "ICQ") == 0) {
		do_set_icq(source, param, id);
	} else if (sstrcmp(cmd, "NOOP") == 0) {
		showopt=do_set_noop(source, param, id);
	} else if (sstrcmp(cmd, "NEVEROP") == 0) {
		showopt=do_set_neverop(source, param, id);
	} else if (sstrcmp(cmd, "PRIVATE") == 0) {
		showopt=do_set_private(source, param, id);
	} else if (sstrcmp(cmd, "MAILBLOCK") == 0) {
		do_set_mailblock(source, param, id);
	} else {
		notice(s_NickServ, source, "Unknown SET option \2%s\2.", strupper((char *)cmd));
		notice(s_NickServ, source, "For more information type "
			"\2/%s HELP SET\2", s_NickServ);
	}

	if (showopt) {
		char buf[512];
		notice(s_NickServ, source, "Options: %s", show_options(buf, showopt));
		update_nick_flags(source, showopt);
	}

}


/****************************************************************************/

static void do_set_password(const char *source, const char *param, unsigned long id)
{
	User *u = finduser(source);

	if (sstrcmp(source, param) == 0 || strlen(param) < 5) {
		notice(s_NickServ, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/%s HELP REGISTER\2 for more information.", s_NickServ);
	} else if (strlen(param) > PASSMAX) {
		notice(s_NickServ, source,
		"Sorry, nick passwords may only be up to %d characters long",
			PASSMAX);

	} else if (check_escape_codes(param)) {
		notice(s_NickServ, source,
			"Your password has invalid characters. Please try again with "
			"common characters and don't use special colour/attribute codes.");


	} else if (AUTH_EMAIL) {
		char *email = sstrdup(row[SET_EMAIL]);

		if (atol(row[SET_AUTH]) == 0) {
			notice(s_NickServ, source,
				"You may not change your password until you have fully "
				"authorized your nick's e-mail.");

#ifdef USE_ENCRYPTION
		} else if (check_password(param, row[SET_PASS]) == 0) {
#else
		} else if (strcmp(row[SET_PASS], param) == 0) {
#endif
			notice(s_NickServ, source,
				"Your password is already set to %s!", param);
		} else if (sql_select("id FROM pending WHERE id = %d", id)) {
			notice(s_NickServ, source,
				"ERROR: Nick '\2%s\2' is already in another authorization phase. "
				"Check your e-mail and complete it first, or wait for the 72 hour pend to "
				"expire. If you need assistance contact a %s representitive",
					source, NETWORK_NAME);
		} else {
			send_auth_email(source, email, param, 2, id);
			notice(s_NickServ, source,
				"An e-mail has just been sent to you with instructions on how "
				"to authorize this password change.");
		}
		free(email);

	} else {
		notice(s_NickServ, source, "Password changed to \2%s\2.", param);
		send_sql("UPDATE nick SET pass = '%s' WHERE id = %d",
			param, id);
		if (u)
			blog("NS_O", "[NS] P %s (%s@%s)",
				u->nick, u->username, u->host);

	}
}

/****************************************************************************/

static void do_set_email(const char *source, const char *param, unsigned long id)
{
	int none=0;


/* check auth for -1 incase user set an email before the authorization system was in place */

	if (row[SET_EMAIL] && sstrcmp(param, row[SET_EMAIL]) == 0 && atoi(row[SET_AUTH]) != -1) {
		notice(s_NickServ, source,
			"Your e-mail address is already set to %s!", param);

	} else if (!((none = sstrcmp(param, "none"))) && AUTH_EMAIL) {
		notice(s_NickServ, source,
			"ERROR: You may not remove your authorized e-mail. "
			"Although, you may change the e-mail or hide it for privacy. "
			"For more information type: \2/%s HELP SET EMAIL\2", s_NickServ);

	} else if (!none && !AUTH_EMAIL) {
		if (row[SET_EMAIL] == NULL)
			notice(s_NickServ, source,
				"You have no email set to disable!");
		else {
			long flags = atol(row[SET_FLAGS]);
			flags &= ~NS_HIDE_EMAIL;
			send_sql("UPDATE nick SET email = NULL, flags = %d, auth=-1  WHERE id = %d",
				flags, id);
			update_nick_flags(source, flags);
			notice(s_NickServ, source,
				"E-Mail has been disabled for nick '\2%s\2'", source);
		}

	} else if (none && !AUTH_EMAIL) {
		if (!check_valid_email(param)) {
			notice(s_NickServ, source, INVALID_EMAIL);
			return;
		}
		notice(s_NickServ, source,
			"E-Mail for '%s' changed to \2%s\2.",
			source, param);
		send_sql("UPDATE nick SET email = '%s' WHERE id = %d",
			param, id);

	} else if (none && AUTH_EMAIL) {
		char *answer = strtok(NULL, "");
		int authed = atol(row[SET_AUTH]);

		if (HIGH_NICK_SECURITY && !answer && authed != -1) {
			notice(s_NickServ, source,
				"You must provide the answer to your personal question to be able to "
				"change your email!.");
			notice(s_NickServ, source, "To see your personal question, type: "
				"\2/%s myquestion", s_NickServ);
			notice(s_NickServ, source, "Syntax: SET EMAIL <new@email.com> <answer>");

		} else if (HIGH_NICK_SECURITY 
			&& (authed != -1) && answer && row[SET_ANSWER] && sstrcmp(answer, row[SET_ANSWER]))
			notice(s_NickServ, source,
				"Sorry, your personal question is incorrect.");

		else if (!check_valid_email(param))
			notice(s_NickServ, source, INVALID_EMAIL);

		else if (check_bad_email(param))
			notice(s_NickServ, source, BAD_EMAIL);

		else if (authed == 0) {
			notice(s_NickServ, source,
				"You may not change your e-mail address until you have fully authorized "
				"your old one. Check your e-mail and follow it's instructions.");

/* First time authorization if registered before authorization system was in place.
 * Set auth to 0 to make sure user can't run this command more than once */
		} else if (atol(row[SET_AUTH]) == -1) {
			send_auth_email(source, param, param, 0, id);
			notice(s_NickServ, source, NS_REGISTER3);
			send_sql("UPDATE nick SET auth = 0 WHERE id = %d", id);


		} else {  /* changing e-mails */
			if (sql_select("id FROM pending WHERE id = %d", id)) {
				notice(s_NickServ, source,
		                        "ERROR: Nick '\2%s\2' is already in another authorization phase. "
	        	               	"Check your e-mail and complete it first, or wait for the 72 hour "
                		        "pend to expire. If you need assistance, contact a %s representative.",
						source, NETWORK_NAME);
			} else {
				send_auth_email(source, param, param, 1, id);
				notice(s_NickServ, source, NS_REGISTER3);
			}
		}


	} else {
		notice(s_NickServ, source, "Syntax: \2SET EMAIL\2 [\37<e-mail>\37%s]",
			AUTH_EMAIL ? "" : "|\37none\37");
		notice(s_NickServ, source,
		"\2/%s HELP SET EMAIL\2 for more information.", s_NickServ);
	}


}


/****************************************************************************/

static void do_set_showemail(const char *source, const char *param, unsigned long id)
{

	if (!sstrcmp(param,"OFF") || !sstrcmp(param, "HIDE")) {
		if (row[SET_EMAIL] == NULL) {
			notice(s_NickServ, source,
				"You have no e-mail address to hide!");
			notice(s_NickServ, source,
				"For more information, type \2/%s HELP SET EMAIL\2.",
				s_NickServ);
			return;
		} else {

			long flags = atol(row[SET_FLAGS]);
			flags |= NS_HIDE_EMAIL;
			send_sql("UPDATE nick SET flags = '%d' WHERE id = %d",
				flags, id);
			update_nick_flags(source, flags);
			notice(s_NickServ, source,
				"Your e-mail address will no longer be shown.");
		}
	} else if (!sstrcmp(param,"ON") || !sstrcmp(param, "show")) {
		if (row[SET_EMAIL] == NULL) {
			notice(s_NickServ, source,
				"You have no e-mail address set!");
			notice(s_NickServ, source,
				"For more information, type \2/%s HELP SET EMAIL\2.",
				s_NickServ);
			return;
		} else {

			long flags = atol(row[SET_FLAGS]);
			flags &= ~NS_HIDE_EMAIL;
			send_sql("UPDATE nick SET flags = %d WHERE id = %d",
				flags, id);
			update_nick_flags(source, flags);
			notice(s_NickServ, source,
				"Your e-mail address will now be visible to all.");
		}

	} else {
		notice(s_NickServ, source, "Syntax: \2SET SHOWEMAIL\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET SHOWEMAIL\2 for more information.", s_NickServ);

	}
}

/****************************************************************************/

static long do_set_enforce(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);

	if (sstrcmp(param, "ON") == 0) {
		flags |= NS_ENFORCE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Nick Enforcement is now \2ON\2.");
	} else if (sstrcmp(param, "OFF") == 0) {
		flags &= ~NS_ENFORCE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Nick Enforcement is now \2OFF\2.");
	} else {
		notice(s_NickServ, source, "Syntax: \2SET ENFORCE\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET ENFORCE\2 for more information.", s_NickServ);
		return 0;

	}
	return flags;
}

/****************************************************************************/

static long do_set_secure(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);

	if (sstrcmp(param, "ON") == 0) {
		flags |= NS_SECURE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Nick security is now \2ON\2.");
	} else if (sstrcmp(param, "OFF") == 0) {
		flags &= ~NS_SECURE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Nick security is now \2OFF\2.");
	} else {
		notice(s_NickServ, source, "Syntax: \2SET SECURE\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET SECURE\2 for more information.", s_NickServ);
		return 0;
	}
	return flags;
}

/****************************************************************************/

static void do_set_url(const char *source, const char *param, unsigned long id)
{

	if (sstrcmp(param,"none")) {
		if (check_valid_url(source,param)) {
			notice(s_NickServ, source,
				"URL for '%s' changed to \2%s\2.",
				source, param);
			send_sql("UPDATE nick SET url = '%s' WHERE id = %d",
				param, id);
		}
	} else if (!sstrcmp(param,"none")) {
		if (row[SET_URL] == NULL)
			notice(s_NickServ, source,
				"You have no URL set to disable!");
		else {
			send_sql("UPDATE nick SET url = NULL WHERE id = %d", id);
			notice(s_NickServ, source,
				"Your URL has been disabled for nick '\2%s\2'", source);
		}
	} else {
		notice(s_NickServ, source, "Syntax: \2SET URL\2 [\37<url://host>\37|\37none\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET URL\2 for more information.", s_NickServ);
	}

}

/****************************************************************************/

static void do_set_icq(const char *source, const char *param, unsigned long id)
{
	if (sstrcmp(param,"none")) {
		if ((strspn(param, "1234567890") != strlen(param)) 
			|| strlen(param) > 20) {
			notice(s_NickServ, source,
				"ERROR: invalid ICQ number. Your UIN may only contain "
				"the numbers 0-9 and be only up to 20 digits in length.");
			return;
		}
		notice(s_NickServ, source,
			"Your nick's ICQ uin is now set to: \2%s\2.", param);
		send_sql("UPDATE nick SET icq = '%s' WHERE id = %d",
			param, id);

	} else if (!sstrcmp(param,"none")) {
		if (row[SET_ICQ] == NULL)
			notice(s_NickServ, source,
				"You have no ICQ uin set to disable!");
		else {
			send_sql("UPDATE nick SET icq = NULL WHERE id = %d",
				id);
			notice(s_NickServ, source,
				"Your ICQ uin has been disabled for nick '\2%s\2'", source);
		}
	} else {
		notice(s_NickServ, source, "Syntax: \2SET ICQ\2 [\37<uin>\37|\37none\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET ICQ\2 for more information.", s_NickServ);
	}

}

/****************************************************************************/

static long do_set_noop(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);

	if (sstrcmp(param, "ON") == 0) {
		flags |= NS_NOOP;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "The \2NO OP\2 setting is now enabled.");
	} else if (sstrcmp(param, "OFF") == 0) {
		flags &= ~NS_NOOP;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "The \2NO OP\2 is now disabled.");
	} else {
		notice(s_NickServ, source, "Syntax: \2SET NOOP\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET NOOP\2 for more information.", s_NickServ);
		return 0;
	}
	return flags;
}

/****************************************************************************/

static long do_set_neverop(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);

	if (sstrcmp(param, "ON") == 0) {
		flags |= NS_NEVEROP;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "The \2NEVER OP\2 setting is now enabled.");
	} else if (sstrcmp(param, "OFF") == 0) {
		flags &= ~NS_NEVEROP;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "The \2NEVER OP\2 setting is now disabled.");
	} else {
		notice(s_NickServ, source, "Syntax: \2SET NEVEROP\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET NEVEROP\2 for more information.", s_NickServ);
		return 0;
	}
	return flags;
}

/****************************************************************************/

static long do_set_private(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);

	if (sstrcmp(param, "ON") == 0) {
		flags |= NS_PRIVATE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Your nick will no longer appear in generated lists.");

	} else if (sstrcmp(param, "OFF") == 0) {
		flags &= ~NS_PRIVATE;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d",
			flags, id);
		notice(s_NickServ, source, "Your nick will now be included in generated lists.");
	} else {
		notice(s_NickServ, source, "Syntax: \2SET PRIVATE\2 [\37ON\37|\37OFF\37]");
		notice(s_NickServ, source,
		"\2/%s HELP SET PRIVATE\2 for more information.", s_NickServ);

	}
	return flags;
}


/****************************************************************************/

static void do_set_mailblock(const char *source, const char *param, unsigned long id)
{
	long flags = atol(row[SET_FLAGS]);


	if (AUTH_EMAIL) {
		char *email = sstrdup(row[SET_EMAIL]);

		if (atol(row[SET_AUTH]) == 0) {
			notice(s_NickServ, source,
				"You may not alter your settings until you have fully "
				"authorized your nick's e-mail.");

		} else if ((sstrcmp(param, "ON") == 0) && (flags & NS_MAILBLOCK)) {
			notice(s_NickServ, source,
				"You have already set MAILBLOCK for your nick.");


		} else if ((sstrcmp(param, "OFF") == 0) && !(flags & NS_MAILBLOCK)) {
			notice(s_NickServ, source,
				"MAILBLOCK is already disabled for your nick.");

		} else if ((sstrcmp(param, "OFF") != 0 && (sstrcmp(param, "ON") != 0))) {
			notice(s_NickServ, source, "Syntax: \2SET MAILBLOCK\2 [\37ON\37|\37OFF\37]");
			notice(s_NickServ, source,
			"\2/%s HELP SET MAILBLOCK\2 for more information.", s_NickServ);

		} else if (sql_select("id FROM pending WHERE id = %d", id)) {
			notice(s_NickServ, source,
				"ERROR: Nick '\2%s\2' is already in another authorization phase. "
				"Check your e-mail and complete it first, or wait for the 72 hour pend to "
				"expire. If you need assistance contact a %s representitive",
					source, NETWORK_NAME);
		} else {
			send_auth_email(source, email, email, 4, id);
			notice(s_NickServ, source,
				"An e-mail has just been sent to you with instructions on how "
				"to authorize this MAILBLOCK setting.");
		}
		free(email);



	} else {
		if (sstrcmp(param, "ON") == 0) {
			flags |= NS_MAILBLOCK;
			send_sql("UPDATE nick SET flags = %d WHERE id = %d",
				flags, id);
			notice(s_NickServ, source, "SENDPASS disabled - your password no longer can be e-mailed to you.");

		} else if (sstrcmp(param, "OFF") == 0) {
			flags &= ~NS_MAILBLOCK;
			send_sql("UPDATE nick SET flags = %d WHERE id = %d",
				flags, id);
			notice(s_NickServ, source, "SENDPASS enabled - you may now receive your password via e-mail.");

		} else {
			notice(s_NickServ, source, "Syntax: \2SET MAILBLOCK\2 [\37ON\37|\37OFF\37]");
			notice(s_NickServ, source,
			"\2/%s HELP SET MAILBLOCK\2 for more information.", s_NickServ);

		}
	}
}

/****************************************************************************/

static void do_access(const char *source)
{
	NickInfo *ni;
	User *u = finduser(source);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	const char *mask = strtok(NULL, " ");
	
	if (!u)
		return;

	if (cmd && mask && !sstrcmp(mask, "LIST") && has_access(source, 2)) {
		if (!(ni = findnick(cmd)))
			notice(s_NickServ, source, "Nick '\2%s\2' is not registered!", cmd);
		else {
			int i, num_rows=sql_select("mask FROM naccess WHERE id = %d", ni->id);
			if (!num_rows)
				notice(s_NickServ, source, "Nick '\2%s\2' has no access list enteries.",
					cmd);
			else {
				notice(s_NickServ, source, "\2Access list for: %s\2", cmd);
				for (i=0; i < num_rows; i++) {
					row = fetch(res);
					notice(s_NickServ, source,
						"  %d)  %s", i+1, row[0]);
				}
				notice(s_NickServ, source, "\2End of Access List.\2");
			}
		}

	} else if (!cmd || ((!sstrcmp(cmd, "add") || !sstrcmp(cmd, "del"))
		&& !mask)) {
		notice(s_NickServ, source,
			"Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
		notice(s_NickServ, source,
			"type: \2/%s HELP ACCESS\2 for more information.",
			s_NickServ);

	} else if (!(ni = findnick(source))) {
		notice(s_NickServ, source,
			"You must register your nick in order to use the ACCESS command.");

	} else if (mask && (strspn(mask, "abcdefghijklmnopqrstuvwxyz1234567890"
			"ABCDEFGHIJKLMNOPQRSTUVWXYZ.@_-*") != strlen(mask))) {
		notice(s_NickServ, source,
			"ERROR: Invalid mask characters.");

	} else if (!sstrcmp(cmd, "LIST") && ((u->mode & U_RECOGNIZED) || is_identified(source,source))) {
/* we don't use the SQL 'LIKE' clause because that would throw off our numeric listing for ACCESS DEL */
		int num_rows=sql_select("mask FROM naccess WHERE id = %d", ni->id);

		if (!num_rows)
			notice(s_NickServ, source,
				"You have no mask entries in your ACCESS list.");
		else {
			int i;

			notice(s_NickServ, source, "Access list:");
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				if (mask && !match_wild_nocase(mask, row[0]))
					continue;
				notice(s_NickServ, source,
					"  %d)  %s", i+1, row[0]);
			}
			notice(s_NickServ, source, "\2End of Access List.\2");
		}
	} else if (!is_identified(source, source)) {
		notice(s_NickServ, source,
			"Password authentication is required for that command");
	} else if (READ_ONLY) {
		notice(s_NickServ, source,
			"Sorry, that command is temporarly disabled while services are in read-only mode.");
	} else if (!sstrcmp(cmd, "ADD")) {
		if (!strchr(mask, '@') || strchr(mask, '!')) {
			notice(s_NickServ, source,
				"Access mask must be in the form of \37user\37@\37host\37.");
			notice(s_NickServ, source,
				"Type: \2/%s HELP ACCESS\2 for more information.",
				 s_NickServ);
		} else if (strlen(mask) > 80) {
			notice(s_NickServ, source, "ACCESS ERROR: Access mask too large!");
		} else if (!sstrcmp(mask, "*@*") || !sstrcmp(mask, "*@*.*")) {
			notice(s_NickServ, source, "Please try to add a more secure access mask");
		} else {
			int num_rows=sql_select("mask FROM naccess WHERE id = %d", ni->id);

			if (num_rows > ACCESS_MAX) {
				notice(s_NickServ, source,
					"ACCESS ERROR: You have exceeded your maximum allowed "
					"access list entries. Delete some to add more.");

			} else if (num_rows) {
				int i;
				for (i=0; i < num_rows; i++) {
					row = fetch(res);
					if (!sstrcmp(mask, row[0])) {
						notice(s_NickServ, source,
							"Mask '%s' already on your access list!",
							mask);
						return;
					}
				}
			}
			add_ni_access(findnick(source), mask);
			send_sql("INSERT into naccess VALUES (%d,'%s')",
				ni->id, mask);
			notice(s_NickServ, source,
				"Mask '%s' added to your access list", mask);
		}
	} else if (!sstrcmp(cmd, "DEL")) {
		int maskid;
		int num_rows=sql_select("mask FROM naccess WHERE id = %d", ni->id);

		if (!num_rows) {
			notice(s_NickServ, source, "Your nick's access list is empty!");

		} else if ((strspn(mask, "1234567890") == strlen(mask))
			&& ((maskid = atoi(mask)) > 0) && maskid <= num_rows) {

			mysql_data_seek(res, maskid-1);
			row = fetch(res);
			notice(s_NickServ, source,
				"Mask '\2%s\2' removed from your access list.", row[0]);
			send_sql("DELETE FROM naccess WHERE id = %d AND mask = '%s'",
				ni->id, row[0]);
			del_ni_access(source, row[0]);

		} else {
			int findmask=sql_select("* FROM naccess WHERE id = %d AND mask = '%s'",
				ni->id, mask);
			if (!findmask)
				notice(s_NickServ, source,
					"Entry '%s' not on access list!", mask);
			else {
				del_ni_access(source, mask);
				send_sql("DELETE from naccess WHERE id = %d AND mask = '%s'",
					ni->id, mask);
				notice(s_NickServ, source,
					"Entry '%s' deleted from your access list.", mask);
			}
		}
	} else if (!sstrcmp(cmd, "WIPE")) {
		int num_rows=sql_select("* FROM naccess WHERE id = %d", ni->id);
		if (!num_rows)
			notice(s_NickServ, source, "Your nick's access list is empty!");
		else {
			del_ni_accesslist(findnick(source));
			send_sql("DELETE FROM naccess WHERE id = %d", ni->id);
			notice(s_NickServ, source, "All masks on your access list have been deleted.");
		}
	} else {
		notice(s_NickServ, source,
			"Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
		notice(s_NickServ, source,
			"type: \2/%s HELP ACCESS\2 for more information.",
			s_NickServ);
	}
}


/****************************************************************************/

static void do_drop(const char *source)
{
	User *u = finduser(source);
	const char *nick = gettok(NULL, " ", NICKMAX);
	int reg;
	unsigned long id = nick_key(nick);

	if (!u)
		return;

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2DROP\2 <nick>");
		notice(s_NickServ, source,
			"For more information type: \2/%s HELP DROP\2",
				s_NickServ);
	} else if (sql_select("type FROM pending WHERE id = %d", id)) {
		notice(s_NickServ, source,
			"ERROR: Nick '\2%s\2' is already in another authorization phase. "
			"Check your e-mail and complete it first, or wait for the 72 hour pend to "
			"expire. If you need assistance, contanct a %s representitive.",
				nick, NETWORK_NAME);

	} else if (!(reg = sql_select("auth,email,flags FROM nick WHERE id = %d", id))) {
		notice(s_NickServ, source, "DROP ERROR: Nick '\2%s\2' isn't registered!", nick);

	} else if (!is_identified(source, nick)) {
		notice(s_NickServ, source, "You are not identified to nick '\2%s\2'. "
			"Password authentication is required.", nick);
		notice(s_NickServ, source,
			"Retry after typing: \2/%s IDENTIFY %s <\37password\37>\2",
				s_NickServ, nick);
		blog("NS_O", "[NS] D* %s (%s!%s@%s) [User hasn't identified]",
			nick, source, u->username, u->host);

	} else if (is_frozen(nick)) {
		notice(s_NickServ, source,
			"DROP ERROR: You may not drop the nick '%s' because it is currently frozen.", nick);

	} else if (nick_flags(nick) & NS_HELD) {
		notice(s_NickServ, source,
			"DROP ERROR: You may not drop the nick '%s' because it is currently held.", nick);


	} else if (is_on_master_list(nick)) {
		notice(s_NickServ, source,
			"DROP ERROR: You may not drop the nick '%s' until it is removed from the m:lines!",nick);

	} else if (!AUTH_EMAIL) {
		del_ni_nick(nick, 1);
		del_sql_nick(nick);
		notice(s_NickServ, source,
			"The nick '\2%s\2' has been successfully deleted.", nick);

	} else if (AUTH_EMAIL) {
		row = fetch(res);

		if (atol(row[0]) < 1) {
			notice(s_NickServ, source,
				"You may not drop the nick '\2%s\2' until you have "
				"properly authorized your e-mail.", nick);
		} else {
			if (check_email(nick, BLOCK_AUTH_EMAIL, 0)) {
				notice(s_NickServ, source,
					"DROP ERROR: A drop attempt has already been made recently. "
					"For security reasons, you may not attempt to drop this nick "
					"again for a disclosed period of time.");
				return;
			} else
				add_email(nick, BLOCK_AUTH_EMAIL);

			notice(s_NickServ, source, "A drop authorization code has just been "
				"e-mailed to you. Check your e-mail shortly for instructions "
				"on how to complete the nick drop.");


			send_auth_email(source, row[1], row[1],3, id);

		}
	}
}

/****************************************************************************/

static void do_acc(const char *source) {

	const char *nick = gettok(NULL, " ", NICKMAX);


	if (!nick) {
		notice(s_NickServ, source, "Syntax: \2ACC\2 <nick>");
		notice(s_NickServ, source,
		"Type: \2/%s HELP ACC\2 for more information.", s_NickServ);
		return;
	} else if (!nick_registered(nick)) {
		notice(s_NickServ, source,
			"%s ACC: 0 [UnRegistered]", nick);
	} else if (nick_flags(nick) < 0) {
		notice(s_NickServ, source,
			"%s ACC: 5 [Forbidden]", nick);
	} else if (!finduser(nick)) {
		notice(s_NickServ, source,
			"%s ACC: 1 [Registered/Offline]", nick);
	} else if (is_identified(nick, nick)) {
		notice(s_NickServ, source,
			"%s ACC: 3 [Identified]", nick);
	} else if (is_recognized(nick, NULL)) {
		notice(s_NickServ, source,
			"%s ACC: 4 [Recognized]", nick);
	} else {
		notice(s_NickServ, source,
			"%s ACC: 2 [UnIdentified/UnRecognized]", nick);
	}

}

/****************************************************************************/

static void do_ghost(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *pass = strtok(NULL, " ");
	User *u = finduser(source);
	int rec = 0;
	unsigned long id = nick_key(nick);

	if (!u) {
		return;

	} else if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2GHOST\2 <\37nickname\37> [<\37password\37>]");
		notice(s_NickServ, source,
			"\2/%s HELP GHOST\2 for more information.", s_NickServ);

	} else if (!(sql_select("pass FROM nick WHERE id = %d", id))) {
		notice(s_NickServ, source,
			"GHOST ERROR: Nick '\2%s\2' is not registered.", nick);

	} else if (!finduser(nick)) {
		notice(s_NickServ, source,
			"GHOST ERROR: Nick '\2%s\2' is not currently in use.", nick);

	} else if (sstrcmp(nick, source) == 0) {
		notice(s_NickServ, source,
			"GHOST ERROR: You may not ghost yourself!");
	} else if (!pass && (rec=is_recognized(source, nick))) {
		char buf[256];
		sprintf(buf, "Ghost command used by \2'%s'\2", source);
		notice(s_NickServ, source,
			"Successfully removed ghost nick '\2%s\2'", nick);
		kill_user(s_NickServ, nick, buf);
		blog("NS_M", "[NS] K %s (%s!%s@%s)",
			nick, source, u->username, u->host);
	} else if (!pass && !rec) {
		notice(s_NickServ, source,
			"You do not have access to GHOST the nick '\2%s\2'.", nick);
		notice(s_NickServ, source,
			"Try typing: \2/%s GHOST %s <\37password\37>\2",
				s_NickServ, nick);
	} else if (pass) {
		row = fetch(res);

		if (strcmp(pass, row[0])) {   /* case sensitive */
			notice(s_NickServ, source,
				"GHOST ERROR: Invalid password for nick '\2%s\2'", nick);
			blog("NS_M", "[NS] K* %s (%s!%s@%s)",
				nick, source, u->username, u->host);
		} else {
			char buf[50 + NICKMAX];
			sprintf(buf, "Ghost command used by \2'%s'\2", source);
			add_idnick(u, nick); generateid(u, nick, 0);
			notice(s_NickServ, source,
				"Successfully removed ghost nick '\2%s\2'", nick);
			kill_user(s_NickServ, nick, buf);
			blog("NS_M", "[NS] K %s (%s!%s@%s)",
				nick, source, u->username, u->host);
		}
	} else
		notice(s_NickServ, source, "Unknown GHOST error");
}

/****************************************************************************/

static void do_sendpass(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *email = strtok(NULL, " ");
	User *u = finduser(source);
	unsigned long id = nick_key(nick);


#ifdef USE_ENCRYPTION
	notice(s_NickServ, source,
		"SENDPASS has been disabled by the network.");
	return;
#endif


	if (!u)
		return;

	else if (!AUTH_EMAIL)
		notice(s_NickServ, source,
			"E-Mail systems have been disabled by the network.");

	else if (!nick || !email)
		notice(s_NickServ, source,
			"Syntax: \2SENDPASS\2 <nickname> <your@email.com>");

	else if (!sql_select("email,auth,email,pass,flags FROM nick WHERE id = %d", id))
		notice(s_NickServ, source,
			"SENDPASS ERROR: Nick '\2%s\2' is not registered!", nick);

	else {
		row = fetch(res);
		if (!row[0])
			notice(s_NickServ, source,
				"SENDPASS ERROR: Could not retrieve e-mail information for nick '%s'.", nick);

		else if (atol(row[4]) < 0)
			notice(s_NickServ, source,
				"Nick '\2%s\2' may not be used.", nick);
		else if (atol(row[1]) != 1) {
			blog("NS_S", "[NS] S* %s (%s!%s@%s) [Not authorized]", 
				nick, source, u->username, u->host);
			notice(s_NickServ, source,
				"SENDPASS ERROR: Nick '\2%s\2' has not authorized yet.", nick);
		} else if (atol(row[4]) & NS_MAILBLOCK) {
			notice(s_NickServ, source,
				"SENDPASS has been disabled for this nick.");

		} else if (sstrcmp(row[2], email) != 0) {
			blog("NS_S", "[NS] S* %s (%s!%s@%s) [E-Mail discrepancy]",
				nick, source, u->username, u->host);
			notice(s_NickServ, source,
				"SENDPASS ERROR: The e-mail you specified does not match "
				"with the one on file!. This attempt has been logged.");
		} else {
			char emsg[1024];
			char esubj[128];

			if (check_email(nick, BLOCK_SENDPASS, 0)) {
				notice(s_NickServ, source,
					"SENDPASS ERROR: A SENDPASS attempt has already been made recently. "
					"For security reasons, you may not attempt to SENDPASS this nick "
					"again for a disclosed period of time.");
				return;
			} else
				add_email(nick, BLOCK_SENDPASS);


			blog("NS_S", "[NS] S %s (%s!%s@%s)", nick, source, u->username, u->host);
			globops(s_NickServ, "%s used SENDPASS on %s", source, nick);
			notice(s_NickServ, source,
				"Password has been sent via e-mail to %s", email);


			snprintf(emsg, sizeof(emsg),
				"Someone, possibly you, has requested the password for the nick '%s' to "
				"be sent via e-mail. "
				"Your nickname's password is %s. "
				"Log back onto %s using the nick %s and type - "
				"/%s IDENTIFY %s\n"
				"If you did not request this action to be taken, you may wish to have one "
				"of the network administrators investigate this.",
				nick, row[3], NETWORK_NAME, nick, s_NickServ, row[3]);

			snprintf(esubj, sizeof(esubj), "%s password retrieval", NETWORK_NAME);
			e_mail(emsg, esubj, email, EMAIL_NS_REG);
		}
	}

}

/****************************************************************************/

static void do_getpass(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	User *u = finduser(source);
	unsigned long id = nick_key(nick);


#ifdef USE_ENCRYPTION
	notice(s_NickServ, source,
		"GETPASS has been disabled by the network.");
	return;
#endif

	if (!u)
		return;

	else if (!ENABLE_GETPASS)
		notice(s_NickServ, source,
			"GETPASS has been disabled by the network.");

	else if (!nick)
		notice(s_NickServ, source,
			"Syntax: \2GETPASS\2 <nickname>");

	else if (sql_select("instater FROM reason,flags WHERE flags.name='%s' "
		"AND flags.name=reason.name AND MARKED > 0", nick)) {
		row = fetch(res);
		notice(s_NickServ, source,
			"GETPASS ERROR: Nick '\2%s\2' has been marked by '%s' and may not be getpassed!",
				nick, row[0]);

	} else if (is_on_root_list(nick)) {
		notice(s_NickServ, source,
			"GETPASS ERROR: You may not getpass a services root's nick!");
		globops(s_NickServ, "'\2%s\2' tried to GETPASS the services root nick '%s'",
			source, nick);

	} else if (!sql_select("pass,flags FROM nick WHERE id = %d",id))
		notice(s_NickServ, source,
			"GETPASS ERROR: Nick '\2%s\2' is not registered!", nick);


	else {
		row = fetch(res);

		blog("NS_S", "[NS] G %s (%s!%s@%s)", 
			nick, source, u->username, u->host);

		globops(s_NickServ, "%s used GETPASS on nickname '%s'", opertrace(u,0), nick);
		notice(s_NickServ, source,
			"Nick password retrieved for %s: \2%s\2", nick, row[0]);

	}

}

/****************************************************************************/

inline static void timeout_release(Timeout *t)
{
	send_cmd(NULL, ":%s QUIT :nick released.", t->data);
	free(t->data);
}


/****************************************************************************/

static void do_recover(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *pass = strtok(NULL, " ");
	User *u = finduser(source);
	int recover=0;
	unsigned long id = nick_key(nick);


	if (!u)
		return;

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2RECOVER\2 <nickname> <password>");
		notice(s_NickServ, source,
		"\2/%s HELP RECOVER\2 for more information.", s_NickServ);

	} else if (!finduser(nick)) {
		notice(s_NickServ, source,
			"RECOVER ERROR: Nick '\2%s\2' is not currently in use.", nick);

	} else if (sstrcmp(source, nick) == 0) {
		notice(s_NickServ, source,
			"RECOVER ERROR: You may not attempt to recover yourself!");

	} else if (!sql_select("pass,flags FROM nick WHERE id = %d", id)) {
		notice(s_NickServ, source,
			"RECOVER ERROR: Nick '\2'%s\2 isn't registered!", nick);

	} else if (!pass && is_recognized(source, nick)) {
		recover=1;

	} else if (!pass && !is_recognized(source, nick)) {
		notice(s_NickServ, source,
			"RECOVER ERROR: You do not have access to recover this nick.");
		notice(s_NickServ, source,
			"Try again supplying a password, type - /%s RECOVER <password>",
				s_NickServ);
		blog("NS_M", "[NS] K* %s (%s!%s@%s)",
			nick, source, u->username, u->host);
	} else if (pass) {
		row = fetch(res);
		if (strcmp(pass, row[0])) {  /* case sensitive */
			notice(s_NickServ, source,
				"RECOVER ERROR: Password invalid to recover nick '\2%s\2'.", nick);
			blog("NS_M", "[NS] K* %s (%s!%s@%s)",
				nick, source, u->username, u->host);
			bad_password(u, source);
		} else
			recover=1;
	}
	if (recover) {
#ifndef BAHAMUT
		Timeout *enforcer = add_timeout(RELEASE_TIMEOUT, timeout_release, 0);
		enforcer->data = sstrdup(nick);
		enforcer->flagged = 1;
#endif

		add_idnick(u, nick); generateid(u, nick, 0);
		notice(s_NickServ, nick,
			"Your nick is being changed due to the owner claiming it back.");

#ifdef BAHAMUT
		send_cmd(NULL, "SVSHOLD %s %d :This nick is currently being protected for the "
			"owner. If you own this nick, type \"/%s RELEASE %s <password>\" to use it.", 
			nick, RELEASE_TIMEOUT, s_NickServ, nick);
		send_cmd(NULL, "SVSNICK %s Guest %d",
			nick, time(NULL));
#else
		send_cmd(NULL, "SVSNICK %s Guest%d %d",
			nick, getrandom(10000,99999), time(NULL));
                send_cmd(NULL, "NICK %s 1 %lu + enforcer %s %s 0 0 :Nick Protection Enforcement",
                        nick, time(NULL), services_host, servername);
#endif
		notice(s_NickServ, source,
			"Your nick '\2%s\2' has been successfully recovered.", nick);
		notice(s_NickServ, source,
			"Type \2/%s RELEASE\2 or wait 1 minute if you would like to use "
			"this nick.", s_NickServ);
		blog("NS_M", "[NS] K %s (%s!%s@%s)",
			nick, source, u->username, u->host);
	}
}

/****************************************************************************/

static void do_release(const char *source)
{
	NickInfo *ni;
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *pass = strtok(NULL, " ");
	int release=0;
	User *u = finduser(source);

	if (!u)
		return;

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2RELEASE\2 <nickname> [<password>]");
		notice(s_NickServ, source,
			"\2/%s HELP RELEASE\2 for more information.", s_NickServ);

	} else if (!(ni = findnick(nick))) {
		notice(s_NickServ, source,
			"RELEASE ERROR: Nick '\2%s\2' isn't registered!", nick);
#ifndef BAHAMUT

	} else if (!check_release(nick)) {
		notice(s_NickServ, source,
			"RELEASE ERROR: Nick '\2%s\2' is not being held!", nick);
#endif

	} else if (!pass && is_recognized(source, nick)) {
		release=1;

	} else if (!pass && !is_recognized(source, nick)) {
		notice(s_NickServ, source,
			"RELEASE ERROR: You do not have access to release this nick.");
		notice(s_NickServ, source,
			"Try again supplying a password, type - /%s RELEASE <password>",
				s_NickServ);
		blog("NS_M", "[NS] K* %s (%s!%s@%s)",
			nick, source, u->username, u->host);

	} else if (pass) {
		sql_select("pass FROM nick WHERE id = %d", ni->id);
		row = fetch(res);
		if (row[0] && strcmp(pass, row[0])) { /* case sensitive */
			notice(s_NickServ, source,
				"RELEASE ERROR: Password invalid to release nick '\2%s\2'.", nick);
			blog("NS_M", "[NS] K* %s (%s!%s@%s)",
				nick, source, u->username, u->host);
			bad_password(u, source);
		} else
			release=1;

	}
	if (release) {
		add_idnick(u, nick); generateid(u, nick, 0);
		notice(s_NickServ, source,
			"Your nick '\2%s\2' has been successfully released.", nick);
#ifdef BAHAMUT
		send_cmd(NULL, "SVSHOLD %s 0", nick);
#else
		send_cmd(nick, "QUIT :released nick.", nick);
		del_ntimeout(nick, 1);
#endif

	}
}


/****************************************************************************/

static void myquestion(const char *source)
{
	if (!sql_select("question FROM nick WHERE id = %d", nick_key(source)))
		notice(s_NickServ, source, "ERROR: Nick '\2%s\2' isn't registered!", source);

	else if (!is_identified(source, source))
		notice(s_NickServ, source,
			"Password authentication is required before you may issue that command.");
	else {
		row = fetch(res);
		if (row[0] == NULL)
			notice(s_NickServ, source, "You do not have a personal question set.");
		else {
			switch(atoi(row[0])) {
				case 1:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_1);
					break;
				case 2:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_2);
					break;
				case 3:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_3);
					break;
				case 4:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_4);
					break;
				case 5:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_5);
					break;
				case 6:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_6);
					break;
				case 7:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_7);
					break;

			}
		}
	}

}



/****************************************************************************/

static void ohelp(const char *source) {

	Command *c;
	int shown=0;
	unsigned int level;
	User *u = finduser(source);

	if (!u)
		return;

	notice(s_NickServ, source,
		"This is a listing of %s's privileged commands.", s_NickServ);
	notice(s_NickServ, source,
		"Your current access level is: %s.",
		u->os_level==0 ? "IRC Operator" :
		u->os_level==1 ? "Services Operator" :
		u->os_level==2 ? "Services Administrator" :
		u->os_level==3 ? "Services Root" :
		u->os_level>3 ? "Services Root Master" : "");


	notice(s_NickServ, source, "For a detailed explanation for a command, type:");
	notice(s_NickServ, source, "\2/%s HELP <command>", s_NickServ);

	for (level=1; level < 5; level++, shown=0) {
		notice(s_NickServ, source, " ");
		notice(s_NickServ, source, "%s (%s)",
			level==1 ? "Commands available for \2IRC Operators\2:" :
			level==2 ? "Commands available for \2Services Operators\2:" :
			level==3 ? "Commands available for \2Services Administrators\2:" :
			level>3 ? "Commands available for \2Services Roots\2:" : "",
			u->os_level+1 >= level ? "Access Granted" : "Access Denied");
		notice(s_NickServ, source, " ");

		for (c = commands; c->cmd; c++) {
			if (c->access == level) {
				notice(s_NickServ, source, "    \2%s\2   %s\n", c->cmd, c->desc);
				shown++;
			}
		}
		if (!shown)
			notice(s_NickServ, source, "    No commands available.");

	}
	notice(s_NickServ, source, " ");
	notice(s_NickServ, source, "Note: Most of these commands are logged!");
	notice(s_NickServ, source, "*** END OF HELP ***");
}

/****************************************************************************/

static void help(const char *source)
{
	char *param = gettok(NULL, "", CMDMAX);
	char *s, *topic;
	int level=0;


	if (param) {
		s = sstrdup(param);
		topic = strtok(s, " ");
		if (topic) {
			int i=0;
			if ((i=check_ns_command(topic)) > -1)
				level = commands[i].access;

		}
		free(s);
	}
	if (!AUTH_EMAIL && param && sstrcmp(param, "REGISTER") == 0)
		do_help(s_NickServ, source, "register2", level);
	else
		do_help(s_NickServ, source, param, level);
}

/****************************************************************************/


static void timeout_warn(Timeout *t)
{
#ifndef BAHAMUT
	Timeout *enforcer;
#endif

	if (finduser(t->data) && !is_recognized(t->data, NULL)) {
#ifdef BAHAMUT
		send_cmd(NULL, "SVSHOLD %s %d :This nick is currently being protected for the "
			"owner. If you own this nick, type \"/%s RELEASE %s <password>\" to use it.", 
			t->data, RELEASE_TIMEOUT, s_NickServ, t->data);
		send_cmd(NULL, "SVSNICK %s Guest %d",
			t->data, time(NULL));
#else
		send_cmd(NULL, "SVSNICK %s Guest%d %d",
			t->data, getrandom(10000,99999), time(NULL));
#endif
	}

#ifndef BAHAMUT
	enforcer = add_timeout(RELEASE_TIMEOUT, timeout_release, 0);
	enforcer->data = sstrdup(t->data);
	enforcer->flagged = 1;
#endif

	free(t->data);
}


static void timeout_enforce(Timeout *t)
{
	Timeout *warn;

	if (finduser(t->data) && !is_recognized(t->data, NULL)) {
		warn = add_timeout(30, timeout_warn, 0);
		warn->data = sstrdup(t->data);
		notice(s_NickServ, t->data,
			"If you do not choose another nick within 30 seconds your "
			"nick will be changed.");
	}

	free(t->data);
}

/****************************************************************************/

void validate_user(User *user, int newuser)
{
	NickInfo *ni;
        int num_rows, i;
        char *buf;
	Timeout *t;

        if (ENABLE_R_MODES && (ENABLE_R_MODES != 2))
                if (check_r_mode(user)) {  /* Timestamp verifies already identified */
			user->mode |= U_RECOGNIZED;
                        return;
		}

        if (!(ni = findnick(user->nick))) {
		if (GENERIC_NOTICE && !user->timestamp && newuser)
			notice(s_Global, user->nick, WELCOME_NOTICE,
				NETWORK_NAME, user->nick, NETWORK_NAME, s_NickServ, s_ChanServ);
                return;
	}

	if (ni->flags < 0) {      /* forbidden nick */
		if (is_oper(user->nick) && !FORBID_OPERS) {
			notice(s_NickServ, user->nick,
				"WARNING: You are using a forbidden nick. However, I will allow this "
				"since you are an IRC Operator.");
			return;
		}
		notice(s_NickServ, user->nick,
			"The nick '%s' is forbidden to be used, your nick will now be changed.",
			user->nick);
#ifdef BAHAMUT
		send_cmd(NULL, "SVSHOLD %s 500 :This nick is forbidden", user->nick);
		send_cmd(NULL, "SVSNICK %s Guest %d", user->nick, time(NULL));
#else
		send_cmd(NULL, "SVSNICK %s Guest%d %d",
			user->nick, getrandom(10000,99999), time(NULL));
		send_cmd(NULL, "NICK %s 1 %lu + enforcer %s %s 0 0 :Forbidden nick",
			user->nick, time(NULL), services_host, servername);
		t = add_timeout(500, timeout_release, 0);
		t->data = sstrdup(user->nick);
		t->flagged = 1;
#endif
		return;
	}

	if (!(ni->flags & NS_SECURE)) {   /* No Security, allow recognition */
		buf = smalloc(strlen(user->username) + strlen(user->host) + 2);
		sprintf(buf, "%s@%s", user->username, user->host);

		num_rows = sql_select("mask FROM naccess WHERE id = %d", ni->id);
		if (num_rows) {
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				if (match_wild_nocase(row[0], buf)) {
					user->mode |= U_RECOGNIZED;
					free(buf);
					return;
				}
			}
		}
		free(buf);
	}

	notice(s_NickServ, user->nick,
		"This nick is %sregistered to someone else. Please choose another. "
		"If you are indeed the owner, please type: \2/%s IDENTIFY <your-password>\2.",
		(ni->flags & NS_SECURE) ? "protected and " : "", s_NickServ);


	if (ni->flags & NS_ENFORCE) {
		notice(s_NickServ, user->nick,
			"If you do not choose another within one minute, your nick will be changed.");
		t = add_timeout(30, timeout_enforce, 0);
		t->data = sstrdup(user->nick);
	}
}

/****************************************************************************/

void cancel_user(User *u)
{
	del_ntimeout(u->nick,0);
}


/****************************************************************************/

static void send_reminder(const char *nick, const char *email)
{
	NickInfo *ni = findnick(nick);
	char emsg[512 + NICKMAX];
	char esubj[40];

	if (!ni || ni->remind)
		return;

	ni->remind=1;

	snprintf(esubj, sizeof(esubj),
		"Your registration is about to expire");


	snprintf(emsg, sizeof(emsg),
		"Your nick '%s' on %s will expire in %d days. If you do not log on and identify "
		"to this nick by then, your registration will be lost.",
		nick, NETWORK_NAME, REMINDER_DAYS);

	e_mail(emsg, esubj, email, 1);

}

/****************************************************************************/

int expire_pends()
{
	time_t expire = time(NULL) - (XAUTH_TIME*60*60*24);
	int num_rows=0, i=0;
	char buf[512 + NICKMAX];


	snprintf(buf, sizeof(buf), "SELECT nick.name FROM nick,pending "
		" WHERE pending.time <= %ld AND pending.type = 0 AND pending.id=nick.id", expire);

	if (!mysql_query(&mysql, buf)) {
		if ((res3 = mysql_store_result(&mysql)) == NULL)
			return 0;
		num_rows = mysql_num_rows(res3);
	}


	for (i=0; i < num_rows; i++) {
		row3 = fetch(res3);
		del_ni_nick(row3[0], 0);
		del_sql_nick(row3[0]);
	}
	send_sql("DELETE FROM pending WHERE time <= %d", expire);

	if (num_rows)
		mysql_free_result(res3);

	return num_rows;
}


/****************************************************************************/

void expire_nicks(int xauth)
{
	time_t expire = time(NULL) - (NICK_EXPIRE*60*60*24);
	time_t near_expire = expire + (24*60*60*REMINDER_DAYS);
	unsigned long regs = sql_select("name FROM nick") + xauth;
	unsigned long num_rows=0,i;
	char buf[512];


	send_sql("DELETE from flags WHERE held=0 AND marked=0 AND forbidden=0 AND frozen=0");


	snprintf(buf, sizeof(buf), "SELECT name FROM nick WHERE lasttime <= %ld AND flags < %d AND flags <> -1",
		 expire, HELD_ID);

	if (!mysql_query(&mysql, buf)) {
		if ((res3 = mysql_store_result(&mysql)) != NULL)
			num_rows = mysql_num_rows(res3);
	}

	for (i=0; i < num_rows; i++) {
		row3 = fetch(res3);
		del_ni_nick(row3[0], 0);
		del_sql_nick(row3[0]);
		freeze(row3[0], DEL);
		blog("NS_X", "[NS] X %s", row3[0]);
	}

	send_sql("DELETE from nick WHERE lasttime <= %d AND flags < %d AND flags <> -1", expire, HELD_ID);


	if (num_rows)
		mysql_free_result(res3);


	if (DISPLAY_UPDATES)
		globops(NULL, "Completed nick expire: (%ld/%ld)",
			(num_rows+xauth), regs);


               /* Let's make sure we delete any special flags along with expired nicks */
	num_rows=sql_select("flags.name,nick.name FROM flags LEFT JOIN nick ON flags.name=nick.name");

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		if (row[1] == NULL) {
			send_sql("DELETE FROM flags WHERE name = '%s'", row[0]);
			send_sql("DELETE FROM reason WHERE name = '%s'", row[0]);
		}
	}

	num_rows=sql_select("name,email FROM nick WHERE lasttime <= %d AND flags < %d AND flags <> -1", 
		near_expire, HELD_ID);

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		if (!row[1])
			continue;
		send_reminder(row[0], row[1]);
	}

}

/****************************************************************************/

void do_list(const char *source)
{
	char *param = strtok(NULL, "");
	char *ptr;
	int num_rows=0, i=0, isoper = is_oper(source);
	long x=0;


	if (((ALLOW_LIST==2 && !is_oper(source))) || !ALLOW_LIST) {
		notice(s_NickServ, source,
			"Sorry, the network has disabled the LIST command.");
		return;
	}


	notice(s_NickServ, source, "LIST of nicks (Search paramater = %s)",
		param ? param : "*");


	if (param) {
		if (strlen(param) > NICKMAX) {
			notice(s_NickServ, source, "ERROR: List paramater too long");
			return;
		}
		if (!strchr(param, '@')) {      /* nick search */
			if (strspn(param, "abcdefghijklmnopqrstuvwxyz"
				"ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890\\`_-[]{}^|*%") != strlen(param)) {
				notice(s_NickServ, source,
					"Invalid characters in nick search pattern.");
				return;
			}
		} else if (!check_valid_mask(param)) {     /* host search */
				notice(s_NickServ, source,
					"Invalid host search pattern.");
				return;
		}

		ptr = param;
		while (*ptr) {
			if (*ptr == '*')
				*ptr = '%';
			*ptr++;
		}

	} else
		param = "%";

	num_rows=sql_select("name,usermask,lasttime,flags FROM nick "
		"WHERE %s LIKE '%s' AND flags <> -1 ORDER BY name",
			strchr(param, '@') ? "usermask" : "name", param);


	if (num_rows) {
		unsigned int shown=0, private=0;
		for (i=0; i < num_rows && i < LIST_MAX; i++) {
			row = fetch(res);
			if (atol(row[3]) & NS_PRIVATE) {
				if (!isoper) {	private++; continue;}
				private++;
			}
			shown++;
			x = time(NULL) - atol(row[2]);
			notice(s_NickServ, source,
				"%d) %s   (%s) [%s]", shown, row[0], row[1],
				x<3600 ? "minutes" : 
				x<86400 ? "hours" :
				x<604800 ? "days" :
				x<18144000 ? "weeks" :
				x>=18144000 ? "months" : "");
		}
		notice(s_NickServ, source,
			"Displayed %d of %d match results. (%d private) %s",
			shown, num_rows, private, LIST_MAX<num_rows ?
			"Exceeded maximum result limit, please try refining your search pattern" : "");
	} else {
		notice(s_NickServ, source, "No registered nicks for that match pattern.");
		return;
	}
	notice(s_NickServ, source, "*** End of list ***");

}

/****************************************************************************/

static void do_flags(const char *source, int flag)
{
	NickInfo *ni;
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *reason = strtok(NULL, "");
	char *cmd, *lcmd;
	User *u = finduser(source);


	if (flag==FLAG_HOLD)         { cmd = "HELD"; lcmd = "held"; }
	else if (flag==FLAG_MARK)    { cmd = "MARKED"; lcmd = "marked"; }
	else if (flag==FLAG_FORBID)  { cmd = "FORBIDDEN"; lcmd="forbidden"; }
	else if (flag==FLAG_FREEZE)  { cmd = "FROZEN"; lcmd="frozen"; }
	else { lcmd = cmd = ""; }


	if (!nick || (!reason && ((*nick != '-') && sstrcmp(nick, "*")))) {
		notice(s_NickServ, source,
			"Syntax: \2%s\2 [-]<\37nick\37> <\37reason\37>",
			flag==1 ? "HOLD" : flag==2 ? "MARK" : flag==3 ? "FORBID" :
			flag==4 ? "FREEZE" : "");
		notice(s_NickServ, source,
			"\2/%s HELP %s\2 for more information.", s_NickServ,
			flag==1 ? "HOLD" : flag==2 ? "MARK" : flag==3 ? "FORBID" : 
			flag==4 ? "FREEZE" : "");
	} else if (*nick == '*') {  /* list all nicks */
		int num_rows=0, i, found=0;
		notice(s_NickServ, source, "\2%s nick listing:\2", cmd);

		num_rows=sql_select("flags.name,reason.instater,reason.reason,flags.%s FROM flags,reason "
			"WHERE flags.%s > 0 AND flags.name=reason.name AND reason.type = '%s'",
			lcmd,lcmd,cmd);

		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(s_NickServ, source,
				"%d) %s  [Added by: %s] Reason: \"%s\"   Placed: %s",
					i+1, row[0],row[1],row[2],timebuf(atol(row[3])));
			found++;
		}
		if (!num_rows)
			notice(s_NickServ, source, "No nicks are currently %s.", cmd);
		else
			notice(s_NickServ, source, "\2End of %s list. (%d nicks found)", cmd, found);

	} else if (*nick == '-') {
		*nick++;
		if (!sql_select("name FROM flags WHERE name = '%s' AND %s > 0",
			nick, lcmd)) {
			notice(s_NickServ, source,
				"ERROR: nick '\2%s\2' is not %s!", nick, cmd);
		} else {
/* If all flags are now set to 0 expire_nicks will clean this up on next update. */
			send_sql("UPDATE flags SET %s = 0 WHERE name = '%s'", lcmd, nick);
			send_sql("DELETE from reason WHERE name = '%s' AND type = '%s'",
				nick, cmd);
			if (flag==FLAG_FORBID) {
				send_sql("DELETE FROM nick WHERE name = '%s'", nick);
				send_sql("DELETE FROM reason WHERE name = '%s' AND type = 'FORBIDDEN'", nick);
				send_sql("UPDATE flags SET forbidden = 0 WHERE name = '%s'", nick);
				del_ni_nick(nick, 1);
			} else if (flag==FLAG_HOLD) {
				if ((ni = findnick(nick))) {
					ni->flags &= ~NS_HELD;
					send_sql("UPDATE nick SET flags = %d WHERE id = %d", ni->flags, ni->id);
				}
			} else if (flag==FLAG_FREEZE)
				freeze(nick, DEL);

			blog("NS_F", "[NS] %s- %s (%s!%s@%s)",
				flag==FLAG_HOLD ? "H" : flag==FLAG_MARK ? "M" :
				flag==FLAG_FORBID ? "FB" : flag==FLAG_FREEZE ? "FZ" : "",
				nick, u->os_nick, u->username, u->host);
			notice(s_NickServ, source,
				"Nick '\2%s\2' is no longer %s.", nick, cmd);
			globops(s_NickServ, "%s has removed the %s flag from nick '%s'",
				opertrace(u,0), cmd, nick);

		}

	} else if (strlen(reason) > 255) {
		notice(s_NickServ, source,
			"ERROR: Your reason must not exceed 255 characters in length.");

	} else if (flag == FLAG_FORBID) {
		add_forbid(source, u, nick, reason);


	} else if (nick_flags(nick) < 0) {
		notice(s_NickServ, source,
			"ERROR: Nick \2%s\2' is forbidden!", nick);

	} else if ((flag == FLAG_FREEZE) && is_on_root_list(nick)) {
		notice(s_NickServ, source,
			"ERROR: You may not freeze a service's root nick.");

	} else {
		if (!(ni = findnick(nick))) {
			notice(s_NickServ, source,
				"%s ERROR: Nick '\2%s\2' is not registered!", cmd, nick);

		} else {
			if (sql_select("%s FROM flags WHERE name='%s'", lcmd, nick)) {
				row = fetch(res);
				if (atol(row[0]) > 0) {
					notice(s_NickServ, source,
						"ERROR: Nick '\2%s\2' is already %s!",
						nick, cmd);
					return;
				} else {
					send_sql("UPDATE flags SET %s=%d WHERE name='%s'",
						lcmd,time(NULL),nick);
				}
			} else {
				send_sql("INSERT INTO flags VALUES "
					"('%s','%d','%d','%d','%d')", nick,
					flag==FLAG_HOLD ? time(NULL) : 0,
					flag==FLAG_MARK ? time(NULL) : 0,
					flag==FLAG_FORBID ? time(NULL) : 0,
					flag==FLAG_FREEZE ? time(NULL) : 0);
			}
			if (flag==FLAG_FREEZE)
				freeze(nick, ADD);
			else if (flag==FLAG_HOLD) {
				ni->flags |= NS_HELD;
				send_sql("UPDATE nick SET flags = %d WHERE id = %d", ni->flags, ni->id);
			}

			send_sql("INSERT INTO reason VALUES "
				"('%s','%s','%s','%s')", nick, cmd, reason, 
					opertrace(finduser(source),1));
			blog("NS_F", "[NS] %s+ %s (%s!%s@%s)",
				flag==FLAG_HOLD ? "H" : flag==FLAG_MARK ? "M" :
				flag==FLAG_FORBID ? "FB" : flag==FLAG_FREEZE ? "FZ" : "",
				nick, u->os_nick, u->username, u->host);
			notice(s_NickServ, source,
				"Nick '\2%s\2' added to %s list", nick, cmd);
			globops(s_NickServ, "%s has set the %s flag for nick '%s'.",
				opertrace(u,0), cmd, nick);

		}
	}

}


/****************************************************************************/

void add_forbid(const char *source, User *u, const char *nick, const char *reason)
{
	int del=0;

	if (*nick != '+') {
		if (sql_select("flags FROM nick WHERE name = '%s' AND flags <> -1", nick)) {
			notice(s_NickServ, source,
				"WARNING: nick '\2%s\2' is registered! "
				"If you wish to carry out this action, type: "
				"\2/%s FORBID +%s %s\2", nick, s_NickServ, nick, reason);
			return;
		}
	}

	if (*nick == '+') { *nick++; del=1; }

	if (is_on_root_list(nick)) {
		notice(s_NickServ, source,
			"ERROR: You may not forbid a services root's nick.");
		globops(s_NickServ, "'\2%s\2' tried to forbid the services root nick '%s'",
			source, nick);

	} else if (sql_select("flags FROM nick WHERE name = '%s' AND flags = -1", nick))
		notice(s_NickServ, source,
			"ERROR: nick '\2%s\2 is already forbidden!", nick);
	else {
		NickInfo *ni = scalloc(sizeof(NickInfo), 1);

		if (del) {
			del_ni_nick(nick, 1);
			del_sql_nick(nick);
		}
		ni->nick = sstrdup(nick);
		ni->flags = -1;
		add_ni_nick(ni);

		send_sql("INSERT into nick VALUES "
			"(%d,'%s','*','*','*','*','*',0,0,-1,0,NULL,NULL,0,0,0)",
			generate_primary_id("nick"), nick);
		send_sql("INSERT INTO flags VALUES "
			"('%s',0,0,'%d',0)",
			nick, time(NULL));
		send_sql("INSERT INTO reason VALUES "
			"('%s','FORBIDDEN','%s','%s')", nick, reason, opertrace(finduser(source),1));


		blog("NS_F", "[NS] FB+ %s (%s!%s@%s)", nick,
			u->os_nick, u->username, u->host);
		notice(s_NickServ, source,
			"Nick '\2%s\2' is now forbidden and may not be used.", nick);
		globops(s_NickServ, "%s has forbidden the nick '%s'", opertrace(u,0), nick);

		if (finduser(nick)) {
			notice(s_NickServ, nick, "This nick is now forbidden and will be changed.");
#ifdef BAHAMUT
			send_cmd(NULL, "SVSHOLD %s 60 :This nick is forbidden.",nick);
			send_cmd(NULL, "SVSNICK %s Guest %d", nick, time(NULL));
#else
			send_cmd(NULL, "SVSNICK %s Guest%d %d",
				nick, getrandom(10000,99999), time(NULL));
#endif
		}
	}
}

/****************************************************************************/

static void do_delete(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);

	if (!nick) {
		notice(s_NickServ, source, "Syntax: \2DELETE\2 <nick>");
		notice(s_NickServ, source,
			"\2/%s HELP DELETE\2 for more information.",
				s_NickServ);

	} else if (is_on_root_list(nick)) {
		notice(s_NickServ, source,
			"DELETE ERROR: You may not delete a services root account!");
		globops(s_NickServ, "'\2%s\2' tried to DELETE the services root nick '%s'",
			source, nick);

	} else if (!findnick(nick)) {
		notice(s_NickServ, source,
			"DELETE ERROR: Nick '\2%s\2' is not registered!", nick);

	} else if (is_frozen(nick)) {
		notice(s_NickServ, source,
			"DELETE ERROR: Nick '\2%s\2' is currently frozen!", nick);

	} else {
		User *u = finduser(source);
		del_ni_nick(nick, 1);
		del_sql_nick(nick);
		blog("NS_O", "[NS] D+ %s (%s!%s@%s)",
			nick, u->os_nick, u->username, u->host);
		notice(s_NickServ, source,
			"Nick '\2%s\2' deleted from all database records.", nick);
		globops(s_NickServ, "%s has deleted nick '%s' from all database records.",
			opertrace(u,0), nick);
	}

}

/****************************************************************************/

static void do_flist(const char *source)
{
	char *cmd = strtok(NULL, " ");

	if (!cmd || (sstrcmp(cmd, "HELD") && sstrcmp(cmd, "MARKED") &&
		sstrcmp(cmd, "FORBIDDEN") && sstrcmp(cmd, "FROZEN"))) {
		notice(s_NickServ, source,
			"Syntax: \2FLIST\2 [HELD|MARKED|FORBIDDEN|FROZEN]");
		notice(s_NickServ, source,
			"\2/%s HELP FLIST\2 for more information.", s_NickServ);

	} else {

		int num_rows=0, i, found=0;
		char *ucmd = sstrdup(cmd);

		notice(s_NickServ, source, "\2%s nick listing:\2", strupper(cmd));


		num_rows=sql_select("flags.name,reason.instater,reason.reason,flags.%s FROM flags,reason "
			"WHERE flags.%s > 0 AND flags.name=reason.name AND reason.type = '%s'",
			strlower(cmd),cmd,strupper(ucmd));
		free(ucmd);


		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(s_NickServ, source,
				"%d) %s  [Added by: %s] Reason: \"%s\"   Placed: %s",
					i+1, row[0],row[1],row[2],timebuf(atol(row[3])));
			found++;
		}
		if (!num_rows)
			notice(s_NickServ, source, "No nicks are currently %s.", cmd);
		else
			notice(s_NickServ, source, "\2End of %s list. (%d nicks found)", cmd, found);

	}
}

/****************************************************************************/

static void do_flush(const char *source)
{
	NickInfo *ni;
	const char *nick = gettok(NULL, " ", NICKMAX);
	User *u = finduser(source);

	if (!u)
		return;

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2FLUSH\2 <\37nick\37>");
		notice(s_NickServ, source,
			"\2/%s HELP FLUSH\2 for more information.", s_NickServ);

	} else if (!(ni = findnick(nick))) {
		notice(s_NickServ, source,
			"FLUSH ERROR: Nick '\2%s\2' is not registered!", nick);
	} else {
		send_sql("UPDATE nick SET auth = -1 WHERE id = %d", ni->id);
		send_sql("DELETE FROM pending WHERE id = %d", ni->id);
		notice(s_NickServ, source,
			"Successfully flushed authorization information for nick '\2%s\2'", nick);
		globops(s_NickServ, "%s has flushed authorization information for nick '\2%s\2'",
			opertrace(u,0), nick);
	}

}

/****************************************************************************/

static void do_getemail(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	User *u = finduser(source);

	if (!u)
		return;

	else if (!nick)
		notice(s_NickServ, source,
			"Syntax: \2GETEMAIL\2 <nickname>");

	else if (is_on_root_list(nick)) {
		notice(s_NickServ, source,
			"GETEMAIL error: You may not getemail a services root nick!");
		globops(NULL, "\2%s\2 tried to use GETEMAIL on the services root nick: %s", source, nick);

	} else if (sql_select("instater FROM reason,flags WHERE name='%s' AND MARKED > 0", nick)) {
		row = fetch(res);
		notice(s_NickServ, source,
			"Nick \2%s\2 is currently marked by %s. Unable to GETEMAIL.", row[0]);

	} else if (!sql_select("email,flags FROM nick WHERE id = %d", nick_key(nick)))
		notice(s_NickServ, source,
			"GETEMAIL error: Nick \2%s\2 is not registered!", nick);
	else {
		row = fetch(res);

		globops(NULL, "%s used GETEMAIL on nick: %s", opertrace(u,0), nick);
		notice(s_NickServ, source, "E-mail address for nick %s is: \2%s\2", nick, row[0]);
		blog("NS_S", "[NS] E %s (%s!%s@%s)", nick, source, u->username, u->host);
	}


}

/****************************************************************************/

static void do_listchans(const char *source)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	NickInfo *ni;

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2LISTCHANS\2 <\37nick\37>");
		notice(s_NickServ, source,
			"\2/%s HELP LISTCHANS\2 for more information.", s_NickServ);

	} else if (!(ni = findnick(nick))) {
		notice(s_NickServ, source,
			"LISTCHANS ERROR: Nick '%s' is not registered!", nick);

	} else if (!is_identified(source, nick) && !is_oper(source)) {
		notice(s_NickServ, source,
			"LISTCHANS ERROR: You must be identified to the nick '%s' before using LISTCHANS.", nick);

	} else {
		int num_founder=sql_select("name FROM chan WHERE fid=%d", ni->id);
		int num_rows;
		int i, level=0, nchans=0;

		if (num_founder) {
			notice(s_NickServ, source, "\2LISTCHANS result:\2");
			for (i=0; i < num_founder; i++) {
				row = fetch(res);
				notice(s_NickServ, source,
					" %d) %s -- Access: Founder", ++nchans, row[0]);
			}
		}
		num_rows=sql_select("name FROM chan WHERE successor='%s'", ni->nick);
		if (num_rows) {
			if (!nchans)
				notice(s_NickServ, source, "\2LISTCHANS result:\2");
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				notice(s_NickServ, source,
					" %d) %s -- Access: Successor", ++nchans, row[0]);
			}
		}

		num_rows=sql_select("caccess.*, chan.name FROM caccess,chan WHERE "
			"caccess.entry='%s' AND caccess.chan=chan.id", nick);

		if (!num_rows && !num_founder) {
			notice(s_NickServ, source,
				"Currently the nick '%s' does not have access to any channels.", nick);
			return;
		}
		if (!num_rows) {
			notice(s_NickServ, source, "\2End of LISTCHANS\2");
			return;
		}


		if (!nchans)
			notice(s_NickServ, source, "\2LISTCHANS result:\2");

		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			level = atol(row[2]);
			notice(s_NickServ, source,
				" %d) %s -- Access: %s    Added by: %s %s", ++nchans, row[6],
				level==3 ?  "Auto-Voice" : level==5 ?  "Auto-Op   " :
				level==10 ? "Super-Op  " : level==13 ? "Co-Founder" : "",
				row[3], timebuf(atol(row[4])));
		}
	notice(s_NickServ, source, "\2End of LISTCHANS\2");
	}


}

/****************************************************************************/

static void do_authinfo(const char *source)
{
	NickInfo *ni;
	const char *nick = gettok(NULL, " ", NICKMAX);

	if (!nick) {
		notice(s_NickServ, source,
			"Syntax: \2AUTHINFO\2 <\37nick\37>");
		notice(s_NickServ, source,
			"\2/%s HELP AUTHINFO\2 for more information.", s_NickServ);

	} else if (!(ni = findnick(nick))) {
		notice(s_NickServ, source,
			"AUTHINFO ERROR: Nick '%s' is not registered!", nick);
	} else {
		int num_rows, i;

		globops(s_NickServ, "%s has used AUTHINFO on '%s'",
			opertrace(finduser(source),0), nick);

		notice(s_NickServ, source, "AUTHINFO for nick: %s", nick);

		num_rows=sql_select("* FROM pending WHERE id = %d", ni->id);

		if (!num_rows)
			notice(s_NickServ, source,
				"Nick '%s' has no current authorizations pending.", nick);
		else {
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				notice(s_NickServ, source,
					"Auth Phase: [type: %s]  %s  [code: %s]",
						atoi(row[4])==0 ? "registration authorization" :
						atoi(row[4])==1 ? "E-mail change" :
						atoi(row[4])==2 ? "password change" :
						atoi(row[4])==3 ? "drop registration" :
						atoi(row[4])==4 ? "mailblock change" : "N/A",
					timebuf(atol(row[3])), row[2]);
			}
		}

		if (!HIGH_NICK_SECURITY)
			return;

		if ((num_rows=sql_select("question,answer FROM nick WHERE id = %d", ni->id)))
			row = fetch(res);

		if (!num_rows || row[0] == NULL || row[1] == NULL) {
			notice(s_NickServ, source,
				"Nick '%s' has no personal question & answer on file.", nick);
			return;
		}
		switch(atoi(row[0])) {
				case 1:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_1);
					break;
				case 2:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_2);
					break;
				case 3:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_3);
					break;
				case 4:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_4);
					break;
				case 5:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_5);
					break;
				case 6:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_6);
					break;
				case 7:
					notice(s_NickServ, source, "Personal Question: %s", PERSONAL_7);
					break;
				default:
					notice(s_NickServ, source, "Personal Question: Unknown.");
					break;
		}
		notice(s_NickServ, source, "Answer: %s", row[1]);
	}
}

/****************************************************************************/

static void do_listemail(const char *source)
{
	const char *email = gettok(NULL, " ", 100);

	if (!email) {
		notice(s_NickServ, source,
			"Syntax: \2LISTEMAIL\2 <\37email@host\37>");
		notice(s_NickServ, source,
			"\2/%s HELP LISTEMAIL\2 for more information.", s_NickServ);


	} else {
		char *cstr = sstrdup(email), *ptr = cstr;
		int num_rows, i=0;


		while (*ptr) {
			if (*ptr == '*')
				*ptr = '%';
			*ptr++;
		}

		globops(s_NickServ, "%s has used LISTEMAIL on '%s'",
			opertrace(finduser(source),0), email);

		num_rows=sql_select("name,time FROM nick WHERE email LIKE '%s'", cstr);

		if (!num_rows)
			notice(s_NickServ, source,
				"There are no registrations on file for the E-mail address: %s", email);
		else {
			notice(s_NickServ, source, "\2LISTEMAIL for: %s\2", email);
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				notice(s_NickServ, source,
					"  %d) %-30s  Time Registered: %s",
					i+1, row[0], timebuf(atol(row[1])));
			}
			notice(s_NickServ, source, "\2End of LISTEMAIL\2 (%d match%s)",
				i, i==1 ? "" : "es");
		}
		free(cstr);
	}

}

/****************************************************************************/
/****************************************************************************/
