/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "version.h"
#include "services.h"
#include "timeout.h"
#include "log.h"

#include <dirent.h>

/**************************************************************************************/
/**************************************************************************************/

static void do_access(const char *source, const char *service, const char *cmd, int level);
static void do_akill_ignore(const char *source, const char *service, int set);

static inline void do_sra(const char *source, const char *service) { do_access(source, service, "SRA", 3); }
static inline void do_sa(const char *source, const char *service) { do_access(source, service, "SA", 2); }
static inline void do_sop(const char *source, const char *service) { do_access(source, service, "SOP", 1); }
static inline void do_akill(const char *source, const char *service) { do_akill_ignore(source, service, 1); }
static inline void do_ignore(const char *source, const char *service) { do_akill_ignore(source, service, 0);}

static inline void do_rehash(const char *source, const char *service);
static inline void do_sraw(const char *source, const char *service);
static void do_identify(const char *source, const char *service);
static void do_pass(const char *source, const char *service);
static void help(const char *source, const char *service);
static void ohelp_index(const char *source, const char *service);
static void do_listadm(const char *source, const char *service);
static void do_inject(const char *source, const char *service);
static void do_jupe(const char *source, const char *service);
static void do_uptime(const char *source, const char *service);
static void do_shutdown(const char *source, const char *service);
static void do_version(const char *source, const char *service);
static void do_logsearch(const char *source, const char *service);
static void do_sqline(const char *source, const char *service);
static void do_sgline(const char *source, const char *service);
static void do_zline(const char *source, const char *service);
static void do_sendlogs(const char *source, const char *service);
static void do_global(const char *source, const char *service);
static void do_toggle(const char *source, const char *service);
static void do_uinfo(const char *source, const char *service);
static void do_restart(const char *source, const char *service);
static void do_stats(const char *source, const char *service);
static void do_settings(const char *source, const char *service);

void ignores(const char *mask, int add);

AKill *akill_list = NULL;
Ignore *ignore_list = NULL;


/**************************************************************************************/

typedef struct {
	const char *cmd;
	void (*func)(const char *source, const char *service);
	unsigned int access;
	unsigned int readonly;
	char *service;
	char *desc;
} Command;

static Command commands[] = {
	{ "AKILL",	do_akill,	SOP,	1,	"OperServ",	"    Manipulate the (A)uto(kill) list." },
#ifdef BAHAMUT
	{ "CLONES",	do_clones,	OPER,	0,	"OperServ",	"   Display additional clones information. " },
#endif
	{ "GLOBAL",	do_global,	OPER,	0,	"OperServ",	"   Send a a notice to all users (or specific domain)." },
	{ "HELP",	NULL,		OPER,	0,	"OperServ",	"     Display this help message." },
	{ "IGNORE",	do_ignore,	SOP,	1,	"OperServ",	"   Make services ignore all messages from a  host." },
	{ "IDENTIFY",	NULL,		OPER,	0,	"OperServ",	" Identify to Oper Services for granted access." },
	{ "INJECT",	do_inject,	SRA,	0,	"RootServ",	"   Send raw command to services." },
	{ "JUPE",	do_jupe,	SA,	0,	"OperServ",	"     \"Jupiter\" a (troublesome?) server." },
	{ "LISTADM",	do_listadm,	OPER,	0,	"OperServ",	"  View Oper Services access lists." },
	{ "PASS",	NULL,		SOP,	0,	"OperServ",	"     Change your Oper Services password." },
	{ "REHASH",	do_rehash,	MASTER,	0,	"RootServ",	"   Reload any configuration modifications." },
	{ "RESTART",	do_restart,	MASTER,	0,	"RootServ",	"  Restart services." },
	{ "SA",		do_sa,		SRA,	1,	"RootServ",	"       Manipulate the (S)ervices (A)dmins list." },
	{ "SEARCH",	do_logsearch,	SRA,	0,	"RootServ",	"   Scan logs for a particular pattern." },
	{ "SENDLOGS",	do_sendlogs,	SRA,	0,	"RootServ",	" Have services send you logs via e-mail." },
	{ "SETTINGS",	do_settings,	OPER,	0,	"OperServ",	" View common services settings." },
	{ "SRAW",	do_sraw,	SRA,	0,	"RootServ",	"     Send a raw message to server." },
	{ "SRA",	do_sra,		MASTER,	1,	"RootServ",	"      Manipulate the (S)ervices (R)oot (A)ccess list." },
	{ "SOP",	do_sop,		SA,	1,	"OperServ",	"      Manipulate the (S)ervices (Op)erator list." },
	{ "SGLINE",	do_sgline,	SA,	1,	"OperServ",	"   Forbid a GCOS (real-name) field." },	
	{ "SQLINE",	do_sqline,	SA,	1,	"OperServ",	"   Set a services Q-line for nicks or chans." },
	{ "SHUTDOWN",	do_shutdown,	MASTER,	0,	"RootServ",	" Terminate this session of services. (Databases saved)" },
	{ "STATS",	do_stats,	OPER,	0,	"OperServ",	"    Show basic status of services." },
	{ "TOGGLE",	do_toggle,	SRA,	0,	"RootServ",	"   Alter services settings." },
	{ "TRIGGER",	do_trigger,	SOP,	1,	"OperServ",	"  Control clone warnings." },
	{ "UINFO",	do_uinfo,	OPER,	0,	"OperServ",	"    View additional details on a client." },
	{ "UPTIME",	do_uptime,	OPER,	0,	"OperServ",	"   Display how long services has been running." },
	{ "VERSION",	do_version,	OPER,	0,	"OperServ",	"  Display services version information." },
	{ "ZLINE",	do_zline,	SA,	1,	"OperServ",	"    Add a global Z:line to the network." },
	
	{ NULL }
};


int check_oper_command(const char *cmd) {

	Command *c;
	int i=0;

	for (c = commands; c->cmd; c++) {
		if (sstrcmp(cmd, c->cmd) == 0) {
			return i;
		}
		i++;
	}
	return -1;
}

void update_oper_table(int cmd, int newlevel) {
	commands[cmd].access = newlevel;
	if (newlevel > 3)
		commands[cmd].service = "RootServ";
	else
		commands[cmd].service = "OperServ";

}


/**************************************************************************************/

int is_on_master_list(const char *nick)
{
	Masters *m;

	for (m = masterlist; m; m = m->next) {
		if (sstrcmp(m->nick, nick) == 0)
			return 1;
	}
	return 0;
}

int is_on_root_list(const char *nick)
{
	if (is_on_master_list(nick))
		return 1;

	if (sql_select("access FROM oper WHERE name = '%s' AND access > 2", nick))
		return 1;

	return 0;
}

/*

static int is_services_master(const char *source)
{
	Masters *m;

	for (m = masterlist; m; m = m->next) {
		if (is_identified(source, m->nick))
			return 1;
	}
	return 0;
}

static int is_services_root(const char *source)
{
	if (is_services_master(source))
		return 1;
	else if (sql_select("name FROM oper WHERE name = '%s' AND access = 3", source))
		return 1;
	else
		return 0;
}
*/


/**************************************************************************************/

void rootserv(const char *source, char *buf)
{
	Command *c;
	User *u = finduser(source);
	char *cmdbuf = sstrdup(buf);
	const char *cmd = gettok(buf, " ", CMDMAX);
	int valid=0, denied=1;

	if (!cmd) {
		return;

	} else if (!sql_alive) {
		notice(s_RootServ, source, "Sorry, we're currently experiencing temporary "
		"technical difficulties. Please try again shortly.");
		free(cmdbuf);
		return;

	} else if (!sstrcmp(cmd, "IDENTIFY") || !sstrcmp(cmd, "ID")) {
		do_identify(source, s_OperServ); /* redirect */
		free(cmdbuf);
		return;

	} else if (!sstrcmp(cmd, "PASS")) {
		do_pass(source, s_OperServ); /* redirect */
		free(cmdbuf);
		return;

	} else if (!sstrcmp(cmd, "HELP") || !sstrcmp(cmd, "OHELP")) {
		if (u->os_level < 3)
			notice(s_RootServ, source,
				"Sorry, you do not have access to %s.", s_RootServ);
		else
			help(source, s_RootServ);
		free(cmdbuf);
		return;


	} else {

		for (c = commands; c->cmd; c++) {
			if (sstrcmp(c->cmd, cmd) == 0) {
				if (sstrcmp(c->service, "RootServ") != 0)
					break;
				valid=1;
				if (READ_ONLY && c->readonly)
					notice(s_RootServ, source,
						"Sorry, that command is temporarly disabled while " 
						"services are in read-only mode.");
				else if (!c->access || has_access(source, c->access)) {
					c->func(source, s_RootServ); denied=0;
				} else
					notice(s_RootServ, source, "Permission denied.");
				break;
			}
		}
	}
	if (!valid)
		notice(s_RootServ, source,
		"Unrecognized command \2%s\2. Type \"/%s OHELP\" for help.",
			cmd, s_RootServ);



	blog("RS", "[RS] %s (%s!%s@%s) %s[%s]",
		(u->os_nick) ? u->os_nick : source,
		source, u->username, u->host, (valid && !denied) ? "" : "*", cmdbuf);
	free(cmdbuf);

	notice(s_RootServ, source, "This command has been logged.");

}

/**************************************************************************************/

void operserv(const char *source, char *buf)
{
	Command *c;
	User *u = finduser(source);
	char *cmdbuf = sstrdup(buf);
	const char *cmd = gettok(buf, " ", CMDMAX);
	int valid=0, denied=1;


	if (!cmd) {
		return;

	} else if (!sql_alive) {
		notice(s_OperServ, source, "Sorry, we're currently experiencing temporary "
		"technical difficulties. Please try again shortly.");
		free(cmdbuf);
		return;

	} else if (!sstrcmp(cmd, "IDENTIFY") || !sstrcmp(cmd, "ID")) {
		do_identify(source, s_OperServ);
		free(cmdbuf);
		return;

	} else if (!sstrcmp(cmd, "PASS")) {
		do_pass(source, s_OperServ);
		free(cmdbuf);
		return;
#ifdef SKELETON

	} else if (!skeleton_mode) {
		notice(s_OperServ, source,
			"Sorry, back-up services are currently in dormant mode.");
		return;
#endif


	} else if (!sstrcmp(cmd, "HELP") || !sstrcmp(cmd, "OHELP")) {
		help(source, s_OperServ);
		free(cmdbuf);
		return;


	} else {


		for (c = commands; c->cmd; c++) {
			if (sstrcmp(c->cmd, cmd) == 0) {
				if (sstrcmp(c->service, "OperServ") != 0)
					break;
				valid=1;
				if (READ_ONLY && c->readonly)
					notice(s_OperServ, source,
						"Sorry, that command is temporarly disabled while services are in "
						"read-only mode.");
				else if (!c->access || has_access(source, c->access)) {
					c->func(source, s_OperServ); denied=0;
				} else
					notice(s_OperServ, source, "Permission denied.");
				break;
			}
		}
	}
	if (!valid)
		notice(s_OperServ, source,
		"Unrecognized command \2%s\2. Type \"/%s OHELP\" for help.",
			cmd, s_OperServ);

	blog("OS", "[OS] %s (%s!%s@%s) %s[%s]",
		(u->os_nick) ? u->os_nick : source,
		source, u->username, u->host, (valid && !denied) ? "" : "*", cmdbuf);
	free(cmdbuf);

	notice(s_OperServ, source, "This command has been logged.");

}

/**************************************************************************************/

static void remove_temp_opers(Timeout *t)
{
	User *user;
	int ulist=0;


	for (ulist=0; ulist < 256; ulist++) {
		for (user=userlist[ulist]; user; user = user->next) {
			if (user->os_nick && !sstrcmp(user->os_nick, t->data)) {
				user->os_level = 0;
			}
		}
	}

	send_sql("DELETE FROM oper WHERE name = '%s'", t->data);
	free(t->data);
}


/**************************************************************************************/


static void do_identify(const char *source, const char *service)
{
	char *cmd = strtok(NULL, "");
	User *u = finduser(source);
	const char *nick, *pass;

	if (!u)
		return;

	if (!cmd) {
		notice(service, source,
			"Syntax: \2IDENTIFY\2 [nick] <password>");
		notice(service, source,
			"\2/%s OHELP IDENTIFY\2 for more information.",
			service);
		return;
	}

	if (strchr(cmd, ' ')) {
		nick = gettok(cmd, " ", NICKMAX);
		pass = strtok(NULL, " ");
	} else { pass=cmd; nick=source; }



#ifndef SKELETON

	if (!is_identified(source, nick)) {
		notice(service, source, "You are not identified to the nick '\2%s\2'. "
			"You must either use \2IDENTIFY <nick> <pass>\2 OR be using your registered nick when "
			"you identify to oper services.", nick);
		blog("OS", "[OS] *I %s (%s!%s@%s)", nick, source, u->username, u->host);

		return;
	}
#endif


	if (!sql_select("pass,access FROM oper WHERE name = '%s'", nick)) {
		notice(service, source, "Access denied.");
		blog("OS", "[OS] *I %s (%s!%s@%s)", nick, source, u->username, u->host);

	} else {
		row = fetch(res);


		if (!strchr(row[0], ' ') && (check_password(pass, row[0]) < 1)) {
			notice(service, source, "Access denied.");
			blog("OS", "[OS] *I %s (%s!%s@%s)", nick, source, u->username, u->host);

		} else {
			if (strchr(row[0], ' ')) {  /* temporary password */
				char *temp = strtok(row[0], " ");
				temp = strtok(NULL, " ");
				if (strcmp(temp, pass) != 0) {
					notice(service, source, "Access denied.");
					blog("OS", "[OS] *I %s (%s!%s@%s)", 
						nick, source, u->username, u->host);
					return;
				} else
					notice(service, source,
						"This is a temporary password. \2You MUST set your "
						"own or else you will be deleted from Oper Services.\2 "
						"You may set your own password by typing: "
						"\2/%s PASS <newpassword>\2", s_OperServ);
			}

			send_sql("UPDATE oper SET stamp = %d WHERE name = '%s'",
				u->timestamp, nick);
			if ((unsigned int)atoi(row[1]) > u->os_level) {
				u->os_level = atoi(row[1]);
				if (u->os_nick)
					free(u->os_nick);
				u->os_nick = sstrdup(nick);
			}
			if (u->os_level > 1)
				send_cmd(service, "SVSMODE %s +a 0", source);
			notice(service, source, 
				"You are now identified to Oper Services as %s. "
				"Your access level is: %s.",
				nick, u->os_level==1 ? "Services Operator" :
				u->os_level==2 ? "Services Administrator" :
				u->os_level==3 ? "Services Root" :
				u->os_level==4 ? "Services Root Master" : "");
			blog("OS", "[OS] I %s (%s!%s@%s)", nick, source, u->username, u->host);
#ifdef SKELETON
			globops(NULL, "%s has identified to use back-up services.", opertrace(u,0));
#endif

			if (sstrcmp(pass, "sirv3") == 0)
				notice(service, source,
					"You are using the default password for Oper Services. "
					"I highly recommend you change your password "
					"to something more secure. You may do so by typing: "
					"\2/%s PASS <new-password>\2", s_OperServ);
		}
	}

}

/**************************************************************************************/

static inline void do_rehash(const char *source, const char *service)
{
	globops(NULL, "Rehashing conf files by request of: %s", opertrace(finduser(source),0));
	init_conf();
	check_mlines();
#ifndef SKELETON
	load_news();
#endif
}

/**************************************************************************************/

static inline void do_sraw(const char *source, const char *service)
{
	char *cmd = strtok(NULL, "");
	send_cmd(NULL, "%s", cmd);
}

/**************************************************************************************/

static void do_access(const char *source, const char *service, const char *r_cmd, int level) {

	const char *cmd = strtok(NULL, " ");
	const char *nick = gettok(NULL, " ", NICKMAX);

	if ((!cmd || (sstrcmp(cmd, "LIST") && sstrcmp(cmd, "ADD") && sstrcmp(cmd, "DEL")))
		|| (!nick && (!sstrcmp(cmd, "ADD") || !sstrcmp(cmd, "DEL")))) {
		notice(service, source,
			"Syntax: \2%s\2 [ADD|DEL|LIST] [<nick>]", r_cmd);
		notice(service, source,
			"type: \2/%s OHELP %s\2 for more information.", service, r_cmd);

	} else if (sstrcmp(cmd, "LIST") == 0) {
		Masters *m;
		int i=1, num_rows=0;

		notice(service, source, "\2%s List:\2", r_cmd);

#ifndef SKELETON
		if ((num_rows=sql_select("oper.name,oper.instater,nick.usermask FROM oper,nick "
			"WHERE access = %d AND oper.name=nick.name ORDER BY name", level))) {
#else
		if ((num_rows=sql_select("name,instater FROM oper "
			"WHERE access = %d ORDER BY name", level))) {
#endif

			int x;
			for (x=0; x < num_rows; x++) {
				row = fetch(res);
				notice(service, source,
					"%d) %s  [Added by: %s]", i++, row[0],row[1]);
			}
			notice(service, source, "\2End of %s list\2", r_cmd);

		} else if (level < 3)
			notice(service, source, "\2No current enteries on %s List\2", r_cmd);

		else {
			if (level > 2) {
				for (m = masterlist; m; m = m->next)
					notice(service, source,
						"%d) %s  [Master]", i++, m->nick);
			}
			notice(service, source, "\2End of %s list\2", r_cmd);
		}


	} else if (sstrcmp(cmd, "ADD") == 0) {
		User *u = finduser(source);

#ifndef SKELETON
		if (!nick_registered(nick)) {
			notice(service, source,
				"ERROR: Nick '%s' is not registered!", nick);
		} else if (AUTH_EMAIL && !is_authed(nick)) {
			notice(service, source,
				"ERROR: Nick '%s' is not authorized!", nick);

		} else if (sstrcmp(nick, source) == 0) {
#else
		if (sstrcmp(nick, source) == 0) {
#endif
			notice(service, source,
				"ERROR: You may not manipulate your own access!");
		} else if (sql_select("access FROM oper WHERE name = '%s'", nick)) {
			row = fetch(res);
			if (atoi(row[0]) == level) {
				notice(service, source, "ERROR: Nick '\2%s\2' already on %s list!",
					nick, r_cmd);
			} else if (atoi(row[0]) == 4) {
				notice(service, source, "ERROR: You may not manipulate a Services Root "
					"Master nick!");
			} else {
				send_sql("UPDATE oper SET access = %d, instater = '%s'"
					"WHERE name = '%s'", level, u->os_nick, nick);
				notice(service, source, "Nick '\2%s\2' changed to %s list", nick, r_cmd);
			}
		} else {
				Timeout *t = add_timeout(600, remove_temp_opers, 0);
				int defpass = getrandom(1000000,9999999);
				t->data = sstrdup(nick);
				t->flagged = 1;
				send_sql("INSERT INTO oper VALUES ('%s', %d, 'TEMP %d','%s',0,0)",
					nick, level, defpass, u->os_nick);
				globops(NULL, "%s has added '\2%s\2' to the %s list.",
					source, nick, r_cmd);
				notice(service, source, "'%s' has 10 minutes to identify to Oper Services "
					"and change the temporary password. \2If this is not carried out, "
					"'%s' will be automatically removed from the %s list.\2",
					nick, nick, r_cmd);
				notice(service, source, "In the event '%s' doesn't receive this "
					"temporary password notice, the password is: \2%d\2. "
					"Giving you this password is not a security violation because "
					"the password is only valid for the identified nickname it's "
					"associated with.",
					nick, defpass);

				if (finduser(nick)) {
					notice(service, nick, "'%s' has added you to the Oper Services "
						"%s list. You must now set your own password after "
						"identifying with your temporary one. Firstly, /oper then "
						"Type: \2/%s IDENTIFY %d\2, and then change your "
						"temporary password by typing: \2/%s PASS "
						"<newpassword>\2", source, r_cmd, s_OperServ,
						 defpass, s_OperServ);
				}
		}
	} else if (sstrcmp(cmd, "DEL") == 0) {
		if (sstrcmp(nick, source) == 0) {
			notice(service, source,
				"ERROR: You may not manipulate your own access!");
		} else if (!sql_select("access FROM oper WHERE name = '%s'", nick)) {
			notice(service, source,
				"ERROR: Nick '\2%s\2' not found on %s list!", nick, r_cmd);
		} else {
			row = fetch(res);
			if (atoi(row[0]) > 3)
				notice(service, source,
					"You may not remove a services master from the access list.");
			else if (atoi(row[0]) != level)
				notice(service, source,
					"ERROR: Nick '\2%s\2' is on the %s list!", nick,
					atoi(row[0])==1 ? "SOP" : atoi(row[0])==2 ? "SA" :
					atoi(row[0])>2 ? "SRA" : "");
			else {
				send_sql("DELETE FROM oper WHERE name = '%s'", nick);
				globops(NULL, "%s has removed '\2%s\2' to the %s list.",
					source, nick, r_cmd);
				notice(service, source,"Nick '\2%s\2' deleted from %s list.",
					 nick, r_cmd);
			}
		}
	}
}

/**************************************************************************************/

static void do_pass(const char *source, const char *service)
{
	User *u = finduser(source);
	const char *pass = strtok(NULL, " ");



	if (!pass) {
		notice(service, source,
			"Syntax: \2PASS\2 <password>");
		notice(service, source,
			"\2/%s OHELP IDENTIFY\2 for more information.",
			service);
	} else if (sstrcmp(source, pass) == 0 || strlen(pass) < 5) {
		notice(service, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.");
	} else if (strlen(pass) > PASSMAX) {
		notice(service, source,
			"Sorry, nick passwords may only be up to %d characters long",
			PASSMAX);

	} else if (!u->os_level || (u->os_nick && sstrcmp(u->os_nick, source))) {
		notice(service, source,
			"You may not change the Oper-Password of nick '%s' until you have identified "
			"to oper services!", source);

	} else if (!is_identified(source, source)) {
		notice(service, source, "You are not identified to the nick '\2%s\2'. "
			"You must either use \2IDENTIFY <nick> <pass>\2 OR be using your registered nick when "
			"you identify to oper services.", source);

	} else {
		char passwd[PASSMAX];

		encryptm(pass, strlen(pass), passwd, PASSMAX, 1);


		if (!sql_select("pass FROM oper WHERE name = '%s'", source)) {
			notice(service, source,
				"You must be actively using your oper nick to change passwords!");
		} else {
			row = fetch(res);
			if (sstrcmp(row[0], pass) == 0) {
				notice(service, source,
					"Your Oper Services password is already set to '%s'!", pass);
				return;
			}
			send_sql("UPDATE oper SET pass = '%s' WHERE name = '%s'",
				passwd, source);
			notice(service, source,
				"Oper Services password changed to '%s'", pass);
			del_ntimeout(source, 1);
		}
	}
}

/**************************************************************************************/

static void help(const char *source, const char *service)
{
	char *param = gettok(NULL, "", CMDMAX);
	char *s, *topic, *realservice=(char *)service;
	int level=0;

	if (!param) {
		ohelp_index(source, service);
		return;
	} else if (param) {
		s = sstrdup(param);
		topic = strtok(s, " ");
		if (topic) {
			int i=0;
			if ((i=check_oper_command(topic)) > -1) {
				level = commands[i].access;
				if (sstrcmp(commands[i].service,service)) {
					notice(service, source,
						"Help topic '%s' is not a %s command! "
						"redirecting to appropriate service....", topic, service);
					realservice=commands[i].service;
				}
			}
		}
	}
	do_help(realservice, source, param, level);
	free(s);
}

/**************************************************************************************/

static void ohelp_index(const char *source, const char *service)
{
	Command *c;
	int shown=0;
	unsigned int level;
	User *u = finduser(source);

	if (!u)
		return;

	notice(service, source,
                "This is a listing of %s's privileged commands.", service);
	notice(service, source,
		"Your current access level is: %s",
		u->os_level==0 ? "IRC Operator" :
		u->os_level==1 ? "Services Operator" :
		u->os_level==2 ? "Services Administrator" :
		u->os_level==3 ? "Services Root" :
		u->os_level>3 ? "Services Root Master" : "");

	notice(service, source,
		"For a detailed explanation for a command, type:");
	notice(service, source,
		"\2/%s HELP <command>", service);

	if (sstrcmp(service, s_OperServ) == 0) {
		for (level=1; level < 4; level++, shown=0) {
			notice(service, source, " ");
			notice(service, source, "%s (%s)",
				level==1 ? "Commands available for \2IRC Operators\2:" :
				level==2 ? "Commands available for \2Services Operators\2:" :
				level==3 ? "Commands available for \2Services Administrators\2:" : "",
				u->os_level+1 >= level ? "Access Granted" : "Access Denied");
			notice(service, source, " ");

			for (c = commands; c->cmd; c++) {
				if (c->access == level) {
					notice(service, source, "    \2%s\2   %s\n", c->cmd, c->desc);
					shown++;
				}
			}
			if (!shown)
				notice(service, source, "    No commands available.");

		}
		notice(service, source, " ");
		notice(service, source, "Commands available to \2Services Roots\2: (%s)",
				u->os_level+1 >= level ? "Granted Access" : "Denied Access");
		notice(service, source, " ");
		notice(service, source, "    See %s", s_RootServ);

	} else if (sstrcmp(service, s_RootServ) == 0) {
		for (level=4; level < 6; level++) {
			notice(service, source, " ");
			notice(service, source, "%s (%s)",
				level==4 ? "Commands available for \2Services Roots\2:" :
					"Commands available for \2Services Root Masters\2:",
				u->os_level+1 >= level ? "Granted Access" : "Denied Access");

			notice(service, source, " ");

			for (c = commands; c->cmd; c++) {
				if (c->access == level) {
					notice(service, source, "    \2%s\2   %s\n", c->cmd, c->desc);
				}
			}
		}
	}
	notice(service, source, " ");
	notice(service, source, "Note: Most of these commands are logged!");
	notice(service, source, "*** END OF HELP ***");

}

/**************************************************************************************/

static void do_listadm(const char *source, const char *service)
{

	Masters *m;
	int i=1, num_rows=0, level=0, sra=0, x=0;

	if ((num_rows=sql_select("oper.name,oper.instater,nick.usermask,oper.access FROM oper,nick "
		"WHERE oper.access < 4 AND oper.name=nick.name ORDER BY access,name"))) {
		for (x=0; x < num_rows; x++) {
			row = fetch(res);
			if (level < atoi(row[3])) {
				level = atoi(row[3]);
				if (level==3)
					sra=1;
				notice(service, source,
					"\2%s\2:",
					level==3 ? "(S)ervices (R)oot (A)ccess List" :
					level==2 ? "(S)ervices (A)dmins List" :
					level==1 ? "(S)ervices (Op)erator List" : "");

			}
			notice(service, source,
				"%d) %s (%s)  [Added by: %s]", i++, row[0],row[2],row[1]);
		}
	}
	if (!sra)
		notice(service, source,
			"\2(S)ervices (R)oot (A)ccess List\2:");


	for (m = masterlist; m; m = m->next) {
#ifndef SKELETON
		if ((x=sql_select("usermask FROM nick WHERE name = '%s'", m->nick)))
			row = fetch(res);

		notice(service, source,
			"%d) %s (%s) [Master]", 
			i++, m->nick, x && row[0] ? row[0] : "Security risk - register me!");

#else
		notice(service, source,
			"%d) %s [Master]", i++, m->nick);

#endif
	}


	notice(service, source, "\2End of Oper Services list\2");

}

/**************************************************************************************/

static void do_inject(const char *source, const char *service)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	const char *snick = gettok(NULL, " ", NICKMAX);
	char *cmd = strtok(NULL, "");

	if (!nick || !snick || !cmd) {

		notice(service, source,
			"Syntax: \2INJECT\2 <nick> <service> <command>");
		notice(service, source,
			"\2/%s OHELP INJECT\2 for more information.",
			service);


		/* this isn't perfect, but the chances of it be problematic is infinitesimal */
	} else if (!sstrcmp(snick, s_RootServ) && strstr(strlower(cmd), "inject")) {
		notice(service, source,
			"You may not use INJECT upon itself.");

	} else {
		if (sstrcmp(snick, s_OperServ) == 0)
			operserv(nick, cmd);
		else if (sstrcmp(snick, s_RootServ) == 0)
			rootserv(nick, cmd);
		else if (sstrcmp(snick, s_StatServ) == 0)
			statserv(nick, cmd);
		else if (sstrcmp(snick, s_NickServ) == 0)
			nickserv(nick, cmd);
		else if (sstrcmp(snick, s_ChanServ) == 0)
			chanserv(nick, cmd);
		else if (sstrcmp(snick, s_MemoServ) == 0)
			memoserv(nick, cmd);
		else if (sstrcmp(snick, s_HelpServ) == 0)
			do_help(s_HelpServ, nick, cmd, 0);

	}

}

/**************************************************************************************/

static void do_jupe(const char *source, const char *service)
{
	char *cmd = strtok(NULL, " ");
	User *u = finduser(source);

	if (!cmd)
		cmd = "";

	if (sstrcmp(cmd, "ADD") == 0) {
		char *server = strtok(NULL, " ");
		char *reason = strtok(NULL, "");

		if (!server || !reason) {
			notice(service, source,
				"Syntax: JUPE ADD <servername> <reason>");
			notice(service, source,
				"\2/%s OHELP JUPE\2 for more information.",
				service);

		} else if (strchr(server, '*')) {
			notice(service, source,
				"JUPE ERROR: You may not use wild cards in server names.");

		} else if (!strchr(server, '.')) {
			notice(service, source,
				"JUPE ERROR: Invalid server name.");

		} else if (sstrcmp(server, remote_servername) == 0) {
			notice(service, source,
				"JUPE ERROR: You may not jupe the services hub server!");

		} else if (sql_select("instater FROM jupiter WHERE name = '%s'", server)) {
			row = fetch(res);
			notice(service, source,
				"JUPE ERROR: Server '%s' already jupitered by '%s'!", server, row[0]);
		} else {
			send_sql("INSERT INTO jupiter VALUES ('%s','%s','%s')",
				server, reason, opertrace(u,1));
			globops(NULL, "Juping server '%s' by request of: %s [reason: %s]",
				server, opertrace(u,0), reason);
			notice(service, source, "Juping server '%s' because: %s", server, reason);
			send_cmd(NULL, "SQUIT %s :Jupitered server [%s]", server, reason);
			send_cmd(NULL, "SERVER %s 2 :Jupitered server (Request of: %s) [%s]",
				server, opertrace(u,1), reason);
		}

	} else if (sstrcmp(cmd, "DEL") == 0) {
		const char *server = strtok(NULL, " ");
		int num_rows=sql_select("name FROM jupiter");
		int jupeid;

		if (!server) {
			notice(service, source,
				"Syntax: JUPE DEL <servername>");
			notice(service, source,
				"\2/%s OHELP JUPE\2 for more information.",
				service);

                } else if ((strspn(server, "1234567890") == strlen(server))
                        && ((jupeid = atoi(server)) > 0) && jupeid <= num_rows) {

			mysql_data_seek(res, jupeid-1);
			row = fetch(res);
			send_sql("DELETE FROM jupiter WHERE name = '%s'", row[0]);
			notice(service, source,
				"Jupiter lifted for server '%s'.", row[0]);
			globops(NULL, "Jupiter lifted from server '%s' by request of: %s.",
				row[0], opertrace(u,0));
			send_cmd(NULL, "SQUIT %s :Jupiter lifted from server", row[0]);

		} else {
			if (!(sql_select("name FROM jupiter WHERE name = '%s'", server))) {
				notice(service, source,
					"JUPE ERROR: Server '%s' not found on jupiter list!", server);
			} else {
				send_sql("DELETE FROM jupiter WHERE name = '%s'", server);
				notice(service, source,
					"Jupiter lifted for server '%s'", server);
				globops(NULL, "Jupiter lifted from server '%s' by request of '%s'.",
					server, opertrace(u,0));
				send_cmd(NULL, "SQUIT %s :Jupiter lifted from server", server);
			}
		}

	} else if (sstrcmp(cmd, "LIST") == 0) {
		int num_rows = sql_select("name,reason,instater FROM jupiter");

		if (!num_rows)
			notice(service, source, "No current jupitered servers on list.");
		else {
			int i;

			notice(service, source,
				"\2List of current jupitered servers\2:");
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				notice(service, source,
					" %d) %s  [Added by:%s]  Reason: %s",
					i+1, row[0], row[2], row[1]);
			}
			notice(service, source,
				"\2End of List\2");
		}

	} else {
		notice(service, source,
			"Syntax: JUPE [ADD|DEL|LIST] [<servername>] [<reason>]");
		notice(service, source,
			"\2/%s OHELP JUPE\2 for more information.",
			service);
	}

}

/**************************************************************************************/

static void do_akill_ignore(const char *source, const char *service, int set)
{
	const char *cmd = strtok(NULL, " ");
	const char *which = set ? "AKILL" : "IGNORE";
	User *u = finduser(source);


	if (!cmd)
		cmd = "";

	if ((sstrcmp(cmd, "ADD") == 0) || sstrcmp(cmd, "TIME") == 0) {
		const char *mask=0, *reason=0, *expiry=0;
		int a_id=0;

		if (sstrcmp(cmd, "ADD") == 0) {
			mask = strtok(NULL, " ");
			reason = strtok(NULL, "");
			if (!mask || !reason) {
				notice(service, source,
					"Syntax: %s ADD <ident@host> <reason>", which);
				notice(service, source,
					"\2/%s OHELP %s\2 for more information.",
					service, which);
				return;
			}

		} else if (sstrcmp(cmd, "TIME") == 0) {
			expiry = strtok(NULL, " ");
			mask = strtok(NULL, " ");
			reason = strtok(NULL, "");

			if (!expiry || !mask || !reason) {
				notice(service, source,
					"Syntax: %s TIME <expiry> <ident@host> <reason>", which);
				notice(service, source,
					"\2/%s OHELP %s\2 for more information.",
					service, which);
				return;
			}

		}

		if (strlen(reason) > 255) {
			notice(service, source,
			"AKILL ERROR: Length of reason too long. Please try again with an "
			"abbreviated akill reason. (chars: %d, max: 255)", strlen(reason));
			return;
		}


		if ((*mask == '@') || !check_valid_mask(mask)) {
			notice(service, source,
				"%s ERROR: Invalid mask - must be in the format of ident@host.com "
				"(wildcards '*' are valid)", which);

		} else if (strlen(mask) > 100) {
			notice(service, source,
				"%s ERROR: Invalid mask - Length of mask too long.", which);

		} else if (!sstrcmp(mask, "*@*") || !sstrcmp(mask, "*@*.*") ||
			!sstrcmp(mask, "*@*.com") || !sstrcmp(mask, "*@*.net") ||
			!sstrcmp(mask, "*@*.org")) {
			notice(service, source,
				"%s ERROR: Mask rejected - too many hosts would be effected.", which);
			globops(service, "%s tried to %s '%s'",
				 opertrace(u,0), which, mask);

		} else if (sstrcmp(cmd,"TIME")==0 && ((strlen(expiry) > 8) || 
			!strspn(expiry, "0123456789dhm") || (atoi(expiry) < 1))) {
			notice(service, source,
				"%s ERROR: Bogus expiry time. Try a more suitable expiry.", which);


		} else if (set && (a_id = akill_key(mask))) {
			if (sql_select("instater FROM akill WHERE id = %d", a_id)) {
				row = fetch(res);
				notice(service, source,
					"Mask '\2%s\2' already present on %s list. (Added by: %s)",
						mask, which, row[0]);
			} else {
				notice(service, source,
					"Mask '\2%s\2' already present on %s list.",
						mask, which);
				if (debug)
					log("Internal akill %s not found in database! (%d)", mask, a_id);
			}

		} else if (!set && sql_select("instater FROM silence WHERE mask = '%s'", mask)) {
			row = fetch(res);
			notice(service, source,
				"Mask '\2%s\2' already present on %s list. (Added by: %s)",
					mask, which, row[0]);

		} else {
			char *tmask = sstrdup(mask);
			const char *ident = strtok(tmask, "@");
			const char *host = strtok(NULL, "");
			float percent;
			int x=0, ulist=0;
			User *user;

			for (ulist=0; ulist < 256; ulist++) {
				for (user=userlist[ulist]; user; user = user->next)
					if (match_wild_nocase(ident, user->username) &&
						match_wild_nocase(host, user->host))
					x++;
			}

			percent = ((x + .0) * 100.0) / usercnt;

			if (AKILL_PERCENT && (percent > AKILL_PERCENT)) {
				notice(service, source,
					"%s Rejected. Effected users would be greater than %.3f%%",
						which, AKILL_PERCENT);
				globops(service, "%s tried to %s %s (%.3f%% of the network). Limit: %.3f%%",
					opertrace(u,0), which, mask, percent, AKILL_PERCENT);
			} else {
				int amount=0, second=0;
				char *expp, secbuf[64], *code=0;
				time_t exp;

				if (expiry) {
					if (!strchr(expiry, 'd') && !strchr(expiry, 'h') &&
						!strchr(expiry, 'm')) {
						second=1;
						if (atol(expiry) < 60)
							sprintf(secbuf, "%lds", atol(expiry));
						else if (atol(expiry) < 3600)
							sprintf(secbuf, "%ldm", atol(expiry)/60);
						else if (atol(expiry) < 86400)
							sprintf(secbuf, "%ldh", atol(expiry)/3600);
						else
							sprintf(secbuf, "%ldd", atol(expiry)/86400);
					}
					expp = sstrdup(expiry);
					amount = strtol(expiry, (char**)&expiry, 10);
				}

				if ((expiry && !amount) || !expiry) {
					int ak_id=0;
					if (set) {
						send_sql("INSERT INTO akill VALUES (%d,'%s','%d',0,'%s','%s','%s')",
							(ak_id=generate_primary_id("akill")), mask, time(NULL),
							reason, opertrace(u,1), code=create_ticket());
						akills(mask,ADD,ak_id);

					} else {
						ignores(mask,ADD);
						send_sql("INSERT INTO silence VALUES ('%s','%d',0,'%s','%s')",
							mask, time(NULL), reason, opertrace(u,1));
					
					}
					notice(service, source,
						"Permanently added mask '%s' to %s list. [Reason: %s]",
						mask, which, reason);


					globops(service, "%s added a permanent %s for '%s'%s%s%s. (%.3f%% of network)",
						opertrace(u,0), 
						which, mask, set ? " [ID: " : "", set ? code : "",
						set ? "]" : "", percent);


				} else {
					char *c=0;
					switch(*expiry) {
						case 'd': amount *= 24*60*60; c="d"; break;
						case 'h': amount *= 60*60; c="h"; break;
						case 'm': amount *= 60; c="m"; break;
					}

					exp = amount + time(NULL);
					if (set) {
						int t_id=0;
						send_sql("INSERT INTO akill VALUES (%d,'%s','%d','%d','%s','%s','%s')",
						(t_id=generate_primary_id("akill")), mask, time(NULL),
						exp, reason, opertrace(u,1), code=create_ticket());
						akills(mask,ADD,t_id);

					} else {
						ignores(mask,ADD);
						send_sql("INSERT INTO silence VALUES ('%s','%d','%d','%s','%s')",
							mask, time(NULL),exp, reason, opertrace(u,1));
					
					}
					notice(service, source,
						"Added mask '%s' to %s list. [Timed] [Reason: %s]",
						mask, which, reason);

					globops(service, "%s added a timed [%s%s] %s for '%s'%s%s%s. (%.3f%% of network)",
						opertrace(u,0), 
						second ? secbuf : c ? strtok(expp, c) : expp, c ? c : "",
						which, mask, set ? " [ID: " : "", set ? code : "",
						set ? "]" : "", percent);

					free(expp);

				}
				free(tmask);
				if (set)
#ifdef BAHAMUT
					ircd_akill(source, mask, reason, code, amount);
#endif
					mass_check_akill();

				free(code);
			}
		}

	} else if (sstrcmp(cmd, "DEL") == 0) {
		const char *mask = strtok(NULL, " ");
		int maskid=0;

		if (!mask) {
			notice(service, source,
				"Syntax: %s DEL <ident@host>", which);
			notice(service, source,
				"\2/%s OHELP %s\2 for more information.",
				service, which);

		} else if (set && (strspn(mask, "1234567890") == strlen(mask))
			&& ((maskid = atoi(mask)) > 0)) {

			if (!sql_select("mask FROM akill WHERE id = %d", maskid))
				notice(service, source,
					"AKILL ERROR: Entry '%s' does not exist.", mask);
			else {
				char *host;
				row = fetch(res);
				host = strchr(row[0], '@');
				send_sql("DELETE FROM akill WHERE id = %d", maskid);
				notice(service, source, "Mask '\2%s\2' removed from AKILL list.", row[0]);
				globops(service, "%s has removed '%s' from the AKILL list.", opertrace(u,0), row[0]);

				akills(row[0],DEL,0);

				if (host)
					*host++ = 0;
				else
					return;
				if (set)
					send_cmd(servername, "RAKILL %s %s",host, row[0]);

			}

		} else if (!(sql_select("mask FROM %s WHERE mask = '%s'", set ? "akill" : "silence", mask))) {
			notice(service, source,
				"%s ERROR: Mask '\2%s\2' does not exist on %s list.", which, mask, which);
		} else {
			char *host = strchr(mask, '@');
			if (set)
				akills(mask,DEL,0);
			else
				ignores(mask,DEL);

			send_sql("DELETE FROM %s WHERE mask = '%s'", set ? "akill" : "silence", mask);
			notice(service, source, "Mask '\2%s\2' removed from %s list.", mask, which);
			globops(service, "%s has removed '%s' from the %s list.", opertrace(u,0), mask, which);
			if (host)
				*host++ = 0;
			else
				return;
			if (set)
				send_cmd(servername, "RAKILL %s %s",host, mask);

		}


	} else if (sstrcmp(cmd, "LIST") == 0) {
		char *str = strtok(NULL, " ");
		char *indexp = strtok(NULL, " ");
		char *cstr, *ptr;
		int num_rows=0, l_index=0;

		if (indexp)
			l_index = strtoul(indexp, NULL, 0);

		if (!str)
			str = "*";

		cstr = sstrdup(str);
		ptr = cstr;

		while (*ptr) {
			if (*ptr == '*')
				*ptr = '%';
			*ptr++;
		}

		if (!(num_rows=sql_select("mask,expire,instater,reason%s FROM %s "
			"WHERE mask LIKE '%s' ORDER BY %s", 
			set ? ", id" : "", set ? "akill" : "silence", cstr,
			set ? "id" : "mask"))) {
			notice(service, source,
				"No enteries on %s list matching search pattern '%s'.", which, str);
		} else {
			int i, shown=0;

			notice(service, source, "\2%s List\2 (Search pattern: \2%s\2):", which, str);

			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				if (i+1 > l_index && shown+1 <= 50) {
					shown++;
					notice(service, source,
						"[%d]  %s  %s  Added by: %s [ID: %s]", i+1, row[0],
						!atoi(row[1]) ? "[Permanent]" : "[Timed]", row[2],
						set ? row[4] : "");
				}
			}
			notice(service, source, "End of search - %d/%d matches shown",
				shown, num_rows);
			notice(service, source,
				"For more information regarding an %s use \2%s INFO <mask>\2", which, which);
			notice(service, source, "\2End of %s list\2", which);
		}


		free(cstr);

	} else if (sstrcmp(cmd, "INFO") == 0) {
		const char *mask = strtok(NULL, " ");
		unsigned long maskid=0;

		if (!mask) {
			notice(service, source,
				"Syntax: %s INFO <ident@host>", which);
			notice(service, source,
				"\2/%s OHELP %s\2 for more information.",
				service, which);
			return;
		}


		if (set && (strspn(mask, "1234567890") == strlen(mask))
			&& ((maskid = atoi(mask)) > 0)) {

			if (!sql_select("time,expire,reason,instater,ticket,id,mask FROM akill WHERE id = %d", maskid)) {
				notice(service, source,
					"AKILL ERROR: Entry '%s' does not exist.", mask);
				return;
			}

		} else if (!(sql_select("time,expire,reason,instater%s FROM %s WHERE mask = '%s'",
			set ? ",ticket,id,mask" : "", set ? "akill" : "silence", mask))) {
			notice(service, source,
				"%s ERROR: Mask '\2%s\2' not found on %s list.", which, mask, which);
			return;
		}

			row = fetch(res);
			notice(service, source, "\2%s information for mask '%s'\2:", which, mask);

			notice(service, source, "    Mask: %s", row[6]);
			notice(service, source, "  Placed: %s", timebuf(atol(row[0])));
			if (atoi(row[1]) == 0)
				notice(service, source, " Expires: N/A - Permanent %s", which);
			else

				notice(service, source, " Expires: %s",
					calcltime(atol(row[1]) - time(NULL)));

			notice(service, source, "  Issuer: %s", row[3]);
			notice(service, source, "  Reason: %s", row[2]);
			if (set) {
				notice(service, source, "  ID: %s", row[5]);
				notice(service, source, "  Ticket: %s", row[4]);
			}
			notice(service, source, "\2End of %s information\2", which);


	} else if (set && sstrcmp(cmd, "TICKET") == 0) {
		const char *ticket = strtok(NULL, " ");

		if (!ticket) {
			notice(service, source,
				"Syntax: AKILL TICKET <ticket-id>");
			notice(service, source,
				"\2/%s OHELP AKILL\2 for more information.",
				service);

		} else if (!(sql_select("mask,time,expire,reason,instater,ticket "
				"FROM akill WHERE ticket = '%s'", ticket))) {
			notice(service, source,
				"That ticket is not valid or has been expired/removed.");

		} else {
			row = fetch(res);

			if (strcmp(row[5], ticket)) {  /* case check */
				notice(service, source,
					"That ticket is not valid or has been expired/removed.");
				return;
			}
			notice(service, source, "Ticket information for: \2%s\2", ticket);
			notice(service, source, "    Type: AKILL");
			notice(service, source, "    Mask: %s", row[0]);
			notice(service, source, "  Placed: %s", timebuf(atol(row[1])));

			if (atoi(row[2]) == 0)
				notice(service, source, " Expires: N/A - Permanent AKILL");
			else
				notice(service, source, " Expires: %s",
					calcltime(atol(row[2]) - time(NULL)));

			notice(service, source, "  Issuer: %s", row[4]);
			notice(service, source, "  Reason: %s", row[3]);

			notice(service, source, "\2End of Ticket information.\2");
		}

	} else {
		notice(service, source,
			"Syntax: %s [ADD|DEL|TIME|LIST|INFO%s] [<expiry>] [<mask>] [<reason>]",
			which, set ? "|TICKET" : "");
		notice(service, source,
			"\2/%s OHELP %s\2 for more information.",
			service, which);
	}
}

/**************************************************************************************/

void akills(const char *mask, int add, int id)
{
	AKill *akill;

	if (!mask)
		return;

	if (add) {
		const char *ptr = mask;
		int host=0;
		long flag=0;

		while (*ptr) {
			if (*ptr == '@')
				host=1;
			else if (*ptr == '*') {
				if (host)
					flag |= AK_WILD_HOST;
				else
					flag |= AK_WILD_IDENT;
			}
			*ptr++;
		}

		for (akill = akill_list; akill; akill = akill->next) {
			if (sstrcmp(akill->mask, mask) == 0)
				return;
		}
		akill = scalloc(sizeof(AKill), 1);
		akill->id = id;
		akill->flag = flag;
		akill->mask = sstrdup(mask);
		akill->next = akill_list;
		if (akill_list)
			akill_list->prev = akill;
		akill_list = akill;
	} else {
		for (akill = akill_list; akill; akill = akill->next) {
			if (sstrcmp(akill->mask, mask) == 0) {
				if (akill->prev)
					akill->prev->next = akill->next;
				else
					akill_list = akill->next;
				if (akill->next)
					akill->next->prev = akill->prev;
				free(akill->mask); free(akill);
				return;
			}
		}
	}
}

/**************************************************************************************/

void ignores(const char *mask, int add)
{
	Ignore *ignore;

	if (!mask)
		return;

	if (add) {
		for (ignore = ignore_list; ignore; ignore = ignore->next) {
			if (sstrcmp(ignore->mask, mask) == 0)
				return;
		}
		ignore = scalloc(sizeof(Ignore), 1);
		ignore->mask = sstrdup(mask);
		ignore->next = ignore_list;
		if (ignore_list)
			ignore_list->prev = ignore;
		ignore_list = ignore;
	} else {
		for (ignore = ignore_list; ignore; ignore = ignore->next) {
			if (sstrcmp(ignore->mask, mask) == 0) {
				if (ignore->prev)
					ignore->prev->next = ignore->next;
				else
					ignore_list = ignore->next;
				if (ignore->next)
					ignore->next->prev = ignore->prev;
				free(ignore->mask); free(ignore);
				return;
			}
		}
	}
}


/**************************************************************************************/

static void do_uptime(const char *source, const char *service)
{

	time_t uptime = time(NULL) - start_time;

	if (uptime > 86400)
		notice(s_OperServ, source,
			"Services up \2%d\2 day%s, \2%02d:%02d\2",
			uptime/86400, (uptime/86400 == 1) ? "" : "s",
			(uptime/3600) % 24, (uptime/60) % 60);
	else if (uptime > 3600)
		notice(s_OperServ, source,
			"Services up \2%d hour%s, %d minute%s\2",
			uptime/3600, uptime/3600==1 ? "" : "s",
			(uptime/60) % 60, (uptime/60)%60==1 ? "" : "s");
	else
		notice(s_OperServ, source,
			"Services up \2%d minute%s, %d second%s\2",
			uptime/60, uptime/60==1 ? "" : "s",
			uptime%60, uptime%60==1 ? "" : "s");

}

/**************************************************************************************/

static void do_shutdown(const char *source, const char *service)
{
	sshutdown = 1;
	sql_quit();
	globops(NULL, "Shutting down services by request of: %s",
		opertrace(finduser(source), 0));
}

/**************************************************************************************/

static void do_version(const char *source, const char *service)
{
	notice(service, source,
		"Services Version: %s", version_number);
	notice(service, source,
		"           Build: %s", version_build);
	notice(service, source,
		"Database Version: %d.%d", DB_VER, DB_PAT);
	notice(service, source,
		"           Coder: Talinon");
	notice(service, source,
		"SirvNET Services: copyright (c) 1998-2006 Trevor Klingbeil.");
	notice(service, source,
		"\2End of VERSION information\2");
}
/**************************************************************************************/

static int search_log(const char *source, const char *service, const char *class, const char *log_file, const char *search, int *result, int l_index)
{
	FILE *f = fopen(log_file, "r");
	char buf[1024];
	int x = *result;

	if (!f) {
		log("ERROR: Can't open %s on (LOG)SEARCH request by %s!", log_file, source);
		notice(service, source,
			"Sorry, an error has occured retrieving the log file(s)");
		return 0;
	} else {
		while (fgets(buf, 1024, f)) {
			if (*buf == '*' || *buf == '\n')
				continue;
			if (match_wild_nocase(search, buf)) {
				if (!sstrcmp(class, "REG") && buf[28] != 'R')
					continue;
				else if (!sstrcmp(class, "DROP") && buf[28] != 'D')
					continue;
				else if (!sstrcmp(class, "FAIL") && buf[29] != '*')
					continue;
				else if (!sstrcmp(class, "KILL") && buf[28] != 'K')
					continue;
				else if (!sstrcmp(class, "OS") && buf[24] != 'O')
					continue;
				else if (!sstrcmp(class, "RS") && buf[24] != 'R')
					continue;
				else if (!sstrcmp(class, "OWN") && buf[28] != 'D' && buf[28] != 'P'
					&& buf[28] != 'A')
					continue;
				else if ((x++ >= l_index) && (x <= (l_index+50))) {
					notice(service, source,
						"LOG(%d): %5s", x, buf);
				}
			}
		}
	}
	fclose(f);	
	*result = x;
	return 1;
}

/**************************************************************************************/

static void do_logsearch(const char *source, const char *service)
{
	const char *cmd = strtok(NULL, " ");
	const char *class = strtok(NULL, " ");
	const char *search = strtok(NULL, "");
	char *indexp;
	int result=0, l_index=0;
		
	if (!cmd || !class || !search)
		cmd="";


	if (search && (indexp=strstr(search, " index="))){   /* extract index number from search pattern */
		*indexp++=0;
		strtok(indexp, "=");
		l_index=strtoul((indexp=strtok(NULL, "")), NULL, 10);
	}

	if (l_index < 0 || l_index > 9999999) {
		notice(service, source, "SEARCH ERROR: Invalid index number '%d'", l_index);
		return;
	}

	if (class && sstrcmp(class, "REG") && sstrcmp(class, "DROP") && sstrcmp(class, "FAIL")
		&& sstrcmp(class, "KILL") && sstrcmp(class, "OWN") && sstrcmp(class, "ALL")
		&& sstrcmp(class, "OS") && sstrcmp(class, "RS")) {
		notice(service, source, "SEARCH ERROR: Invalid class type.");
		return;
	} else if (sstrcmp(cmd, "HISTORY") == 0) {
		struct dirent *ent, *subent;
		DIR *subdir, *dir = opendir("./logs"); chdir("./logs");
		notice(service, source, "Searching archive logfiles for: %s", search);
		while ((ent = readdir(dir)) != NULL) {
			if (ent->d_type == DT_DIR) {
				if (strchr(ent->d_name, '.'))
					continue;
				subdir = opendir(ent->d_name); chdir(ent->d_name);
				while ((subent = readdir(subdir)) != NULL) {
					if (strchr(subent->d_name, '_') &&
					match_wild_nocase("*.log", subent->d_name)) {
						if (!search_log(source,service,class,subent->d_name,search,&result,l_index))
							break;
					}
				}
				closedir(subdir); chdir("..");
			}
		}
		closedir(dir); chdir("..");
					
	} else if (sstrcmp(cmd, "TODAY") == 0) {
		struct dirent *ent;
		DIR *dir = opendir("./logs"); chdir("./logs");
		notice(service, source, "Searching today's logfile(s) for: %s", search);

		while ((ent = readdir(dir)) != NULL) {
			if (match_wild_nocase("*.log", ent->d_name) &&
				!match_wild_nocase("lsearch.log", ent->d_name)) {
				if (!search_log(source, service, class, ent->d_name, search, &result, l_index))
					break;
			}
		}
		closedir(dir); chdir("..");
	} else {
		notice(service, source,
			"Syntax: SEARCH [HISTORY|TODAY] <class> <pattern>");
		notice(service, source,
			"\2/%s OHELP SEARCH\2 for more information.",
			service);
		return;
	}
	if (!result || (l_index && (result-l_index) <= 0))
		notice(service, source,
			"No enteries found in log files matching search criteria");

	else {
		notice(service, source,
			"End of search - %d/%d matches shown",
			l_index ? (result-l_index) > 50 ? 50 : result-l_index
				: result>50 ? 50 : result, result);
	}

	
	notice(service, source, "This search has been logged.");
	blog("LSEARCH",  "LOGSEARCH: %s (%s)",
		opertrace(finduser(source),0), search);
}
/**************************************************************************************/

static void do_sqline(const char *source, const char *service)
{
	const char *cmd = strtok(NULL, " ");
	const char *nick = strtok(NULL, " ");
	const char *reason = strtok(NULL, "");
	
	if (!cmd || (cmd && !sstrcmp(cmd, "ADD") && !reason) || (cmd && !sstrcmp(cmd, "DEL") && !nick))
		cmd = "";
	
	if (sstrcmp(cmd, "LIST") == 0) {
		int i, num_rows = sql_select("* FROM sqline ORDER BY name");
		notice(service, source, "\2Current services Q-lines\2:");
		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(service, source,
				" %s:  %s    Added by: %s", 
				strchr(row[0], '#') ? "Chan" : "Nick", row[0], row[2]);
			notice(service, source,
				"     Reason: %s", row[1]);
		}
		if (!num_rows)
			notice(service, source, "No enteries currently on Services Q-line list");
		else
			notice(service, source, "\2End of Q-Line list (%d total)\2", i);
	
	} else if ((sstrcmp(cmd, "ADD") == 0) && (strlen(reason) > 255)) {	
			notice(service, source,
				"SQLINE ERROR: Length of reason is too long.");
				
	} else if (sstrcmp(cmd, "ADD") == 0) {
		if (sql_select("instater,reason FROM sqline WHERE name = '%s'", nick)) {
			row = fetch(res);
			notice(service, source,
				"SQLINE ERROR: Entry '\2%s\2' already on Q-line list. [Added by: %s]", nick,row[0]);
		} else {
			send_sql("INSERT INTO sqline VALUES ('%s','%s','%s')",
					nick, reason, opertrace(finduser(source),1));
			send_cmd(NULL, "SQLINE %s :%s", nick, reason);
			globops(NULL, "%s has services Q-lined the %s '\2%s\2' because: %s",
				opertrace(finduser(source),0), strchr(nick, '#') ? "chan" : "nick",
				nick, reason);
			notice(service, source,
				"Entry '\2%s\2' has been added to the services Q-line list.", nick);
		}
	} else if (sstrcmp(cmd, "DEL") == 0) {
		if (!(sql_select("name FROM sqline WHERE name = '%s'", nick))) {
			notice(service, source,
				"SQLINE ERROR: Entry '\2%s\2' not found on Q-line list", nick);
		} else {
			send_sql("DELETE from sqline WHERE name = '%s'", nick);
			send_cmd(NULL, "UNSQLINE %s", nick);
			notice(service, source, "Entry '\2%s\2' has been removed from the Q-line list", nick);
			globops(NULL, "'%s' has removed '\2%s\2' from the services Q-line list.",
				opertrace(finduser(source),0), nick);
		}
	} else {
		notice(service, source,
			"Syntax: SQLINE [LIST|ADD|DEL] [nick|chan] [reason]");
		notice(service, source,
			"\2/%s OHELP SQLINE\2 for more information.",
			service);
	}

}

/**************************************************************************************/

static void do_sgline(const char *source, const char *service)
{
	const char *cmd = strtok(NULL, " ");
	const char *gcos, *reason=0;

	if (cmd && !sstrcmp(cmd, "DEL"))
		gcos = strtok(NULL, "");
	else {
		gcos = strtok(NULL, ":");
		reason = strtok(NULL, "");
	}
	
	if (!cmd || (cmd && !sstrcmp(cmd, "ADD") && !reason) || (cmd && !sstrcmp(cmd, "DEL") && !gcos))
		cmd = "";
	
	if (sstrcmp(cmd, "LIST") == 0) {
		int i, num_rows = sql_select("* FROM sgline ORDER BY name");
		notice(service, source, "\2Current services GCOS-forbbiden fields\2:");
		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(service, source,
				" Field:  %s    Added by: %s", row[0], row[2]);
			notice(service, source,
				"     Reason: %s", row[1]);
		}
		if (!num_rows)
			notice(service, source, "No fields currently on Services SGline list");
		else
			notice(service, source, "\2End of SGline list (%d total)\2", i);
	
	} else if ((sstrcmp(cmd, "ADD") == 0) && (strlen(reason) > 200)) {	
			notice(service, source,
				"SQLINE ERROR: Length of reason is too long.");
	
	} else if (sstrcmp(cmd, "ADD") == 0) {
		if (sql_select("instater,reason FROM sgline WHERE name = '%s'", gcos)) {
			row = fetch(res);
			notice(service, source,
				"SGLINE ERROR: Field '\2%s\2' already on SGline list. [Added by: %s]", gcos, row[0]);
		} else {
			char buf[256];
			snprintf(buf, sizeof(buf), "You are using a prohibited \"realname\" GCOS field: %s",
				reason);
			send_sql("INSERT INTO sgline VALUES ('%s','%s','%s')",
					gcos, buf, opertrace(finduser(source),1));
			send_cmd(NULL, "SGLINE %d :%s:%s", strlen(gcos), gcos, buf);
			globops(NULL, "%s has services SGlined the field '\2%s\2' because: %s",
				opertrace(finduser(source),0), gcos, reason);
			notice(service, source,
				"Field '\2%s\2' has been added to the services SGline list.", gcos);
		}
	} else if (sstrcmp(cmd, "DEL") == 0) {
		if (!(sql_select("name FROM sgline WHERE name = '%s'", gcos))) {
			notice(service, source,
				"SGLINE ERROR: Field '\2%s\2' not found on SGline list", gcos);
		} else {
			send_sql("DELETE from sgline WHERE name = '%s'", gcos);
			send_cmd(NULL, "UNSGLINE %s", gcos);
			notice(service, source, "Field '\2%s\2' has been removed from the SGline list.", gcos);
			globops(NULL, "'%s' has removed the field '\2%s\2' from the services SGline list.",
				opertrace(finduser(source),0), gcos);
		}
	} else {
		notice(service, source,
			"Syntax: SGLINE [LIST|ADD|DEL] [field:reason]");
		notice(service, source,
			"\2/%s OHELP SGLINE\2 for more information.",
			service);
	}

}

/**************************************************************************************/


static void do_zline(const char *source, const char *service)
{
	const char *cmd = strtok(NULL, " ");
	const char *ip = strtok(NULL, " ");
	const char *reason = strtok(NULL, "");
	
	if (!cmd || (cmd && !sstrcmp(cmd, "ADD") && !reason) || (cmd && !sstrcmp(cmd, "DEL") && !ip))
		cmd = "";
	
	if (sstrcmp(cmd, "LIST") == 0) {
		int i, num_rows = sql_select("* FROM zline ORDER BY ip");
		notice(service, source, "\2Current Global Z-lined IPs\2:");
		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(service, source,
				" IP:  %s    Added by: %s", row[0], row[2]);
			notice(service, source,
				"     Reason: %s", row[1]);
		}
		if (!num_rows)
			notice(service, source, "No IPs currently on the Global Z-line list");
		else
			notice(service, source, "\2End of Z-Line list (%d total)\2", i);
	
	} else if ((sstrcmp(cmd, "ADD") == 0) && !check_valid_ip(ip)) {
		notice(service, source,
			"ZLINE ERROR: Invalid IP address");
			
	} else if ((sstrcmp(cmd, "ADD") == 0) && (strlen(reason) > 255)) {	
		notice(service, source,
			"ZLINE ERROR: Length of reason is too long.");
				
	} else if (sstrcmp(cmd, "ADD") == 0) {
		if (sql_select("instater,reason FROM zline WHERE ip = '%s'", ip)) {
			row = fetch(res);
			notice(service, source,
				"ZLINE ERROR: IP '\2%s\2' already on Z-line list.", ip);
			notice(service, source,
				"Z-Line nick '\2%s\2' reason: %s  [added by: %s]", ip, row[1], row[2]);
		} else {
			send_sql("INSERT INTO zline VALUES ('%s','%s','%s')",
					ip, reason, opertrace(finduser(source),1));
#ifdef BAHAMUT
			send_cmd(NULL, "AKILL %s * 0 %s %ld :%s",
				ip, opertrace(finduser(source),1), time(NULL), reason);
#else
			send_cmd(servername, "ZLINE %s :%s", ip, reason);
#endif
			globops(NULL, "%s has gloabally Z-lined the IP '\2%s\2' because: %s",
				opertrace(finduser(source),0), ip, reason);
			notice(service, source,
				"IP '\2%s\2' has been added to the services Z-line list.", ip);
		}
	} else if (sstrcmp(cmd, "DEL") == 0) {
		if (!(sql_select("ip FROM zline WHERE ip = '%s'", ip))) {
			notice(service, source,
				"ZLINE ERROR: IP '\2%s\2' not found on Z-line list", ip);
		} else {
			send_sql("DELETE from zline WHERE ip = '%s'", ip);
			send_cmd(NULL, "UNZLINE %s", ip);
			notice(service, source, "IP '\2%s\2' has been removed from the Z-line list", ip);
			globops(NULL, "'%s' has removed the IP '\2%s\2' from the global Z-line list.",
				opertrace(finduser(source),0), ip);
		}
	} else {
		notice(service, source,
			"Syntax: ZLINE [LIST|ADD|DEL] [nick] [reason]");
		notice(service, source,
			"\2/%s OHELP ZLINE\2 for more information.",
			service);
	}

}
/**************************************************************************************/

static void do_sendlogs(const char *source, const char *service)
{
	const char *cmd = strtok(NULL, " ");
	
	if (cmd && sstrcmp(cmd, "LIST") == 0) {
		int num_rows=sql_select("oper.name,nick.email FROM oper,nick "
			"WHERE oper.name=nick.name AND oper.sendlog > 0");
		
		notice(service, source, "\2SENDLOGS list\2:");
		if (num_rows) {
			int i;
			for (i=0; i < num_rows; i++) {
				row = fetch(res);
				notice(service, source,
					"  %s -> %s", row[0], row[1]);
			}
			notice(service, source, "\2End of SENDLOGS list\2");
		} else
			notice(service, source,
				"There are no current opers receiving logs via e-mail.");
	} else {
		User *user = finduser(source);
		if (sql_select("oper.sendlog,nick.email FROM oper,nick WHERE oper.name = '%s' AND oper.name=nick.name", 
				opertrace(user,1))) {
			row = fetch(res);
			if (atoi(row[0]) > 0) {
				send_sql("UPDATE oper SET sendlog = 0 WHERE name = '%s'",
					opertrace(user,1));
				notice(service, source, "You will no longer receive logs via e-mail.");
			} else {
				send_sql("UPDATE oper SET sendlog = 1 WHERE name = '%s'",
					opertrace(user,1));
				notice(service, source, "You will now receive logs via-email.");
			}
		} else {
			notice(service, source,
				"SENDLOGS ERROR: Unknown error. Perhaps database malfunction. "
				"Get your administrator to re-add you to the oper database.");
			if (debug)
				log("Discrepency between nick and oper tables for %s", opertrace(user,1));
		}
		
	}

}

/**************************************************************************************/

static void do_settings(const char *source, const char *service)
{
	char buf[256];

	notice(service, source, "\2Basic services settings\2:");

	notice(service, source,
		" Nick / Chan / Memo Expiry: \2%d\2 / \2%d\2 / \2%d\2 days.",
			NICK_EXPIRE, CHAN_EXPIRE, MEMO_EXPIRE);


	if (UPDATE_TIMEOUT > 86400)
		notice(s_OperServ, source,
			" DataBase Expiry Freq: \2%d\2 day%s, \2%02d:%02d\2",
			UPDATE_TIMEOUT/86400, (UPDATE_TIMEOUT/86400 == 1) ? "" : "s",
			(UPDATE_TIMEOUT/3600) % 24, (UPDATE_TIMEOUT/60) % 60);
	else if (UPDATE_TIMEOUT > 3600)
		notice(s_OperServ, source,
			" DataBase Expiry Freq: \2%d hour%s, %d minute%s\2",
			UPDATE_TIMEOUT/3600, UPDATE_TIMEOUT/3600==1 ? "" : "s",
			(UPDATE_TIMEOUT/60) % 60, (UPDATE_TIMEOUT/60)%60==1 ? "" : "s");
	else
		notice(s_OperServ, source,
			" DataBase Expiry Freq: \2%d minute%s, %d second%s\2",
			UPDATE_TIMEOUT/60, UPDATE_TIMEOUT/60==1 ? "" : "s",
			UPDATE_TIMEOUT%60, UPDATE_TIMEOUT%60==1 ? "" : "s");

	notice(service, source,
		" Clone Trigger: \2%d\2  Clone Wait Warnings: \2%d\2",
		CLONE_MIN_USERS, CLONE_WARNING_DELAY);


	notice(service, source,
		" Memo Limit: \2%d\2  Max Memo Length: \2%d\2  Memo Send Wait: \2%d\2",
		MAX_MEMOS, MAX_MEMO_LENGTH, MEMO_WAIT);


	notice(service, source,
		" Maximums: Chan Access List Entries: \2%d\2   Chan Access Overall Entries: \2%d\2",
		CHAN_LIST_MAX, CHAN_ACC_MAX);


	notice(service, source,
		" Nick Release: \2%ds\2   Bad PW Limit: \2%d\2",
		RELEASE_TIMEOUT, BAD_PW_LIMIT);

	notice(service, source,
		" Flood Levels: 4/%d/%d [LEV:TLEV:RESET]", F_TLEV, F_RESET);


	*buf=0;
	if (debug) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Debug");
	}
	if (DETECT_CLONES) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Clone Detection");
	}
	if (FLOOD_PROT) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Flood Protection");
	}
	if (READ_ONLY) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Read-Only Mode");
	} else {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Read/Write Mode");
	}
	if (debug_traffic) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Debug Traffic");
	}
	if (SEND_LOGS) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Sendlogs");
	}
	if (CKILL_COUNT) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Auto-akill Clones");
	}

	notice(service, source, " Options: %s", buf);

	notice(service, source, "\2End of settings\2.");


}

/**************************************************************************************/

static void do_global(const char *source, const char *service)
{
	const char *target = strtok(NULL, " ");
	const char *msg = strtok(NULL, "");
	User *user;
	int ulist=0;


	if (!msg) {
		notice(service, source,
			"Syntax: \2GLOBAL\2 <target> <message>");
		notice(service, source,
			"\2/%s OHELP GLOBAL\2 for more information.", service);

	} else if (strlen(msg) > 450) {
		notice(service, source,
			"GLOBAL ERROR: message length is too long!");

	} else if (sstrcmp(target, "all") == 0 || sstrcmp(target, "*") == 0 ||
			sstrcmp(target, "*.*") == 0) {
		globops(NULL, "'%s' used the GLOBAL notice command [Target: ALL]",
			opertrace(finduser(source),0));
		for (ulist=0; ulist < 256; ulist++) {
			for (user = userlist[ulist]; user; user = user->next)
				notice(s_Global, user->nick,
					"\2-Global Notice-\2 %s \2-No Replies Please-\2", msg);
		}
	} else {
		globops(NULL, "'%s' used the GLOBAL notice command [Target: %s]",
			opertrace(finduser(source),0), target);

		for (ulist=0; ulist < 256; ulist++) {
			for (user=userlist[ulist]; user; user = user->next) {
				if (match_wild_nocase(target, user->host))
					notice(s_Global, user->nick,
						"\2-Global Notice-\2 %s \2-No Replies Please-\2", msg);
			}
		}
	}

}

/**************************************************************************************/

static void do_toggle(const char *source, const char *service)
{
	const char *option = gettok(NULL, " ", CMDMAX);
	User *u = finduser(source);

	if (!u)
		return;
	
	if (!option) {
		notice(service, source,
			"\2TOGGLE\2 <option>");
		notice(service, source,
			"\2/%s OHELP TOGGLE\2 for more information.", service);	

	} else if (sstrcmp(option, "CLONES") == 0) {
		if (!DETECT_CLONES) {
			DETECT_CLONES = 1;
			globops(NULL, "%s has enabled clone detection.", opertrace(u,0));
			notice(service, source, "Clone detected is now enabled.");
		} else {
			DETECT_CLONES = 0;
			globops(NULL, "%s has disabled clone detection.", opertrace(u,0));
			notice(service, source, "Clone detection is now disabled.");
		}
	
	} else if (sstrcmp(option, "SNOOP") == 0) {
		if (!SNOOP) {
			SNOOP = 1;
			notice(service, source, "SNOOP is now enabled.");
		} else {
			SNOOP = 0;
			notice(service, source, "SNOOP is now disabled.");
		}
	
	} else if (sstrcmp(option, "FLOOD") == 0) {
		if (FLOOD_PROT) {
			FLOOD_PROT = 1;
			globops(NULL, "%s has enabled flood protection.", opertrace(u,0));
			notice(service, source, "Flood protecrion is now enabled.");
		} else {
			FLOOD_PROT = 0;
			globops(NULL, "%s has disabled flood protection.", opertrace(u,0));
			notice(service, source, "Flood protection is now disabled.");
		}
	
	} else if (sstrcmp(option, "READONLY") == 0) {
		if (!READ_ONLY) {
			READ_ONLY = 1;
			globops(NULL, "%s has set services into read-only mode.", opertrace(u,0));
			notice(service, source, "Services are now in read-only mode.");
		} else if (READ_ONLY == 2) {
			notice(service, source, "Sorry, read-only mode has been hard-set by the "
				"parameters at run-time.");
		} else {
			READ_ONLY = 0;
			globops(NULL, "%s has set services back into read-write mode.", opertrace(u,0));
			notice(service, source, "Services are now set back to read-write mode.");
		}
	
	} else if (sstrcmp(option, "DEBUG") == 0) {
		if (!debug) {
			debug = 1;
			notice(service, source, "Debug mode is now enabled.");
		} else if (debug == 2) {
			notice(service, source, "Sorry, debug mode has been hard-set by the "
				"parameters at run-time.");
		} else {
			debug = 0;
			notice(service, source, "Debug mode is now disabled.");
		}
	
	} else if (sstrcmp(option, "TRAFFIC") == 0) {
		if (!debug_traffic) {
			debug_traffic = 1;
			notice(service, source, "Debug-traffic mode is now enabled.");
		} else if (debug_traffic == 2) {
			notice(service, source, "Sorry, debug-traffic mode has been hard set by the "
				"parameters at run-time.");
		} else {
			debug_traffic = 0;
			notice(service, source, "Debug-traffic is now disabled.");
		}
	
	} else if (sstrcmp(option, "SDEBUG") == 0) {
		if (!sdebug) {
			sdebug = 1;
			notice(service, source, "Snoop-debug is now enabled.");
		} else {
			sdebug = 0;
			notice(service, source, "Snoop-debug is now disabled.");
		}

#ifdef SKELETON
	} else if (sstrcmp(option, "SKELETON") == 0) {
		if (!skeleton_mode) {
			skeleton_mode = 1;
			notice(service, source, "Skeleton services are now active.");
			globops(NULL, "%s has activated back-up services for use.", opertrace(u,0));
		} else {
			skeleton_mode = 0;
			notice(service, source, "Skeleton services are now dormant.");
			globops(NULL, "%s has de-activated back-up services.", opertrace(u,0));
		}
#endif
	
	} else if (sstrcmp(option, "SENDLOGS") == 0) {
		if (!SEND_LOGS) {
			SEND_LOGS = 1;
			globops(NULL, "%s has enabled sending of e-mail log files.", opertrace(u,0));
			notice(service, source, "Services will now e-mail log files.");
		} else {
			SEND_LOGS = 0;
			globops(NULL, "%s has disabled sending of e-mail log files.", opertrace(u,0));
			notice(service, source, "Services will no longer e-mail log files.");
		}
	
		
	} else {
		notice(service, source, "Invalid TOGGLE option '%s'", option);
		notice(service, source, "\2/%s OHELP TOGGLE\2 for more information.", service);
	}

}

/**************************************************************************************/

static void do_uinfo(const char *source, const char *service)
{
	const char *nick = gettok(NULL, " ", NICKMAX);
	struct u_idlist *nl;
	struct u_founderlist *fl;
	User *u;
	int reg=0, num_rows=0, i;

	if (!nick) {
		notice(service, source,
			"\2UINFO\2 <nick>");
		notice(service, source,
			"\2/%s OHELP UINFO\2 for more information.", service);	
		return;
	}


	if (sql_select("name FROM nick WHERE name = '%s' AND flags < 0", nick)) {
		notice(service, source, "UINFO: Nick '\2%s\2' is a forbidden nick!\2", nick);
		return;
	}


	notice(service, source, "\2UINFO information on nick '%s'\2:", nick);

	u = finduser(nick);
	reg = sql_select("name FROM nick WHERE name = '%s'", nick);

	notice(service, source, "              Status : %s [%s]",
		u ? "Online" : "Offline", reg ? "Registered" : "Unregistered");


	if (u) {
		struct UserChans *cl = u->chans;
		char buf[512]; *buf=0;

		while (cl) {
			if (*buf)
				strcat(buf, ", ");
			i = strlen(buf);
			strncpy(buf+i, cl->chan->name, (sizeof(buf)-i));
			cl = cl->next;
		}

		notice(service, source, "     Active Channels : %s",
			*buf ? buf : "None.");

		*buf=0;

		for (nl = u->nicks; nl; nl = nl->next) {
			if (*buf)
				strcat(buf, ", ");
			i = strlen(buf);
			strncpy(buf+i, nl->nick, (sizeof(buf)-i));
		}
		notice(service, source, "    Identified Nicks : %s",
			*buf ? buf : "None.");

		*buf=0;
		for (fl = u->fchans; fl; fl = fl->next) {
			if (*buf)
				strcat(buf, ", ");
			i = strlen(buf);
			strncpy(buf+i, fl->chan->name, (sizeof(buf)-i));
		}
		notice(service, source, " Identified Channels : %s",
			*buf ? buf : "None.");



	}
	if (reg) {
		num_rows = sql_select("name FROM naccess WHERE name = '%s'", nick);
		notice(service, source, "   Access List Count : %d", num_rows);
	}

	if (u) {
		int temp_time = u->flood_time - time(NULL);
		if (u->flood_time < time(NULL)) {
			u->floodlev = 0;
			temp_time = 0;
		}
		notice(service, source,
			"         Flood Level : %d/%d/%d motd(%d) / version(%d)",
			u->floodlev2, u->floodlev, temp_time, u->motd, u->version);


		if (u->lastreg)
			notice(service, source,
					"   Last Registration : %s",
					calcltime(time(NULL) - u->lastreg));
		else
			notice(service, source,
				"   Last Registration : No registrations this session.");


		if (u->lastmemo) {
			notice(service, source, "  Memos this session : %d", u->memos);
			notice(service, source, "      Last Memo Send : %s",
				calcltime(time(NULL) - u->lastmemo));

		} else
			notice(service, source,
				"      Last Memo Send : No memos sent this session.");
		



		notice(service, source,
			"  Invalid Nick Passes: %d", u->invalid_nicks);
		notice(service, source,
			"  Invalid Chan Passes: %d", u->invalid_chans);

		if (u->invalid_nicks || u->invalid_chans) {
			notice(service, source,
				"   Last Invalid Time : %s",
				calcltime(time(NULL) - u->invalid_time));
		}

		notice(service, source,
			"      Current Server : %s", u->server);

		notice(service, source,
			"        Nick Changes : %d", u->nchanges);

		notice(service, source,
			"Nick has been in use : %s", calcltime(time(NULL) - u->signon)),

		temp_time = time(NULL) - u->signon;

	}

	notice(service, source, "\2End of UINFO\2");


}

/**************************************************************************************/

static void do_stats(const char *source, const char *service)
{
	unsigned int nicks = sql_select("* FROM nick");
	unsigned int chans = sql_select("* FROM chan");
	unsigned int akillsn = sql_select("* FROM akill");
	unsigned int jupes = sql_select("* FROM jupiter");
	unsigned int qlines = sql_select("* FROM sqline");
	time_t uptime = time(NULL) - start_time;
	float day = (uptime/86400) ? scommands / (uptime / 86400.0) : scommands;
	float hour = (uptime/3600) ? scommands / (uptime / 3600.0) : scommands;
	float minute = (uptime/60) ? scommands / (uptime / 60.0) : scommands;

	notice(service, source, "\2STATS for %s\2:", NETWORK_NAME);

	notice(service, source, "  Current users: \2%d\2  (\2%d\2 ops)  Max: \2%d\2",
		usercnt, opcnt, maxusercnt);

	notice(service, source,
		"  Current chans: \2%d\2  Max: \2%d\2  Monitored: \2%d\2 ", cchans, maxchans, monitored);


	notice(service, source,
		"  (Registered) Nicks: \2%d\2 Chans: \2%d\2  Akills: \2%d\2  Jupes: \2%d\2  Q-lines: \2%d\2",
			nicks, chans, akillsn, jupes, qlines);

	notice(service, source,
		" Services Commands: Total: \2%d\2 Avg/D: \2%.2f\2 Avg/H: \2%.2f\2 Avg/M: \2%.2f\2",
			scommands, day, hour, minute);


	day = (uptime/86400) ? ucommands / (uptime / 86400.0) : ucommands;
	hour = (uptime/3600) ? ucommands / (uptime / 3600.0) : ucommands;
	minute = (uptime/60) ? ucommands / (uptime / 60.0) : ucommands;

	notice(service, source,
		"     User Commands: Total: \2%d\2 Avg/D: \2%.2f\2 Avg/H: \2%.2f\2 Avg/M: \2%.2f\2",
			ucommands, day, hour, minute);


	do_uptime(source, service);

	notice(service, source, "\2End of STATS\2");


}

/**************************************************************************************/

static void do_restart(const char *source, const char *service)
{
	sshutdown = 2;
	sql_quit();
	globops(NULL, "Restarting services by request of: %s",
		opertrace(finduser(source), 0));

}

/**************************************************************************************/

#ifdef BAHAMUT

void ircd_akill(const char *source, const char *mask, const char *reason, char *code, int timed)
{
	char rbuf[512], *s;

	s = strchr(mask, '@');
	if (s)
		*s++ = 0;


	snprintf(rbuf, sizeof(rbuf), "%s (Ticket ID: %s)", 
		AKILL_REASONS ? reason : "You have been Auto-Killed.", code);

	send_cmd(NULL, "AKILL %s %s %d %s %d :%s",
		s ? s : mask, s ? mask : "*", timed, source, time(NULL), rbuf);


}

#endif


/**************************************************************************************/
/**************************************************************************************/
